////////////////////////////////////////////////////////////////////////////////
//
//      Title:          PCIDIO24.c
//
//      Date Created:   9 August 1999
//
//      Created By:     Ben-z Lawrence
//      Organization:   Control Systems Research
//
//      Target OS:      Windows 2000
//      Compiler:       Microsoft Visual C++ 6.0
//
//      Description:    Port I/O Driver for Industrial Computer Source
//                      PCI-DIO24 card series (includes H, H/CL, D, and D/CL).
//                      The PCI-DIO24 contains one Intel 8255 PPI which
//                      contains 24 TTL lines (24-bits I/O).  The ?/CL models
//                      also contain three Intel 8254 Programmable Interval
//                      Timers which provides 3 sets of 3 16-bit counters.
//                      The card has hard-wired the OUT line of the second
//                      counter of each timer to the CLOCK line of the
//                      third counter of each timer.  In addition, the
//                      user may short pins in the connector to set the
//                      OUT line of the first counter of each timer to the
//                      CLOCK line of the second counter of each timer.
//                      With certain timer counter modes, this will provide
//                      32-bit and 48-bit timers.  Also, the OUT line of the
//                      third counter of the first timer is hard-wired to the
//                      processor interrupt line.  An interrupt line may
//                      be set for both one TTL and the timer (A).  The base
//                      address and the interrupt are set by PnP and must
//                      be retrieved by the driver entry.
//
//      Modifications:
//
//      Version 1.1:    Ben-z Lawrence; August 15, 2000
//
//                      Error discovered in methodology of ISR.  Added DPC
//                      to complete ISR processing (as required by NT/2000)
//                      [function DIO_ISRDPC] and moved actual processing
//                      of thread control to the DPC.  Placed
//                      multiprocessor-safe code in DPC to prevent re-entrancy
//                      and a function [DPCInterlockOperation] to prevent
//                      concurrent access to the variable used to prevent
//                      re-entrancy.  Added queuing of the interrupts received
//                      for processing of the interrupts by the DPC and a
//                      function [DPCGetQueuedInterrupt] to retrieve queued
//                      interrupts in a thread/multiprocessor safe method.
//
//                      Added multiprocessor-safe code and functions to
//                      counter/timer set/read functions [DIOSyncCounterSet,
//                      DIOSyncCounterRead, and DIOSyncTimerRead].
//
//                      Added check to ensure interrupt events are not added
//                      or removed while interrupts are enabled.
//
//      Version 1.2:    Ben-z Lawrence; August 14, 2001
//
//                      Required capability of synchronizing the timer
//                      interrupt to the external interrupt and changing the
//                      timer counter reload values at termination of current
//                      count with new values without stopping the counters.
//
//                      The DIOWriteCounter was added.
//
//                      The DIOSetTimerInterruptPeriod was modified to set
//                      Counter A-1 to 2 and set the counter A-2 to the
//                      desired millisecond time divided by 2.  This allowed
//                      for drift compensation with a resolution of 2 ticks.
//
//      Version 1.3:    Ben-z Lawrence; December 10, 2001
//
//                      Required capability of synchronizing all the interrupt
//                      events timing with the external interrupt.  The
//                      internal tick counters can be reset (will generally
//                      be called by the user's external interrupt event
//                      handler).
//
//                      The DIOResetInterruptEvents was added.
//
////////////////////////////////////////////////////////////////////////////////
#include "precomp.h"

//
// All global variables are placed in one structure.
//
DIO_GLOBALS DIOGlobals;

// Define code sections for paging and disposal.
//
// INIT -        Only needed during initialization and then can be disposed of.
//               Used for DriverEntry() and DriverEntry supporting functions.
// DIOPAGED -    Always paged (never locked).  Used for code that is not often
//               called and has nothing to do with I/O performance.
// DIOLOCKED -   Must be locked when a device is open, else paged.  Used for
//               code that needs to be locked for performance and IRQL reasons.
//
// Note that the alloc_text pragma MUST be after function declaration (in
// header) and before function definitions (in source).  Also, the alloc_text
// pragma MUST be in the same module as the function definition.
//

#pragma alloc_text(INIT,DriverEntry)
#pragma alloc_text(DIO_PAGED,DIOSyncCompletion)
#pragma alloc_text(DIO_PAGED,DIOPoRequestComplete)
#pragma alloc_text(DIO_PAGED,DIOPowerCompletionRoutine)
#pragma alloc_text(DIO_PAGED,DIOProcessQueuedRequests)
#pragma alloc_text(DIO_PAGED,DIOAddDevice)
#pragma alloc_text(DIO_PAGED,DIOPnpDispatch)
#pragma alloc_text(DIO_PAGED,DIOPowerDispatch)
#pragma alloc_text(DIO_PAGED,DIOCreateClose)
#pragma alloc_text(DIO_PAGED,DIOCleanup)
#pragma alloc_text(DIO_PAGED,DIOUnload)
#pragma alloc_text(DIO_LOCKED,DIOIoIncrement)
#pragma alloc_text(DIO_LOCKED,DIOIoDecrement)
#pragma alloc_text(DIO_LOCKED,DIOQueueRequest)
#pragma alloc_text(DIO_LOCKED,DIOCancelQueued)
#pragma alloc_text(DIO_LOCKED,DIODispatch)
#pragma alloc_text(DIO_LOCKED,DIOSetMode)
#pragma alloc_text(DIO_LOCKED,DIOWrite)
#pragma alloc_text(DIO_LOCKED,DIOSet)
#pragma alloc_text(DIO_LOCKED,DIOReset)
#pragma alloc_text(DIO_LOCKED,DIORead)
#pragma alloc_text(DIO_LOCKED,DIOInitCounter)
#pragma alloc_text(DIO_LOCKED,DIOWriteCounter)
#pragma alloc_text(DIO_LOCKED,DIOSetCounter)
#pragma alloc_text(DIO_LOCKED,DIOSyncCounterSet)
#pragma alloc_text(DIO_LOCKED,DIOReadCounter)
#pragma alloc_text(DIO_LOCKED,DIOGetCounter)
#pragma alloc_text(DIO_LOCKED,DIOSyncCounterRead)
#pragma alloc_text(DIO_LOCKED,DIOReadTimer)
#pragma alloc_text(DIO_LOCKED,DIOGetTimer)
#pragma alloc_text(DIO_LOCKED,DIOSyncTimerRead)
#pragma alloc_text(DIO_LOCKED,DIOReadWrite)
#pragma alloc_text(DIO_LOCKED,LogDIOError)
#pragma alloc_text(DIO_LOCKED,DIOSetTimerInterruptPeriod)
#pragma alloc_text(DIO_LOCKED,DIOAddInterruptEvent)
#pragma alloc_text(DIO_LOCKED,DIORemoveInterruptEvent)
#pragma alloc_text(DIO_LOCKED,DIOResetInterruptEvents)
#pragma alloc_text(DIO_LOCKED,DIOEnableInterrupt)
#pragma alloc_text(DIO_LOCKED,DIODisableInterrupt)
#pragma alloc_text(DIO_LOCKED,DIOGetEventInfo)
#pragma alloc_text(DIO_LOCKED,DIOResetEvent)
#pragma alloc_text(DIO_LOCKED,DPCInterlockOperation)
#pragma alloc_text(DIO_LOCKED,DPCGetQueuedInterrupt)

////////////////////////////////////////////////////////////////////////////////
// Entry point for the driver.  Set the dispatch entry points in
// the driver object and initialize the global driver settings for
// driver installation by the PnP dispatch.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DriverEntry(DRIVER_OBJECT *driverObject,
                     UNICODE_STRING *registryPath)
{

    // **** Local Declarations ****

    // Status to return from NT calls.
    NTSTATUS status;

    // Pointer to locked section.
    void *lockPtr;

    // **** Code Statements ****

    // Clear the DIOGlobals.
    RtlZeroMemory(&DIOGlobals, sizeof(DIOGlobals));

    // Lock the paged code in their frames
    lockPtr = MmLockPagableCodeSection(DIOIoIncrement);

    // Save the handle to the locked paged code.
    ASSERT(DIOGlobals.PAGEDIO_Handle == NULL);
    DIOGlobals.PAGEDIO_Handle = lockPtr;
    DIOGlobals.IsLocked = TRUE;

    // Store both UNICODE_STRING registry entries.  First, the exact
    // parameter passed.
    DIOGlobals.RegistryPath.MaximumLength = registryPath->MaximumLength;
    DIOGlobals.RegistryPath.Length = registryPath->Length;
    DIOGlobals.RegistryPath.Buffer = ExAllocatePool(PagedPool,
        DIOGlobals.RegistryPath.MaximumLength);

    // If error allocating unicode string, return error status and log the
    // error.
    if (DIOGlobals.RegistryPath.Buffer == NULL)
    {
        MmUnlockPagableImageSection(lockPtr);
        DIOGlobals.IsLocked = FALSE;
        LogDIOError(driverObject, NULL, PhysicalZero, PhysicalZero, 0,
                    0, 0, 0x01, STATUS_INSUFFICIENT_RESOURCES,
                    DIO_INSUFFICIENT_RESOURCES, 0, NULL, 0, NULL);
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    // Copy the registry path to the global area.
    RtlCopyUnicodeString(&DIOGlobals.RegistryPath, registryPath);

    // The driver must have a pointer to the unload routine.  Called when
    // the driver is unloaded by the system.
    driverObject->DriverUnload = DIOUnload;

    // The driver must have an add device routine to add the device to the
    // driver.
    driverObject->DriverExtension->AddDevice = DIOAddDevice;

    // The driver object must have several dispatch routines.  Setup
    // those routines that are required.  The following are the
    // dispatch routines:
    //      IRP_MJ_CREATE ---------- Handles request for handle to device.
    //      IRP_MJ_CLOSE ----------- Handles request to close.
    //      IRP_MJ_CLEANUP --------- Handles request to get ready for a close.
    //      IRP_MJ_PNP ------------- Handles PnP requests.
    //      IRP_MJ_POWER ----------- Handles Power requests.
    //      IRP_MJ_DEVICE_CONTROL -- Handles DeviceIoControl function call.
    //      IRP_MJ_READ ------------ Handles read requests.
    //      IRP_MJ_WRITE ----------- Handles write requests.
    driverObject->MajorFunction[IRP_MJ_CREATE] = DIOCreateClose;
    driverObject->MajorFunction[IRP_MJ_CLOSE] = DIOCreateClose;
    driverObject->MajorFunction[IRP_MJ_CLEANUP] = DIOCleanup;
    driverObject->MajorFunction[IRP_MJ_PNP] = DIOPnpDispatch;
    driverObject->MajorFunction[IRP_MJ_POWER] = DIOPowerDispatch;
    driverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DIODispatch;
    driverObject->MajorFunction[IRP_MJ_READ] = DIOReadWrite;
    driverObject->MajorFunction[IRP_MJ_WRITE] = DIOReadWrite;

    // A device handle is not opened, so unlock the performance code.
    MmUnlockPagableImageSection(lockPtr);
    DIOGlobals.IsLocked = FALSE;

    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Increments the number of requests the device receives.
////////////////////////////////////////////////////////////////////////////////
LONG DIOIoIncrement(DIO_DEVICE_EXTENSION *deviceInfo)
{

    // **** Local Declarations ****

    // Value of current count.
    LONG result;

    // **** Code Statements ****

    result = InterlockedIncrement(&deviceInfo->OutstandingIO);

    // If the result is 2, the StopEvent needs to be cleared.
    if (result == 2)
        KeClearEvent(&deviceInfo->StopEvent);

    return result;
}

////////////////////////////////////////////////////////////////////////////////
// Decrements the number of requests as the request completes.
////////////////////////////////////////////////////////////////////////////////
LONG DIOIoDecrement(DIO_DEVICE_EXTENSION *deviceInfo)
{

    // **** Local Declarations ****

    // Value of current count.
    LONG            result;

    // **** Code Statements ****

    result = InterlockedDecrement(&deviceInfo->OutstandingIO);

    // Set the stop event when the value is 1.
    if (result == 1)
        KeSetEvent(&deviceInfo->StopEvent, IO_NO_INCREMENT, FALSE);
    
    // Can only be set if a remove IRP is received.  Set the remove
    // event so the device object can be deleted.
    else if (result == 0)
        KeSetEvent(&deviceInfo->RemoveEvent, IO_NO_INCREMENT, FALSE);

    return result;
}

////////////////////////////////////////////////////////////////////////////////
// Synchronization completion call-back - set event flag when complete.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOSyncCompletion(DEVICE_OBJECT *deviceObject, IRP *pIRP,
                           KEVENT *syncEvent)
{

    // **** Code Statements ****

    if (pIRP->PendingReturned)
    {
        IoMarkIrpPending(pIRP);
    }
    KeSetEvent(syncEvent, IO_NO_INCREMENT, FALSE);
    return STATUS_MORE_PROCESSING_REQUIRED;
}

////////////////////////////////////////////////////////////////////////////////
// Power Synchronization completion call-back from PoRequestPowerIrp.
////////////////////////////////////////////////////////////////////////////////
void DIOPoRequestComplete(DEVICE_OBJECT *deviceObject, UCHAR minorFunction,
                           POWER_STATE devicePowerState, IRP *pIRP,
                           IO_STATUS_BLOCK *ioStatus)
{

    // **** Local Declarations ****

    // Pointer to caller's stack location in the given IRP.  Required
    // to get parameters for the current request.
    IO_STACK_LOCATION *pIRPStack;

    // Pointer to device specific info.
    DIO_DEVICE_EXTENSION *deviceInfo;

    // System power information.
    POWER_STATE_TYPE systemPowerType;
    POWER_STATE systemPowerState;

    // **** Code Statements ****

    // Get the IRP stack (user parameters).
    pIRPStack = IoGetCurrentIrpStackLocation(pIRP);

    // Get the local data structure for the PCI-DIO24 device.
    deviceInfo = (DIO_DEVICE_EXTENSION *) deviceObject->DeviceExtension;

    // Set the system power information.
    systemPowerType = pIRPStack->Parameters.Power.Type;
    systemPowerState = pIRPStack->Parameters.Power.State;

    // Powering down.
    if (devicePowerState.DeviceState == PowerDeviceD3)
    {

        // Since a device power IRP was requested before
        // sending the system power IRP down the stack,
        // the system power IRP must be sent down now
        // and caught on the way back up.
        IoCopyCurrentIrpStackLocationToNext(pIRP);
        IoSetCompletionRoutine(pIRP, DIOPowerCompletionRoutine,
                               NULL, TRUE, TRUE, TRUE);
        PoCallDriver(deviceInfo->LowerDeviceObject, pIRP);
    }

    // Powering up.
    else
    {
    
        // Since we sent the system power IRP down the stack and requested the
        // device power IRP on the way back up the stack, complete the system
        // power IRP now.
        PoSetPowerState(deviceObject, systemPowerType, systemPowerState);
        deviceInfo->SystemPowerState = systemPowerState.SystemState;
        PoStartNextPowerIrp(pIRP);
        pIRP->IoStatus.Status = STATUS_SUCCESS;
        IoCompleteRequest(pIRP, IO_NO_INCREMENT);
        DIOIoDecrement(deviceInfo);
    }
}

////////////////////////////////////////////////////////////////////////////////
// Power completion routine call-back from PoCallDriver.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOPowerCompletionRoutine(DEVICE_OBJECT *deviceObject, IRP *pIRP,
                                   void *context)
{

    // **** Local Declarations ****

    // Pointer to caller's stack location in the given IRP.  Required
    // to get parameters for the current request.
    IO_STACK_LOCATION *pIRPStack;

    // Pointer to device specific info.
    DIO_DEVICE_EXTENSION *deviceInfo;

    // System power information.
    POWER_STATE_TYPE powerType;
    POWER_STATE powerState;

    // **** Code Statements ****

    // Get the IRP stack (user parameters).
    pIRPStack = IoGetCurrentIrpStackLocation(pIRP);

    // Get the local data structure for the PCI-DIO24 device.
    deviceInfo = (DIO_DEVICE_EXTENSION *) deviceObject->DeviceExtension;

    // Set the system power information.
    powerType = pIRPStack->Parameters.Power.Type;
    powerState = pIRPStack->Parameters.Power.State;

    // Process only if the status is STATUS_SUCCESS.
    if (NT_SUCCESS(pIRP->IoStatus.Status))
    {

        // Switch on system or device power change.
        switch (powerType)
        {

            // System power change.
            case SystemPowerState:
            {

                // Powering down.  At the completion of a system power IRP.
                // Set the system power state in the device information.
                if (powerState.SystemState > PowerSystemWorking)
                {
                    deviceInfo->SystemPowerState = powerState.SystemState;
                    PoSetPowerState(deviceInfo->DeviceObject, powerType,
                        powerState);
                }

                // Powering up.  Request a device power IRP and do not
                // complete this IRP until the device IRP is complete.
                else
                {

                    powerState.DeviceState = PowerDeviceD0;
                    PoRequestPowerIrp(deviceInfo->DeviceObject,
                        IRP_MN_SET_POWER, powerState, DIOPoRequestComplete,
                        pIRP, NULL);
                    return STATUS_MORE_PROCESSING_REQUIRED;
                }
                break;
            }  // case SystemPowerState:

            // Device power change.  Must be a power up.
            case DevicePowerState:
            {

                // Set the device to normal operation.
                deviceInfo->DevicePowerState = powerState.DeviceState;
                PoSetPowerState(deviceInfo->DeviceObject, powerType,
                    powerState);
                deviceInfo->HoldNewRequests = FALSE;
                DIOProcessQueuedRequests(deviceInfo);
                break;
            }  // case DevicePowerState:
        }  // switch (powerType)

        // Enable future power IRPs.
        PoStartNextPowerIrp(pIRP);
        DIOIoDecrement(deviceInfo);
    }
    return pIRP->IoStatus.Status;
}

////////////////////////////////////////////////////////////////////////////////
// Queue IRP.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOQueueRequest(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP)
{

    // **** Local Declarations ****

    // Status to return from NT calls.
    NTSTATUS status;

    // Current IRQ Level.
    KIRQL oldIrql;

    // **** Code Statements ****

    // Acquire a spinlock to manipulate the queue list.
    KeAcquireSpinLock(&deviceInfo->QueueLock, &oldIrql);

    // Must set the cancel routine before checking the
    // cancel flag.
    IoSetCancelRoutine(pIRP, DIOCancelQueued);

    // Check if IRP has already been cancelled.
    if (pIRP->Cancel && IoSetCancelRoutine(pIRP, NULL))
    {

        // IRP has been cancelled, but the cancel routine is not
        // yet called.  Relase the spinlock and complete the IRP.
        KeReleaseSpinLock(&deviceInfo->QueueLock, oldIrql);
        pIRP->IoStatus.Status = STATUS_CANCELLED;
        pIRP->IoStatus.Information = 0;
        IoCompleteRequest(pIRP, IO_NO_INCREMENT);
        status = STATUS_CANCELLED;
    }
    else
    {

        // Mark the IRP as pending.
        IoMarkIrpPending(pIRP);
        InsertTailList(&deviceInfo->NewRequestsQueue,
            &pIRP->Tail.Overlay.ListEntry);
        KeReleaseSpinLock(&deviceInfo->QueueLock, oldIrql);
        status = STATUS_SUCCESS;
    }
    return status;
}

////////////////////////////////////////////////////////////////////////////////
// Cancel queued IRP call-back.
////////////////////////////////////////////////////////////////////////////////
void DIOCancelQueued(DEVICE_OBJECT *deviceObject, IRP *pIRP)
{

    // **** Local Declarations ****

    // Status of kernel calls.
    NTSTATUS status;

    // Pointer to device specific info.
    DIO_DEVICE_EXTENSION *deviceInfo;

    // Pointers to IRQs in list.
    LIST_ENTRY *thisEntry;
    LIST_ENTRY *nextEntry;
    LIST_ENTRY *listHead;

    // Pointers to pending IRQs.
    IRP *pendingIRP;
    IRP *cancelIRP;

    // **** Code Statements ****

    // Get the local data structure for the PCI-DIO24 device.
    deviceInfo = (DIO_DEVICE_EXTENSION *) deviceObject->DeviceExtension;

    // Get a local spinlock before releasing the global cancel
    // spinlock.
    KeAcquireSpinLockAtDpcLevel(&deviceInfo->QueueLock);
    IoReleaseCancelSpinLock(KeGetCurrentIrql());

    // Go though the queue and find this cancelled IRP.  If
    // not found in the queue, ignore this request (should
    // not happen!).
    listHead = &deviceInfo->NewRequestsQueue;
    cancelIRP = NULL;
    for (thisEntry = listHead->Flink;
        thisEntry != listHead;
        thisEntry = thisEntry->Flink)
    {
        pendingIRP = CONTAINING_RECORD(thisEntry, IRP, Tail.Overlay.ListEntry);

        // If this is the IRP to be cancelled, remove from the queue.
        if (pendingIRP == pIRP)
        {
            RemoveEntryList(thisEntry);
            cancelIRP = pendingIRP;
            break;
        }
    }
 
    // Release the spinlock.
    KeReleaseSpinLock(&deviceInfo->QueueLock, pIRP->CancelIrql);

    // If the cancelled IRP was found in the queue, complete the IRP.
    if (cancelIRP)
    {
        cancelIRP->IoStatus.Information = 0;
        cancelIRP->IoStatus.Status = STATUS_CANCELLED;
        IoCompleteRequest(cancelIRP, IO_NO_INCREMENT);
    }
    return;
} 

////////////////////////////////////////////////////////////////////////////////
// Process queued IRPs.
////////////////////////////////////////////////////////////////////////////////
VOID DIOProcessQueuedRequests(DIO_DEVICE_EXTENSION *deviceInfo)
{

    // **** Local Declarations ****

    // Current IRQ Level.
    KIRQL oldIRQL;

    // Pointer to IRQ in list.
    PLIST_ENTRY listEntry;

    // Pointer to pending IRQ.
    IRP *nextIRP;

    // **** Code Statements ****

    // Acquire the queue lock before manipulation the list.
    KeAcquireSpinLock(&deviceInfo->QueueLock, &oldIRQL);

    // Loop through all the queued IRPs.
    while (!IsListEmpty(&deviceInfo->NewRequestsQueue))
    {

        // Get next IRP from queue.
        listEntry = RemoveHeadList(&deviceInfo->NewRequestsQueue);
        nextIRP = CONTAINING_RECORD(listEntry, IRP, Tail.Overlay.ListEntry);

        // If IRP was cancelled but the cancel routine not called so
        // cancel the IRP now.
        if (nextIRP->Cancel && IoSetCancelRoutine(nextIRP, NULL))
        {

            // Release the spinlock, complete the cancel and reinstate
            // the spinlock.
            KeReleaseSpinLock(&deviceInfo->QueueLock, oldIRQL);
            nextIRP->IoStatus.Information = 0;
            nextIRP->IoStatus.Status = STATUS_CANCELLED;
            IoCompleteRequest(nextIRP, IO_NO_INCREMENT);
            KeAcquireSpinLock(&deviceInfo->QueueLock, &oldIRQL);
        }

        // Process the IRP.  If the device had been deleted, fail the IRP.
        else
        {
            DIOIoIncrement(deviceInfo);
            if (deviceInfo->DevicePnPState == Deleted)
            {
                nextIRP->IoStatus.Information = 0;
                nextIRP->IoStatus.Status = STATUS_DELETE_PENDING;
                IoCompleteRequest(nextIRP, IO_NO_INCREMENT);
            }
            else
            {
                IoSkipCurrentIrpStackLocation(nextIRP);
                IoCallDriver(deviceInfo->LowerDeviceObject, nextIRP);
            }
            DIOIoDecrement(deviceInfo);
        }
    }
    KeReleaseSpinLock(&deviceInfo->QueueLock, oldIRQL);
    return;
}

////////////////////////////////////////////////////////////////////////////////
// Create the device object.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOAddDevice(DRIVER_OBJECT *driverObject,
                      DEVICE_OBJECT *physicalDeviceObject)
{

    // **** Local Declarations ****

    // Status of kernel calls.
    NTSTATUS status;

    // Pointer to new device object.
    DEVICE_OBJECT *deviceObject;

    // Pointer to device specific info.
    DIO_DEVICE_EXTENSION *deviceInfo;

    // Desired power state of the device.
    POWER_STATE powerState;

    // **** Code Statements ****

    // Ensure there is a device to attach.
    if (physicalDeviceObject == NULL)
    {
        LogDIOError(driverObject, NULL, PhysicalZero, PhysicalZero, 0,
                    0, 0, 0x10, STATUS_NO_MORE_ENTRIES,
                    DIO_CANNOT_ADD_DEVICE, 0, NULL, 0, NULL);
        return STATUS_NO_MORE_ENTRIES;
    }

    // Create the device object
    status = IoCreateDevice(driverObject, sizeof(DIO_DEVICE_EXTENSION),
                            NULL, DIO_TYPE, 0, FALSE, &deviceObject);
    if (!NT_SUCCESS(status))
    {
        LogDIOError(driverObject, deviceObject, PhysicalZero, PhysicalZero, 0,
            0, 0, 0x11, status, DIO_CANNOT_CREATE_DEVICE, 0, NULL, 0, NULL);
        return status;
    }

    // Get a pointer to the device extension which was created by
    // IoCreateDevice.  This is non-paged and will contain the
    // DIO_DEVICE_EXTENSION structure.  Clear the structure.
    deviceInfo = deviceObject->DeviceExtension;
    RtlZeroMemory(deviceInfo, sizeof(DIO_DEVICE_EXTENSION));

    // Initialize the device extension.
    deviceInfo->DriverObject = driverObject;
    deviceInfo->PhysicalDeviceObject = physicalDeviceObject;
    deviceInfo->DeviceObject = deviceObject;
    deviceInfo->DevicePnPState = NotStarted;
    deviceInfo->CurrentDeviceState = 0;
    deviceInfo->HoldNewRequests = FALSE;
    deviceInfo->InterruptEventInfoCount = 0;
    deviceInfo->InterruptEventInfoList = NULL;
    InitializeListHead(&deviceInfo->NewRequestsQueue);
    KeInitializeSpinLock(&deviceInfo->QueueLock);
    KeInitializeSpinLock(&deviceInfo->InterruptSpinLock);
    KeInitializeEvent(&deviceInfo->RemoveEvent, SynchronizationEvent, FALSE);
    KeInitializeEvent(&deviceInfo->StopEvent, SynchronizationEvent, TRUE);
    deviceInfo->OutstandingIO = 1;

    // Allocate non-paged memory for an array of UCHARs for each port of
    // the PPI.  This will be the PPIPortValue array.  This MUST be freed
    // during any subsequent failure in loading the driver and during driver
    // unload.  Note there are only 4 hardware ports (A, B, C and control)
    // on a 8255 PPI and the PCI-DIO24 only has one PPI.
    deviceInfo->PPIPortValue = (UCHAR *) ExAllocatePool(NonPagedPool, 4);
    if (deviceInfo->PPIPortValue == NULL)
    {
        LogDIOError(driverObject, NULL, PhysicalZero, PhysicalZero, 0,
            0, 0, 0x12, STATUS_INSUFFICIENT_RESOURCES,
            DIO_INSUFFICIENT_RESOURCES, 0, NULL, 0, NULL);
        IoDeleteDevice(deviceObject);
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    // Clear the array.
    RtlZeroMemory(&deviceInfo->PPIPortValue[0], 4);

    // Allocate memory for interrupt processing queue (added by
    // ISR and processed by DPC).
    deviceInfo->InterruptQueue =
        (INTERRUPT_QUEUE *) ExAllocatePool(NonPagedPool,
        (INTERRUPT_QUEUE_LENGTH * sizeof(INTERRUPT_QUEUE)));
    if (deviceInfo->InterruptQueue == NULL)
    {
        LogDIOError(driverObject, NULL, PhysicalZero, PhysicalZero, 0,
            0, 0, 0x13, STATUS_INSUFFICIENT_RESOURCES,
            DIO_INSUFFICIENT_RESOURCES, 0, NULL, 0, NULL);
        ExFreePool(deviceInfo->PPIPortValue);
        IoDeleteDevice(deviceObject);
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    // Clear the queue.
    RtlZeroMemory(deviceInfo->InterruptQueue,
        (INTERRUPT_QUEUE_LENGTH * sizeof(INTERRUPT_QUEUE)));

    // Initialize the first data of the queue to be the top
    // of the queue, the first free position to be the
    // top of the queue, and the end of the queue to be
    // one past the last position in the queue.  The queue
    // will be handled as a circular queue.  If the first
    // free position is the same as the first data
    // data position, the queue is defined to be empty.
    deviceInfo->InterruptQueueFirst = deviceInfo->InterruptQueue;
    deviceInfo->InterruptQueueNext = deviceInfo->InterruptQueue;
    deviceInfo->InterruptQueueEnd = deviceInfo->InterruptQueue +
        INTERRUPT_QUEUE_LENGTH;
    deviceInfo->InterruptsQueued = 0;

    // Specify use of buffered IO and power pagable.
    deviceObject->Flags |= DO_BUFFERED_IO | DO_POWER_PAGABLE;

    // Reset the device initializing flag.
    deviceObject->Flags &= ~DO_DEVICE_INITIALIZING;

    // Initialize the power state as up and running (as this should be
    // the case during the AddDevice routine).
    deviceInfo->DevicePowerState = PowerDeviceD0;
    deviceInfo->SystemPowerState = PowerSystemWorking;

    // Set the initial power state.
    powerState.DeviceState = PowerDeviceD0;
    PoSetPowerState(deviceObject, DevicePowerState, powerState);

    // Attach the device on top of the lower device object.  The
    // returned value is the device object to which it is attached.
    deviceInfo->LowerDeviceObject = IoAttachDeviceToDeviceStack(deviceObject,
        physicalDeviceObject);
    if (deviceInfo->LowerDeviceObject == NULL)
    {
        LogDIOError(driverObject, NULL, PhysicalZero, PhysicalZero, 0,
            0, 0, 0x14, STATUS_UNSUCCESSFUL,
            DIO_CANNOT_ATTACH_DEVICE, 0, NULL, 0, NULL);
        ExFreePool(deviceInfo->InterruptQueue);
        ExFreePool(deviceInfo->PPIPortValue);
        IoDeleteDevice(deviceObject);
        return STATUS_UNSUCCESSFUL;
    }

    // Create a device interface.
    status = IoRegisterDeviceInterface(
        deviceInfo->PhysicalDeviceObject, (LPGUID) &GUID_CLASS_DIO,
        NULL, &deviceInfo->DeviceClassSymbolicName);
    if (!NT_SUCCESS(status))
    {
        LogDIOError(driverObject, deviceObject, PhysicalZero, PhysicalZero, 0,
            0, 0, 0x15, status, DIO_NO_DEVICE_CLASS, 0, NULL, 0, NULL);
        IoDetachDevice(deviceInfo->LowerDeviceObject);
        ExFreePool(deviceInfo->InterruptQueue);
        ExFreePool(deviceInfo->PPIPortValue);
        IoDeleteDevice(deviceObject);
        return status;
    }

    // Initialize the Deferred Procedure Call (DPC) for the Interrupt Service
    // Routine (DPC) (DpcForIsr) function [DIO_ISRDPC] to complete ISR
    // processing at the DISPATCH IRQL.  At least one DPC is required for
    // all Windows NT/2000 ISRs.  This identifies the standard one to be
    // used for the only ISR used by the PCIDIO24 driver.  Note there are
    // no custom DPCs used by the PCIDIO24 driver.
    IoInitializeDpcRequest(deviceObject, DIO_ISRDPC);

    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Dispatch plug and play.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOPnpDispatch(DEVICE_OBJECT *deviceObject, IRP *pIRP)

{

    // **** Local Declarations ****

    // Status of kernel calls.
    NTSTATUS status;

    // Pointer to caller's stack location in the given IRP.  Required
    // to get parameters for the current request.
    IO_STACK_LOCATION *pIRPStack;

    // Pointer to device specific info.
    DIO_DEVICE_EXTENSION *deviceInfo;

    // Points to the associated driver object.
    DRIVER_OBJECT *driverObject;

    // **** Code Statements ****

    // Get the IRP stack (user parameters).
    pIRPStack = IoGetCurrentIrpStackLocation(pIRP);

    // Get the local data structure for the PCI-DIO24 device.
    deviceInfo = (DIO_DEVICE_EXTENSION *) deviceObject->DeviceExtension;

    // Get the associated driver object.
    driverObject = deviceInfo->DriverObject;

    // If the device is removed, fail the IRP.
    if (deviceInfo->DevicePnPState == Deleted)
    {
        pIRP->IoStatus.Status = STATUS_DELETE_PENDING;
        IoCompleteRequest(pIRP, IO_NO_INCREMENT);
        return STATUS_DELETE_PENDING;
    }

    // Increment to show a pending IRP.
    DIOIoIncrement(deviceInfo);

    // Switch on the MinorFunction for the request.
    switch (pIRPStack->MinorFunction)
    {

        // Handle PnP start device request.
        case IRP_MN_START_DEVICE:
        {

            // **** Local Declarations ****

            // Synchronization event.
            KEVENT event;

            // Handle to resource key.
            HANDLE pnpKey;

            // Unicode string of desired resource key.
            UNICODE_STRING keyName;

            // Length of desired resource key.
            ULONG fullKeyInfoLength;

            // Full information structure of desired resource key.
            KEY_VALUE_FULL_INFORMATION *fullKeyInfo;

            // Value of desired resource key.
            ULONG keyValue;

            // Number of I/O ports expected.
            ULONG numPorts;

            // Pointer to full resource descriptor containing
            // all allocated resources.
            CM_FULL_RESOURCE_DESCRIPTOR *fullDesc;

            // Pointer to partial resource list containing
            // all allocated resources.
            CM_PARTIAL_RESOURCE_LIST *partialList;

            // Pointer to partial resource descriptor containing
            // an allocated resource.
            CM_PARTIAL_RESOURCE_DESCRIPTOR *partialDesc;

            // Pointer to full resource descriptor containing
            // all allocated resources translated from bus specific.
            CM_FULL_RESOURCE_DESCRIPTOR *fullTrDesc;

            // Pointer to partial resource list containing
            // all allocated resources translated from bus specific.
            CM_PARTIAL_RESOURCE_LIST *partialTrList;

            // Pointer to partial resource descriptor containing
            // an allocated resource translated from bus specific.
            CM_PARTIAL_RESOURCE_DESCRIPTOR *partialTrDesc;

            // Count and counter of resources in partial resource list.
            ULONG count;
            ULONG i;

            // **** Code Statements ****

            // The device is starting.  Cannot touch the hardware yet, must
            // queue any IRPs.
            deviceInfo->HoldNewRequests = TRUE;

            // Lock the code needed for performance.
            MmLockPagableSectionByHandle(DIOGlobals.PAGEDIO_Handle);
            DIOGlobals.IsLocked = TRUE;

            // Send the IRP to the next lower driver in chain and wait
            // for completion (via DIOSyncCompletion call-back).
            KeInitializeEvent(&event, SynchronizationEvent, FALSE);
            IoCopyCurrentIrpStackLocationToNext(pIRP);
            IoSetCompletionRoutine(pIRP, DIOSyncCompletion, &event,
                TRUE, TRUE, TRUE);
            status = IoCallDriver(deviceInfo->LowerDeviceObject, pIRP);
            if (status == STATUS_PENDING)
            {
                KeWaitForSingleObject(&event, Executive, KernelMode, FALSE,
                    NULL);
                status = pIRP->IoStatus.Status;
            }
            else if (!NT_SUCCESS(status))
            {
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = status;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return status;
            }

            // Get access to the device registry.
            status = IoOpenDeviceRegistryKey(deviceInfo->PhysicalDeviceObject,
                PLUGPLAY_REGKEY_DRIVER, STANDARD_RIGHTS_WRITE, &pnpKey);
            if (!NT_SUCCESS(status))
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x20, status,
                    DIO_CANNOT_GET_DEVICE_REGISTRY_KEY, 0, NULL, 0, NULL);
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = status;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return status;
            }

            // Create a UNICODE_STRING of the key name (use the pointer to
            // the define directly).
            RtlInitUnicodeString(&keyName, HAS_CLOCK_KEY_NAME);

            // Create a buffer for the registry information.
            fullKeyInfoLength = sizeof(KEY_VALUE_FULL_INFORMATION) +
                sizeof(HAS_CLOCK_KEY_NAME) + sizeof(keyValue);
            fullKeyInfo = ExAllocatePool(PagedPool, fullKeyInfoLength);
            if (fullKeyInfo == NULL)
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x21, STATUS_INSUFFICIENT_RESOURCES,
                    DIO_INSUFFICIENT_RESOURCES, 0, NULL, 0, NULL);
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return STATUS_INSUFFICIENT_RESOURCES;
            }

            // Get the registry setting whether a clock is on the device.
            status = ZwQueryValueKey(pnpKey, &keyName, KeyValueFullInformation,
                fullKeyInfo, fullKeyInfoLength, &fullKeyInfoLength);
            if (!NT_SUCCESS(status))
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x22, status,
                    DIO_CANNOT_READ_DEVICE_REGISTRY_KEY,
                    keyName.Length + sizeof(WCHAR), keyName.Buffer, 0, NULL);
                ExFreePool(fullKeyInfo);
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = status;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return status;
            }

            // Check to be sure the data length is proper.
            if (fullKeyInfo->DataLength != sizeof(keyValue))
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x23, STATUS_UNSUCCESSFUL,
                    DIO_INVALID_DEVICE_REGISTRY_KEY,
                    keyName.Length + sizeof(WCHAR), keyName.Buffer, 0, NULL);
                ExFreePool(fullKeyInfo);
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = STATUS_UNSUCCESSFUL;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return STATUS_UNSUCCESSFUL;
            }

            // Get the key value.
            RtlCopyMemory(&keyValue, ((char *) fullKeyInfo) +
                fullKeyInfo->DataOffset, fullKeyInfo->DataLength);

            // Free the memory.
            ExFreePool(fullKeyInfo);

            // Set the clock flag.
            if (keyValue == 0)
                deviceInfo->HasClock = FALSE;
            else
                deviceInfo->HasClock = TRUE;

            // Get the acceptable number of ports for this device.
            if (deviceInfo->HasClock)
                numPorts = NUM_PORTS_WITH_CLOCK;
            else
                numPorts = NUM_PORTS_NO_CLOCK;

            // Be sure the resources and translated resources are not NULL.
            if ((pIRPStack->Parameters.StartDevice.
                    AllocatedResources == NULL) ||
                (pIRPStack->Parameters.StartDevice.
                    AllocatedResourcesTranslated == NULL))
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x24, STATUS_INSUFFICIENT_RESOURCES,
                    DIO_NO_RESOURCE_LIST, 0, NULL, 0, NULL);
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return STATUS_INSUFFICIENT_RESOURCES;
            }

            // Get pointers to full resource descriptor.
            fullDesc = &pIRPStack->Parameters.StartDevice.
                    AllocatedResources->List[0];
            fullTrDesc = &pIRPStack->Parameters.StartDevice.
                    AllocatedResourcesTranslated->List[0];

            // Take apart full resource descriptor for non-translated.
            if (fullDesc)
            {
                partialList = &fullDesc->PartialResourceList;
                partialDesc = partialList->PartialDescriptors;
                count = partialList->Count;
                deviceInfo->InterfaceType = fullDesc->InterfaceType;
                deviceInfo->BusNumber = fullDesc->BusNumber;

                // Loop for all resource descriptors.
                for (i = 0; i < count; i++, partialDesc++)
                {
                    switch (partialDesc->Type)
                    {

                        // Port resource.  Check to see if it is the
                        // correct number of ports and store the port
                        // information.
                        case CmResourceTypePort:
                        {
                            if (partialDesc->u.Port.Length == numPorts)
                            {
                                deviceInfo->DeviceBusAddress =
                                    partialDesc->u.Port.Start;
                                deviceInfo->DeviceBusAddressLength =
                                    partialDesc->u.Port.Length;
                                deviceInfo->PortMemoryType =
                                    partialDesc->Flags;
                            }
                            break;
                        }

                        // Interrupt resource.  Must only be one
                        // interrupt resource.
                        case CmResourceTypeInterrupt:
                        {
                            deviceInfo->DeviceIRQL = (KIRQL)
                                partialDesc->u.Interrupt.Level;
                            deviceInfo->DeviceIRQV =
                                partialDesc->u.Interrupt.Vector;
                            deviceInfo->DeviceIRQAffinity =
                                partialDesc->u.Interrupt.Affinity;
                            if (partialDesc->Flags &
                                CM_RESOURCE_INTERRUPT_LATCHED)
                                deviceInfo->DeviceInterruptMode =
                                    Latched;
                            else
                                deviceInfo->DeviceInterruptMode =
                                    LevelSensitive;
                            break;
                        }

                        default:
                            break;
                    }   // switch (partialDesc->Type)
                }       // for (i = 0; i < count; i++, partialDesc++)
            }           // if (fullDesc)

            // If the resources were not found, it is an error.
            if ((deviceInfo->DeviceBusAddressLength == 0) ||
                (deviceInfo->DeviceIRQL == 0))
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x25, STATUS_INSUFFICIENT_RESOURCES,
                    DIO_UNASSIGNED_RESOURCES, 0, NULL, 0, NULL);
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return STATUS_INSUFFICIENT_RESOURCES;
            }

            // Take apart full resource descriptor for translated.
            if (fullTrDesc)
            {
                partialTrList = &fullTrDesc->PartialResourceList;
                partialTrDesc = partialTrList->PartialDescriptors;
                count = partialTrList->Count;
                deviceInfo->InterfaceType = fullTrDesc->InterfaceType;
                deviceInfo->BusNumber = fullTrDesc->BusNumber;

                // Loop for all resource descriptors.
                for (i = 0; i < count; i++, partialTrDesc++)
                {
                    switch (partialTrDesc->Type)
                    {

                        // Port resource.  Check to see if it is the
                        // correct number of ports and store the port
                        // information.
                        case CmResourceTypePort:
                        {
                            if (partialTrDesc->u.Port.Length == numPorts)
                            {
                                deviceInfo->DeviceTrBusAddress =
                                    partialTrDesc->u.Port.Start;
                                deviceInfo->DeviceTrBusAddressLength =
                                    partialTrDesc->u.Port.Length;
                                deviceInfo->PortMemoryType =
                                    partialTrDesc->Flags;
                            }
                            break;
                        }

                        // Interrupt resource.  Must only be one
                        // interrupt resource.
                        case CmResourceTypeInterrupt:
                        {
                            deviceInfo->DeviceTrIRQL = (KIRQL)
                                partialTrDesc->u.Interrupt.Level;
                            deviceInfo->DeviceTrIRQV =
                                partialTrDesc->u.Interrupt.Vector;
                            deviceInfo->DeviceIRQAffinity =
                                partialTrDesc->u.Interrupt.Affinity;
                            break;
                        }

                        default:
                            break;
                    }   // switch (partialTrDesc->Type)
                }       // for (i = 0; i < count; i++, partialTrDesc++)
            }           // if (fullTrDesc)

            // If the resources were not found, it is an error.
            if ((deviceInfo->DeviceTrBusAddressLength == 0) ||
                (deviceInfo->DeviceTrIRQL == 0))
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x26, STATUS_INSUFFICIENT_RESOURCES,
                    DIO_UNASSIGNED_RESOURCES, 0, NULL, 0, NULL);
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return STATUS_INSUFFICIENT_RESOURCES;
            }

            // Double check all values and log error if not correct.
            if (!deviceInfo->DeviceBusAddress.LowPart)
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x27, STATUS_INSUFFICIENT_RESOURCES,
                    DIO_PROBLEM_RESOURCE, 0, NULL, 0, NULL);
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return STATUS_INSUFFICIENT_RESOURCES;
            }

            if (!deviceInfo->DeviceTrBusAddress.LowPart)
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x28, STATUS_INSUFFICIENT_RESOURCES,
                    DIO_PROBLEM_RESOURCE, 0, NULL, 0, NULL);
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return STATUS_INSUFFICIENT_RESOURCES;
            }

            if (!deviceInfo->DeviceIRQL)
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x29, STATUS_INSUFFICIENT_RESOURCES,
                    DIO_PROBLEM_RESOURCE, 0, NULL, 0, NULL);
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return STATUS_INSUFFICIENT_RESOURCES;
            }

            if (!deviceInfo->DeviceTrIRQL)
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x2a, STATUS_INSUFFICIENT_RESOURCES,
                    DIO_PROBLEM_RESOURCE, 0, NULL, 0, NULL);
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return STATUS_INSUFFICIENT_RESOURCES;
            }

            if (!deviceInfo->DeviceIRQV)
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x2b, STATUS_INSUFFICIENT_RESOURCES,
                    DIO_PROBLEM_RESOURCE, 0, NULL, 0, NULL);
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return STATUS_INSUFFICIENT_RESOURCES;
            }

            if (!deviceInfo->DeviceTrIRQV)
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x2c, STATUS_INSUFFICIENT_RESOURCES,
                    DIO_PROBLEM_RESOURCE, 0, NULL, 0, NULL);
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return STATUS_INSUFFICIENT_RESOURCES;
            }

            if (!deviceInfo->InterfaceType >= MaximumInterfaceType)
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x2d, STATUS_INSUFFICIENT_RESOURCES,
                    DIO_PROBLEM_RESOURCE, 0, NULL, 0, NULL);
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return STATUS_INSUFFICIENT_RESOURCES;
            }

            // The flag return for the port memory type is supposed to only
            // be CM_RESOURCE_PORT_MEMORY (0x00) or CM_RESOURCE_PORT_IO (0x01).
            // However, this ended up being a bit field (not documented as
            // such) with other bits used for undocumented purposes.
            // Therefore, make it either 0 or 1 depending upon the lowest
            // order bit.
            if (deviceInfo->PortMemoryType & 0x00000001)
                deviceInfo->PortMemoryType = 1;
            else
                deviceInfo->PortMemoryType = 0;

            // Map the memory for the ports into virtual memory.
            if (!deviceInfo->PortMemoryType)
            {
                deviceInfo->VirtualAddress = MmMapIoSpace(
                    deviceInfo->DeviceTrBusAddress,
                    deviceInfo->DeviceTrBusAddressLength, FALSE);
                if (deviceInfo->VirtualAddress)
                    deviceInfo->Mapped = TRUE;
                else
                    deviceInfo->Mapped = FALSE;
            }
            else
            {
                deviceInfo->VirtualAddress =
                    (VOID *) deviceInfo->DeviceTrBusAddress.LowPart;
                deviceInfo->Mapped = FALSE;
            }

            // If could not map, generate an error.
            if (!deviceInfo->VirtualAddress)
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x2e, STATUS_INSUFFICIENT_RESOURCES,
                    DIO_NO_VIRTUAL_ADDRESS, 0, NULL, 0, NULL);
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return STATUS_INSUFFICIENT_RESOURCES;
            }

            // In the device information area, place the offsets to the
            // specific hardware addresses.
            deviceInfo->PPIBaseAddress =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         PPI_OFFSET);
            deviceInfo->BufferEnableAddress =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         BUFFER_ENABLE_OFFSET);
            deviceInfo->DisableIRQAddress =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         IRQ_DISABLE_OFFSET);
            deviceInfo->EnableIRQAddress =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         IRQ_ENABLE_OFFSET);
            deviceInfo->ClearIRQAddress =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         IRQ_CLEAR_OFFSET);
            deviceInfo->CounterA0 =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         TIMER_A_OFFSET);
            deviceInfo->CounterA1 =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         TIMER_A_OFFSET + 1);
            deviceInfo->CounterA2 =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         TIMER_A_OFFSET + 2);
            deviceInfo->TimerControlA =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         TIMER_A_OFFSET + 3);
            deviceInfo->CounterB0 =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         TIMER_B_OFFSET);
            deviceInfo->CounterB1 =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         TIMER_B_OFFSET + 1);
            deviceInfo->CounterB2 =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         TIMER_B_OFFSET + 2);
            deviceInfo->TimerControlB =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         TIMER_B_OFFSET + 3);
            deviceInfo->CounterC0 =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         TIMER_C_OFFSET);
            deviceInfo->CounterC1 =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         TIMER_C_OFFSET + 1);
            deviceInfo->CounterC2 =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         TIMER_C_OFFSET + 2);
            deviceInfo->TimerControlC =
                (PVOID) ((ULONG) deviceInfo->VirtualAddress +
                         TIMER_C_OFFSET + 3);

            // Get ISR.
            deviceInfo->DIOIsr = DIO_ISR;
            deviceInfo->DIOIsrContext = deviceObject;

            // Disable interrupt handling by the interrupt handler.
            deviceInfo->InterruptHandled = FALSE;
            deviceInfo->ClockInterruptSavedTimerBValue = 0;

            // There is no queued DPC for the ISR.
            deviceInfo->DPCInterlock = FALSE;

            // Connect to interrupt.
            if ((!deviceInfo->Interrupt) && (deviceInfo->DIOIsr))
            {

                // Connect to the interrupt service routine.
                status = IoConnectInterrupt(&deviceInfo->Interrupt,
                                            deviceInfo->DIOIsr,
                                            deviceInfo->DIOIsrContext,
                                            &deviceInfo->InterruptSpinLock,
                                            deviceInfo->DeviceTrIRQV,
                                            deviceInfo->DeviceTrIRQL,
                                            deviceInfo->DeviceTrIRQL,
                                            deviceInfo->DeviceInterruptMode,
                                            TRUE,
                                            deviceInfo->DeviceIRQAffinity,
                                            FALSE);
                if (!NT_SUCCESS(status))
                {
                    LogDIOError(driverObject, deviceObject, PhysicalZero,
                        PhysicalZero, 0, 0, 0, 0x2f, status,
                        DIO_UNREPORTED_IRQL_CONFLICT, 0, NULL, 0, NULL);
                    deviceInfo->HoldNewRequests = FALSE;
                    pIRP->IoStatus.Status = DIO_UNREPORTED_IRQL_CONFLICT;
                    IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                    DIOIoDecrement(deviceInfo);
                    MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                    DIOGlobals.IsLocked = FALSE;
                    return DIO_UNREPORTED_IRQL_CONFLICT;
                }
            }

            // Mark as started and allow IRPs.
            deviceInfo->DevicePnPState = Started;
            deviceInfo->HoldNewRequests = FALSE;

            // Enable the device interface.
            status = IoSetDeviceInterfaceState(
                &deviceInfo->DeviceClassSymbolicName, TRUE);
            if (!NT_SUCCESS(status))
            {
                LogDIOError(driverObject, deviceObject, PhysicalZero,
                    PhysicalZero, 0, 0, 0, 0x30, status, DIO_NO_ASSOCIATION,
                    0, NULL, 0, NULL);
                pIRP->IoStatus.Status = status;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
                return status;
            }

            // Unlock the code needed for performance.
            MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
            DIOGlobals.IsLocked = FALSE;

            // Process any outstanding queued IRPs.
            DIOProcessQueuedRequests(deviceInfo);
            status = STATUS_SUCCESS;
            break;
        }  // case IRP_MN_START_DEVICE:

        // Handle PnP query stop device request.
        case IRP_MN_QUERY_STOP_DEVICE:
        {

            // Can't stop if device is opened.
            if (deviceInfo->IsOpened > 0)
            {
                pIRP->IoStatus.Status = STATUS_DEVICE_BUSY;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                return STATUS_DEVICE_BUSY;
            }

            // Don't allow requests.
            deviceInfo->DevicePnPState = StopPending;
            deviceInfo->HoldNewRequests = TRUE;

            // Wait for the existing ones to be finished.
            DIOIoDecrement(deviceInfo);
            KeWaitForSingleObject(&deviceInfo->StopEvent, Executive,
                KernelMode, FALSE, NULL);

            // Finish request.
            IoSkipCurrentIrpStackLocation(pIRP);
            status = IoCallDriver(deviceInfo->LowerDeviceObject, pIRP);
            return status;
        }  // case IRP_MN_QUERY_STOP_DEVICE:
    
        // Handle PnP cancel stop device request.
        case IRP_MN_CANCEL_STOP_DEVICE:
        {

            // **** Local Declarations ****

            // Synchronization event.
            KEVENT event;

            // **** Code Statements ****

            // Send the IRP to the next lower driver in chain and wait
            // for completion (via DIOSyncCompletion call-back).
            KeInitializeEvent(&event, SynchronizationEvent, FALSE);
            IoCopyCurrentIrpStackLocationToNext(pIRP);
            IoSetCompletionRoutine(pIRP, DIOSyncCompletion, &event,
                TRUE, TRUE, TRUE);
            status = IoCallDriver(deviceInfo->LowerDeviceObject, pIRP);
            if (status == STATUS_PENDING)
            {
                KeWaitForSingleObject(&event, Executive, KernelMode, FALSE,
                    NULL);
                status = pIRP->IoStatus.Status;
            }
            else if (!NT_SUCCESS(status))
            {
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = status;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                return status;
            }

            // Restart the queue and set the state to started.
            deviceInfo->HoldNewRequests = FALSE;
            deviceInfo->DevicePnPState = Started;

            // Process any queued requests.
            DIOProcessQueuedRequests(deviceInfo);
            break;
        }  // case IRP_MN_CANCEL_STOP_DEVICE:

        // Handle PnP stop device request.
        case IRP_MN_STOP_DEVICE:
        {

            // Disable the device interface.
            IoSetDeviceInterfaceState(&deviceInfo->DeviceClassSymbolicName,
                FALSE);

            // Disconnect to the interrupt service routine.
            if (deviceInfo->Interrupt != NULL)
            {
                IoDisconnectInterrupt(deviceInfo->Interrupt);
                deviceInfo->Interrupt = NULL;
            }

            // Unmap the memory for the ports into virtual memory.
            if (deviceInfo->Mapped)
            {
                MmUnmapIoSpace(deviceInfo->VirtualAddress,
                   deviceInfo->DeviceBusAddressLength);
                deviceInfo->Mapped = FALSE;
            }

            // Mark the device as stopped. 
            deviceInfo->DevicePnPState = Stopped;

            // Pass to lower drivers.
            IoSkipCurrentIrpStackLocation(pIRP);
            status = IoCallDriver(deviceInfo->LowerDeviceObject, pIRP);
            DIOIoDecrement(deviceInfo);
            return status;
        }  // case IRP_MN_STOP_DEVICE:

        // On a query for a device removal, attempt to process any queued
        // IRPs and pass to lower driver(s).
        case IRP_MN_QUERY_REMOVE_DEVICE:
        {

            // Can't remove if device is opened.
            if (deviceInfo->IsOpened > 0)
            {
                pIRP->IoStatus.Status = STATUS_DEVICE_BUSY;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                return STATUS_DEVICE_BUSY;
            }

            // Process any queued requests
            DIOProcessQueuedRequests(deviceInfo);

            // Be sure to hold any new IRPs and set the device state.
            deviceInfo->HoldNewRequests = TRUE;
            deviceInfo->DevicePnPState = RemovePending;
            DIOIoDecrement(deviceInfo);

            // Wait for all IRPs to be complete.
            KeWaitForSingleObject(&deviceInfo->StopEvent, Executive,
                KernelMode, FALSE, NULL);

            // Pass to lower drivers.
            IoSkipCurrentIrpStackLocation(pIRP);
            status = IoCallDriver(deviceInfo->LowerDeviceObject, pIRP);
            return status;
        }  // case IRP_MN_QUERY_REMOVE_DEVICE:
     
        // On a cancel remove device, resume normal activity.
        case IRP_MN_CANCEL_REMOVE_DEVICE:
        {

            // **** Local Declarations ****

            // Synchronization event.
            KEVENT event;

            // **** Code Statements ****

            // Send the IRP to the next lower driver in chain and wait
            // for completion (via DIOSyncCompletion call-back).
            KeInitializeEvent(&event, SynchronizationEvent, FALSE);
            IoCopyCurrentIrpStackLocationToNext(pIRP);
            IoSetCompletionRoutine(pIRP, DIOSyncCompletion, &event,
                TRUE, TRUE, TRUE);
            status = IoCallDriver(deviceInfo->LowerDeviceObject, pIRP);
            if (status == STATUS_PENDING)
            {
                KeWaitForSingleObject(&event, Executive, KernelMode, FALSE,
                    NULL);
                status = pIRP->IoStatus.Status;
            }
            else if (!NT_SUCCESS(status))
            {
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = status;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                return status;
            }

            // Set to normal activity.
            deviceInfo->HoldNewRequests = FALSE;
            deviceInfo->DevicePnPState = Started;
            DIOProcessQueuedRequests(deviceInfo);
            break;
        }  // case IRP_MN_CANCEL_REMOVE_DEVICE:

        // On a surprise removal, remove the device.
        case IRP_MN_SURPRISE_REMOVAL:
        {

            // Set to not hold requests and current state.
            deviceInfo->HoldNewRequests = FALSE;
            deviceInfo->DevicePnPState = SurpriseRemovePending;

            // Disable the device interface.
            IoSetDeviceInterfaceState(&deviceInfo->DeviceClassSymbolicName,
                FALSE);

            // Disconnect to the interrupt service routine.
            if (deviceInfo->Interrupt != NULL)
            {
                IoDisconnectInterrupt(deviceInfo->Interrupt);
                deviceInfo->Interrupt = NULL;
            }

            // Unmap the memory for the ports into virtual memory.
            if (deviceInfo->Mapped)
            {
                MmUnmapIoSpace(deviceInfo->VirtualAddress,
                   deviceInfo->DeviceBusAddressLength);
                deviceInfo->Mapped = FALSE;
            }

            // Send to next lower driver.
            IoSkipCurrentIrpStackLocation(pIRP);
            status = IoCallDriver(deviceInfo->LowerDeviceObject, pIRP);
            DIOIoDecrement(deviceInfo);
            return status;
        }  // case IRP_MN_SURPRISE_REMOVAL:

        // Handle a remove device request.
        case IRP_MN_REMOVE_DEVICE:
        {
            INTERRUPT_EVENT_INFO *savedInterruptEventInfo;

            // If the device state is not in a suprise remove pending and
            // not in a stopped state, must first deallocate resources.
            if ((deviceInfo->DevicePnPState != SurpriseRemovePending) &&
                (deviceInfo->DevicePnPState != Stopped))
            {

                // Set to not hold requests.
                deviceInfo->HoldNewRequests = FALSE;

                // Disable the device interface.
                IoSetDeviceInterfaceState(&deviceInfo->DeviceClassSymbolicName,
                    FALSE);

                // Disconnect to the interrupt service routine.
                if (deviceInfo->Interrupt != NULL)
                {
                    IoDisconnectInterrupt(deviceInfo->Interrupt);
                    deviceInfo->Interrupt = NULL;
                }

                // Unmap the memory for the ports into virtual memory.
                if (deviceInfo->Mapped)
                {
                    MmUnmapIoSpace(deviceInfo->VirtualAddress,
                       deviceInfo->DeviceBusAddressLength);
                    deviceInfo->Mapped = FALSE;
                }
            }

            // Set the status to deleted.
            deviceInfo->DevicePnPState = Deleted;

            // Process queued IRPs (will complete each with a status
            // of STATUS_DELETE_PENDING).
            DIOProcessQueuedRequests(deviceInfo);

            // Send the remove to the next lower device.
            IoSkipCurrentIrpStackLocation(pIRP);
            status = IoCallDriver(deviceInfo->LowerDeviceObject, pIRP);

            // Must decrement twice...once to remove this IRP and
            // another to remove the 1-based value.
            if (DIOIoDecrement(deviceInfo) > 0)
                DIOIoDecrement(deviceInfo);

            // Wait for the remove event to signal.
            KeWaitForSingleObject(&deviceInfo->RemoveEvent, Executive,
                KernelMode, FALSE, NULL);

            // Free the registered symbolic name (created in AddDevice by
            // IoRegisterDeviceInterface where it is our responsibility
            // to free this object).
            RtlFreeUnicodeString(&deviceInfo->DeviceClassSymbolicName);

            // Remove the device from the device stack.
            IoDetachDevice(deviceInfo->LowerDeviceObject);

            // Deallocate the memory for the interrupt queue.
            ExFreePool(deviceInfo->InterruptQueue);

            // Deallocate the memory for the PPI port value array.
            ExFreePool(deviceInfo->PPIPortValue);

            // Ensure all interrupt event info in the list has been removed.
            // Event objects must be dereferenced.
            while (deviceInfo->InterruptEventInfoList != NULL)
            {
                savedInterruptEventInfo = deviceInfo->InterruptEventInfoList;
                deviceInfo->InterruptEventInfoList =
                    deviceInfo->InterruptEventInfoList->NextInterruptEventInfo;
                ObDereferenceObject(
                    savedInterruptEventInfo->InterruptEventObject);
                ExFreePool(savedInterruptEventInfo);
            }
            deviceInfo->InterruptEventInfoCount = 0;

            // Delete the device.
            IoDeleteDevice(deviceInfo->DeviceObject);

            // If the lockable pages were locked, unlock.
            if (DIOGlobals.IsLocked)
            {
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
            }
            return status;
        }  // case IRP_MN_REMOVE_DEVICE:

        // On a capability query, return the lower driver capabilities.
        case IRP_MN_QUERY_CAPABILITIES:
        {

            // **** Local Declarations ****

            // Synchronization event.
            KEVENT event;

            // Structure containing queried device capabilities.
            DEVICE_CAPABILITIES *deviceCapabilities;

            // **** Code Statements ****

            // Get the device capabilities structure.
            deviceCapabilities =
                pIRPStack->Parameters.DeviceCapabilities.Capabilities;

            // Can change or set anything desired.  Nothing in this case
            // so simply pass the IRP down to the next lower driver in
            // chain and wait for completion (via DIOSyncCompletion
            // call-back).
            KeInitializeEvent(&event, SynchronizationEvent, FALSE);
            IoCopyCurrentIrpStackLocationToNext(pIRP);
            IoSetCompletionRoutine(pIRP, DIOSyncCompletion, &event,
                TRUE, TRUE, TRUE);
            status = IoCallDriver(deviceInfo->LowerDeviceObject, pIRP);
            if (status == STATUS_PENDING)
            {
                KeWaitForSingleObject(&event, Executive, KernelMode, FALSE,
                    NULL);
                status = pIRP->IoStatus.Status;
            }
            else if (!NT_SUCCESS(status))
            {
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = status;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                return status;
            }

            // Can do more stuff with capabilities if needed.  Not needed
            // so finished.
            break;
        } // case IRP_MN_QUERY_CAPABILITIES:

        // On a query device state, return the device state.
        case IRP_MN_QUERY_PNP_DEVICE_STATE:
        {

            // **** Local Declarations ****

            // Synchronization event.
            KEVENT event;

            // Bitmask containing device state.
            PNP_DEVICE_STATE *deviceState;

            // **** Code Statements ****

            // Pass the IRP down to the next lower driver in
            // chain and wait for completion (via DIOSyncCompletion
            // call-back).
            KeInitializeEvent(&event, SynchronizationEvent, FALSE);
            IoCopyCurrentIrpStackLocationToNext(pIRP);
            IoSetCompletionRoutine(pIRP, DIOSyncCompletion, &event,
                TRUE, TRUE, TRUE);
            status = IoCallDriver(deviceInfo->LowerDeviceObject, pIRP);
            if (status == STATUS_PENDING)
            {
                KeWaitForSingleObject(&event, Executive, KernelMode, FALSE,
                    NULL);
                status = pIRP->IoStatus.Status;
            }
            else if (!NT_SUCCESS(status))
            {
                deviceInfo->HoldNewRequests = FALSE;
                pIRP->IoStatus.Status = status;
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                return status;
            }

            // A non zero DeviceState means that the device was told to be
            // invalidated.  Reset it back to zero.
            if (deviceInfo->CurrentDeviceState != 0)
            {
                deviceState =
                    (PNP_DEVICE_STATE *) &(pIRP->IoStatus.Information);
                if (deviceState)
                    *deviceState |= deviceInfo->CurrentDeviceState;
                deviceInfo->CurrentDeviceState = 0;
                status = STATUS_SUCCESS;
            }
            break;
        }  // case IRP_MN_QUERY_PNP_DEVICE_STATE:

        // Pass down all unhandled requests.
        default:
        {

            IoSkipCurrentIrpStackLocation(pIRP);
            status = IoCallDriver (deviceInfo->LowerDeviceObject, pIRP);
            DIOIoDecrement(deviceInfo);
            return status;
        }  // default:
    }  // switch (pIRPStack->MinorFunction)

    // Complete the IRP.
    pIRP->IoStatus.Status = status;
    IoCompleteRequest(pIRP, IO_NO_INCREMENT);
    DIOIoDecrement(deviceInfo);
    return status;
}

////////////////////////////////////////////////////////////////////////////////
// Dispatch power requests.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOPowerDispatch(DEVICE_OBJECT *deviceObject, IRP *pIRP)
{

    // **** Local Declarations ****

    // Status of kernel calls.
    NTSTATUS status;

    // Pointer to local data structure (driver specific settings,
    // data, etc) stored with the device object.
    DIO_DEVICE_EXTENSION *deviceInfo;

    // Pointer to caller's stack location in the given IRP.  Required
    // to get parameters for the current request.
    IO_STACK_LOCATION *pIRPStack;

    // **** Code Statements ****

    // Get the local data structure for the PCI-DIO24 device.
    deviceInfo = (DIO_DEVICE_EXTENSION *) deviceObject->DeviceExtension;

    // Get the IRP stack (user parameters).
    pIRPStack = IoGetCurrentIrpStackLocation(pIRP);

    // If the device is in a deleted state, fail the IRP with a
    // STATUS_DELETE_PENDING.
    if (deviceInfo->DevicePnPState == Deleted)
    {
    
        // Although failing the IRP, notify the power manager to
        // send next IRP.
        PoStartNextPowerIrp(pIRP);
        pIRP->IoStatus.Status = STATUS_DELETE_PENDING;
        IoCompleteRequest(pIRP, IO_NO_INCREMENT);
        return STATUS_DELETE_PENDING;
    } 

    // Increment the IRP count (power IRPs are not queued internally,
    // but other IRPs may be).
    DIOIoIncrement(deviceInfo);

    // Switch on the MinorFunction for the request.
    switch (pIRPStack->MinorFunction)
    {

        // Request to set the power. There are two cases when this
        // IRP is generated.  First is when there is a change in the
        // system power state and second to change the power state of
        // a device during the power manager's idle detection.
        case IRP_MN_SET_POWER:
        {

            // **** Local Declarations ****

            // Requested power type.
            POWER_STATE_TYPE powerType;

            // Requested power state.
            POWER_STATE powerState;

            // **** Code Statements ****

            // Get the requested power type and state.
            powerType = pIRPStack->Parameters.Power.Type;
            powerState = pIRPStack->Parameters.Power.State;

            // Switch of the type of power change (system or device).
            switch (powerType)
            {

                // System power change.
                case SystemPowerState:
                {

                    // If the system is powering down, all incoming I/O
                    // requests should be queued.
                    if (powerState.SystemState > PowerSystemWorking)
                    {

                        // Queue future requests.
                        deviceInfo->HoldNewRequests = TRUE;

                        // Wait for the existing IRPs to be complete.
                        // First decrement this IRP from the list and
                        // wait until the stop event is set.
                        DIOIoDecrement(deviceInfo);
                        KeWaitForSingleObject(&deviceInfo->StopEvent,
                            Executive, KernelMode, FALSE, NULL);

                        // Increment this back to send down to the next
                        // lower driver.
                        DIOIoIncrement(deviceInfo);

                        // Mark this IRP as pending for further processing.
                        IoMarkIrpPending(pIRP);

                        // Set this device power state to D3 (a powered off
                        // state).
                        powerState.DeviceState = PowerDeviceD3;

                        // Send a power IRP to ourselfs to power off.
                        status = PoRequestPowerIrp(deviceInfo->DeviceObject,
                            IRP_MN_SET_POWER, powerState,
                            DIOPoRequestComplete, pIRP, NULL);
                        return status;
                    }

                    // Powering up.  Must be handled first by lower
                    // drivers and then up the stack.
                    else
                    {
                        IoCopyCurrentIrpStackLocationToNext(pIRP);
                        IoSetCompletionRoutine(pIRP,
                            DIOPowerCompletionRoutine, NULL, TRUE, TRUE, TRUE);
                        return PoCallDriver(deviceInfo->LowerDeviceObject, pIRP);
                    }
                }  // case SystemPowerState:

                // Device power change.
                case DevicePowerState:
                {

                    // If the device is powering down, all incoming I/O
                    // requests should be queued.
                    if (powerState.DeviceState > PowerDeviceD0)
                    {

                        // Set the power state.
                        PoSetPowerState(deviceObject, powerType, powerState);
                        deviceInfo->DevicePowerState = powerState.DeviceState;

                        // Send to next lower driver and ready for next power
                        // IRP.
                        IoCopyCurrentIrpStackLocationToNext(pIRP);
                        PoStartNextPowerIrp(pIRP);
                        status = PoCallDriver(deviceInfo->LowerDeviceObject,
                            pIRP);
                        DIOIoDecrement(deviceInfo);
                        return status;
                    }

                    // If the device is powering up, must be handled first
                    // by lower drivers and then up the stack.
                    else
                    {
                        IoCopyCurrentIrpStackLocationToNext(pIRP);
                        IoSetCompletionRoutine(pIRP,
                            DIOPowerCompletionRoutine, NULL, TRUE, TRUE, TRUE);
                        return PoCallDriver(deviceInfo->LowerDeviceObject, pIRP);
                    }
                }  // case DevicePowerState:

                // Nothing else implemented.
                default:
                {
                    status = STATUS_NOT_IMPLEMENTED;
                }  // default:
            }  // switch (powerType)

            // Should always be a non-successful status value (from default:)
            if (!NT_SUCCESS(status))
            {
                pIRP->IoStatus.Information = 0;
                pIRP->IoStatus.Status = status;
                PoStartNextPowerIrp(pIRP);
                IoCompleteRequest(pIRP, IO_NO_INCREMENT);
                DIOIoDecrement(deviceInfo);
                return status;
            }
        }  // case IRP_MN_SET_POWER:

        // For all other power requests, pass it down.
        default:
        {

            IoCopyCurrentIrpStackLocationToNext(pIRP);
            PoStartNextPowerIrp(pIRP);
            status = PoCallDriver(deviceInfo->LowerDeviceObject, pIRP);
            DIOIoDecrement(deviceInfo);
            return status;
        }  // default:
    }  // switch (pIRPStack->MinorFunction)

    // All previously returned.
}

////////////////////////////////////////////////////////////////////////////////
// Create or close a device handle (when the instance is created or
// closed by a user).
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOCreateClose(DEVICE_OBJECT *deviceObject, IRP *pIRP)
{

    // **** Local Declarations ****

    // Status of kernel calls.
    NTSTATUS status;

    // Pointer to local data structure (driver specific settings,
    // data, etc) stored with the device object.
    DIO_DEVICE_EXTENSION *deviceInfo;

    // Pointer to caller's stack location in the given IRP.  Required
    // to get parameters for the current request.
    IO_STACK_LOCATION *pIRPStack;

    // Pointer to event information structure for removal.
    INTERRUPT_EVENT_INFO *savedInterruptEventInfo;

    // **** Code Statements ****

    // Get the device information and the driver object.
    deviceInfo = deviceObject->DeviceExtension;

    // Queue the request if the requests are being queued.
    if (deviceInfo->HoldNewRequests)
    {
        status = DIOQueueRequest(deviceInfo, pIRP);
        if (NT_SUCCESS(status))
            status = STATUS_PENDING;
        return status;
    }

    // Increment the number of requests in progress.
    DIOIoIncrement(deviceInfo);

    // Get the IRP stack (user parameters).
    pIRPStack = IoGetCurrentIrpStackLocation(pIRP);

    // Set the opened or closed flag as appropriate.
    switch (pIRPStack->MajorFunction)
    {

        // Open (create) a file handle request.
        case IRP_MJ_CREATE:
        {

            // Make sure a request to create a directory hasn't been sent!
            if (IoGetCurrentIrpStackLocation(pIRP)->Parameters.Create.Options &
                FILE_DIRECTORY_FILE)
            {
                LogDIOError(deviceInfo->DriverObject, deviceObject,
                    PhysicalZero, PhysicalZero, 0, 0, 0, 0x40,
                    STATUS_NOT_A_DIRECTORY, DIO_NOT_A_DIRECTORY,
                    0, NULL, 0, NULL);
                status = STATUS_NOT_A_DIRECTORY;
            }
            else
            {
                deviceInfo->IsOpened++;
                status = STATUS_SUCCESS;

                if (deviceInfo->IsOpened == 1)
                {

                    // Ensure interrupts are disabled.
                    deviceInfo->InterruptHandled = FALSE;
                    if (deviceInfo->PortMemoryType == 1)
                        WRITE_PORT_UCHAR((UCHAR *)
                            deviceInfo->DisableIRQAddress, 0x00);
                    else
                        WRITE_REGISTER_UCHAR((UCHAR *)
                            deviceInfo->DisableIRQAddress, 0x00);

                    // Ensure all interrupt event info in the list has been
                    // removed.  Event objects must be dereferenced.
                    while (deviceInfo->InterruptEventInfoList != NULL)
                    {
                        savedInterruptEventInfo =
                            deviceInfo->InterruptEventInfoList;
                        deviceInfo->InterruptEventInfoList =
                            deviceInfo->InterruptEventInfoList->
                                NextInterruptEventInfo;
                        ObDereferenceObject(
                            savedInterruptEventInfo->InterruptEventObject);
                        ExFreePool(savedInterruptEventInfo);
                    }
                    deviceInfo->InterruptEventInfoCount = 0;
                }

                // If the lockable code section is not locked, lock.
                if (!DIOGlobals.IsLocked)
                {
                    MmLockPagableSectionByHandle(DIOGlobals.PAGEDIO_Handle);
                    DIOGlobals.IsLocked = TRUE;
                }
            }
            break;
        }  // case IRP_MJ_CREATE:

        // Close a file handle request.
        case IRP_MJ_CLOSE:
        {
            deviceInfo->IsOpened--;

            // If there are no more opened handles, ensure interrupts
            // are disabled.
            if (deviceInfo->IsOpened == 0)
            {
                deviceInfo->InterruptHandled = FALSE;
                if (deviceInfo->PortMemoryType == 1)
                    WRITE_PORT_UCHAR((UCHAR *)
                        deviceInfo->DisableIRQAddress, 0x00);
                else
                    WRITE_REGISTER_UCHAR((UCHAR *)
                        deviceInfo->DisableIRQAddress, 0x00);

                // Ensure all interrupt event info in the list has been
                // removed.  Event objects must be dereferenced.
                while (deviceInfo->InterruptEventInfoList != NULL)
                {
                    savedInterruptEventInfo =
                        deviceInfo->InterruptEventInfoList;
                    deviceInfo->InterruptEventInfoList =
                        deviceInfo->InterruptEventInfoList->
                            NextInterruptEventInfo;
                    ObDereferenceObject(
                        savedInterruptEventInfo->InterruptEventObject);
                    ExFreePool(savedInterruptEventInfo);
                }
                deviceInfo->InterruptEventInfoCount = 0;

                // Unlock the lockable code section.
                MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
                DIOGlobals.IsLocked = FALSE;
            }
            status = STATUS_SUCCESS;
            break;
        }  // case IRP_MJ_CLOSE:

        // Should never get here!
        default:
        {
            LogDIOError(deviceInfo->DriverObject, deviceObject, PhysicalZero,
                PhysicalZero, 0, 0, 0, 0x41, STATUS_NOT_IMPLEMENTED,
                DIO_INVALID_FUNCTION, 0, NULL, 0, NULL);
            status = STATUS_NOT_IMPLEMENTED;
            break;
        }  // default:
    }  // switch (pIRPStack->MajorFunction)

    pIRP->IoStatus.Information = 0;
    pIRP->IoStatus.Status = status;
    IoCompleteRequest(pIRP, IO_NO_INCREMENT);
    DIOIoDecrement(deviceInfo);
    return status;
}

////////////////////////////////////////////////////////////////////////////////
// Cleanup the device.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOCleanup(DEVICE_OBJECT *deviceObject, IRP *pIRP)
{

    // **** Local Declarations ****

    // Status of kernel calls.
    NTSTATUS status;

    // Pointer to local data structure (driver specific settings,
    // data, etc) stored with the device object.
    DIO_DEVICE_EXTENSION *deviceInfo;

    // Pointer to caller's stack location in the given IRP.  Required
    // to get parameters for the current request.
    IO_STACK_LOCATION *pIRPStack;

    // Queued list of IRPs to cancel.
    LIST_ENTRY cleanupList;

    // Current IRQ Level.
    KIRQL oldIrql;

    // Pointers to IRQs in list.
    LIST_ENTRY *thisEntry;
    LIST_ENTRY *nextEntry;
    LIST_ENTRY *listHead;

    // Pointer to pending IRQ.
    IRP *pendingIRP;

    // Pointer to stack location of pending IRQ.
    IO_STACK_LOCATION *pendingIRPStack;

    // **** Code Statements ****

    // Get the device information and the driver object.
    deviceInfo = deviceObject->DeviceExtension;

    // Increment the IRP count.
    DIOIoIncrement(deviceInfo);

    // Get the IRP stack (user parameters).
    pIRPStack = IoGetCurrentIrpStackLocation(pIRP);

    // Initialize a list of IRPs to clean up.
    InitializeListHead(&cleanupList);

    // Acquire a cancel spinlock to manipulate the list.
    IoAcquireCancelSpinLock(&oldIrql);
    KeAcquireSpinLockAtDpcLevel(&deviceInfo->QueueLock);

    // Walk through the list of queued IRPs and remove
    // those that belong to the same opened handle (i.e.
    // FileObject) as the current cleanup IRP.
    listHead = &deviceInfo->NewRequestsQueue;
    for (thisEntry = listHead->Flink, nextEntry = thisEntry->Flink;
        thisEntry != listHead;
        thisEntry = nextEntry, nextEntry = thisEntry->Flink)
    {
        pendingIRP = CONTAINING_RECORD(thisEntry, IRP, Tail.Overlay.ListEntry);
        pendingIRPStack = IoGetCurrentIrpStackLocation(pendingIRP);
        if (pIRPStack->FileObject == pendingIRPStack->FileObject)
        {
            RemoveEntryList(thisEntry);
            InsertTailList(&cleanupList, thisEntry);
        }
    }

    // Release the spinlocks.
    KeReleaseSpinLockFromDpcLevel(&deviceInfo->QueueLock);
    IoReleaseCancelSpinLock(oldIrql);

    // Go through the cleanup list and cancel all the IRPs.
    while (!IsListEmpty(&cleanupList))
    {
        thisEntry = RemoveHeadList(&cleanupList);
        pendingIRP = CONTAINING_RECORD(thisEntry, IRP, Tail.Overlay.ListEntry);
        IoSetCancelRoutine(pendingIRP, NULL);
        pendingIRP->IoStatus.Information = 0;
        pendingIRP->IoStatus.Status = STATUS_CANCELLED;
        IoCompleteRequest(pendingIRP, IO_NO_INCREMENT);
    }

    DIOIoDecrement(deviceInfo);
    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Unload the device driver from the system.  Free all resources allocated in
// DriverEntry.
////////////////////////////////////////////////////////////////////////////////
void DIOUnload(DRIVER_OBJECT *driverObject)
{

    // **** Code Statements ****

    // The only thing allocated in the DriverEntry was the registry path.
    ExFreePool(DIOGlobals.RegistryPath.Buffer);

    // Make sure the lockable page is not locked.
    if (DIOGlobals.IsLocked)
    {
        MmUnlockPagableImageSection(DIOGlobals.PAGEDIO_Handle);
        DIOGlobals.IsLocked = FALSE;
    }

    return;
}

////////////////////////////////////////////////////////////////////////////////
// Dispatch handler for the PCI-DIO24 driver.  Handles I/O Request Packets
// (IRP) from the I/O Manager.  Specifically an application call to
// DeviceIoControl is routed to the dispatch handler which performs the
// appropriate action as given in the IO Control Code.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIODispatch(DEVICE_OBJECT *deviceObject, IRP *pIRP)
{

    // **** Local Declarations ****

    // Status to return from NT calls.
    NTSTATUS status;

    // Pointer to local data structure (driver specific settings,
    // data, etc) stored with the device object.
    DIO_DEVICE_EXTENSION *deviceInfo;

    // Pointer to caller's stack location in the given IRP.  Required
    // to get parameters for the current request.
    IO_STACK_LOCATION *pIRPStack;

    // **** Code Statements ****

    // Get the local data structure for the PCI-DIO24 device.
    deviceInfo = (DIO_DEVICE_EXTENSION *) deviceObject->DeviceExtension;

    // If holding requests, queue this IRP.
    if (deviceInfo->HoldNewRequests)
    {
        status = DIOQueueRequest(deviceInfo, pIRP);
        if (NT_SUCCESS(status))
            status = STATUS_PENDING;
        return status;
    }

    // Increment the processing IRP.
    DIOIoIncrement(deviceInfo);

    // Get the IRP stack (user parameters).
    pIRPStack = IoGetCurrentIrpStackLocation(pIRP);

    // Before the switch statement, setup the return values in case
    // the driver is called with an illegal or unimplemented
    // function or IO Control code.  Set the 'Information' field to
    // zero (no bytes transferred) and return status value.  This
    // enables the switch statement to not have 'default' clause.
    pIRP->IoStatus.Information = 0;
    status = STATUS_INVALID_PARAMETER;

    // This is only called with a function code of IRP_MJ_DEVICE_CONTROL
    // so no checking is performed.  Perform a switch on the IO Control
    // code.
    switch (pIRPStack->Parameters.DeviceIoControl.IoControlCode)
    {

        // Set the I/O mode of the given port of the PPI.
        case IOCTL_DIO_SETMODE:
            status = DIOSetMode(deviceInfo, pIRP, pIRPStack);
            break;

        // Write byte to the given port of the PPI.
        case IOCTL_DIO_WRITE:
            status = DIOWrite(deviceInfo, pIRP, pIRPStack);
            break;

        // Set specified bit(s) to the given port of the PPI.
        case IOCTL_DIO_SET:
            status = DIOSet(deviceInfo, pIRP, pIRPStack);
            break;

        // Reset specified bit(s) to the given port of the PPI.
        case IOCTL_DIO_RESET:
            status = DIOReset(deviceInfo, pIRP, pIRPStack);
            break;

        // Read byte from the given port of the PPI.
        case IOCTL_DIO_READ:
            status = DIORead(deviceInfo, pIRP, pIRPStack);
            break;

        // Initialize the given timer counter.
        case IOCTL_DIO_INIT_COUNTER:
            status = DIOInitCounter(deviceInfo, pIRP, pIRPStack);
            break;

        // Set new values in given timer counter.
        case IOCTL_DIO_WRITE_COUNTER:
            status = DIOWriteCounter(deviceInfo, pIRP, pIRPStack);
            break;

        // Input counter value from the given timer counter.
        case IOCTL_DIO_READ_COUNTER:
            status = DIOReadCounter(deviceInfo, pIRP, pIRPStack);
            break;

        // Input all counter values from the given timer.
        case IOCTL_DIO_READ_TIMER:
            status = DIOReadTimer(deviceInfo, pIRP, pIRPStack);
            break;

        // Set the timer interrupt period.
        case IOCTL_DIO_SET_TIMER_PERIOD:
            status = DIOSetTimerInterruptPeriod(deviceInfo, pIRP, pIRPStack);
            break;

        // Add the event handle for timer/external interrupt handler to set.
        case IOCTL_DIO_ADD_INTERRUPT_EVENT:
            status = DIOAddInterruptEvent(deviceInfo, pIRP, pIRPStack);
            break;

        // Remove the event handle for timer/external interrupt handler to set.
        case IOCTL_DIO_REMOVE_INTERRUPT_EVENT:
            status = DIORemoveInterruptEvent(deviceInfo, pIRP, pIRPStack);
            break;

        // Turn on interrupt handler and enable interrupts.
        case IOCTL_DIO_ENABLE_INTERRUPT:
            status = DIOEnableInterrupt(deviceInfo, pIRP, pIRPStack);
            break;

        // Turn off interrupt handler and disable interrupts.
        case IOCTL_DIO_DISABLE_INTERRUPT:
            status = DIODisableInterrupt(deviceInfo, pIRP, pIRPStack);
            break;

        // Get event information.
        case IOCTL_DIO_GET_EVENT_INFO:
            status = DIOGetEventInfo(deviceInfo, pIRP, pIRPStack);
            break;

        // Reset event after completion.
        case IOCTL_DIO_RESET_EVENT:
            status = DIOResetEvent(deviceInfo, pIRP, pIRPStack);
            break;

        // Reset current event tick counters in interrupt events.
        case IOCTL_DIO_RESET_INTERRUPT_EVENTS:
            status = DIOResetInterruptEvents(deviceInfo, pIRP, pIRPStack);
            break;
    }

    // Record the status to the IRP for return.
    pIRP->IoStatus.Status = status;

    // The IRP processing is complete, release the IRP back to the
    // calling thread.  Because the device driver takes very little
    // time, no priority boosting is required.
    IoCompleteRequest(pIRP, IO_NO_INCREMENT);

    // Decrement the processing IRPs.
    DIOIoDecrement(deviceInfo);
    return status;
}

////////////////////////////////////////////////////////////////////////////////
// Set the I/O mode of a PPI Port.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOSetMode(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                    IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Structure containing PPI data from application.
    PPI_IOCTL_INPUT *pPPIInput;

    // Bits to set/reset the appropriate port's mode.
    UCHAR portSetting;

    // **** Code Statements ****

    // No data will be returned.
    pIRP->IoStatus.Information = 0;

    // The input size the IRP is providing MUST be the size of the
    // PPI_IOCTL_INPUT structure.
    if (pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
        sizeof(PPI_IOCTL_INPUT))
        return STATUS_INVALID_PARAMETER;

    // Since there is no data to output, verify the output size the
    // IRP is expected to return is zero.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength != 0)
        return STATUS_INVALID_PARAMETER;

    // The device IOCTL is set as METHOD_BUFFERED.  NT copies the user's
    // input buffer to the IRP's SystemBuffer entry before entry and
    // copies the IRP's SystemBuffer entry to the user's output buffer
    // after return.  The output buffer is not used as there
    // is no output sent to the user.  Point to the SystemBuffer as
    // a PPI_IOCTL_INPUT structure.
    pPPIInput = (PPI_IOCTL_INPUT *) pIRP->AssociatedIrp.SystemBuffer;

    // Check to be sure the PPI port is valid (A, B or C).
    if (pPPIInput->Port > PORT_C)
        return STATUS_INVALID_PARAMETER;

    // Check to be sure the mode is valid (input or output).
    if ((pPPIInput->Mode != PPI_PORT_INPUT) &&
        (pPPIInput->Mode != PPI_PORT_OUTPUT))
        return STATUS_INVALID_PARAMETER;

    // Switch on the port to get the appropriate bits to set or reset.
    switch(pPPIInput->Port)
    {
        case PORT_A:
            portSetting = 0x10;
            break;
        case PORT_B:
            portSetting = 0x02;
            break;
        case PORT_C:
            portSetting = 0x09;
            break;
    };

    // Set or reset the appropriate bits in the current port
    // settings (control port entry in PPIPortValue array).  Note: if the
    // mode is changing to input, clear the PPIPortValue array
    // value for the specified port.
    if (pPPIInput->Mode == PPI_PORT_INPUT)
    {
        deviceInfo->PPIPortValue[3] |= portSetting;
        deviceInfo->PPIPortValue[(ULONG) pPPIInput->Port] = 0;
    }
    else
        deviceInfo->PPIPortValue[3] &= ~portSetting;

    ////////////////////////////////////////////////////////////////////////////
    // IMPORTANT NOTE!!
    //
    // The control port of an 8255 uses several bits for other
    // mode settings.  The PCI-DIO24 driver will ONLY utilize mode
    // zero and uses the PPI's Port C as a single 8-bit port (instead
    // of two 4-bit C-High and C-Low ports).  The driver will use
    // bits 5-7 of the control port entry in the PPIPortValue array
    // to identify if a port has been initialized.  These bits
    // are never sent to the control port but are for informational
    // use by the I/O routines to verify the port has been
    // initialized at least once prior to input or output.  If other
    // modes of the 8255 are required, this methodology of determining
    // initialization must also be changed as these bits will then be
    // required for use by the 8255.  Also note that for mode setting,
    // the high order bit MUST be set (bit 7 of the control register
    // commands the 8255 to set the mode, if bit 7 is zero is controls
    // specific port C modes outside the context of this driver).
    ////////////////////////////////////////////////////////////////////////////

    // Set the respective bit in the control port entry in the PPIPortValue
    // array to signify the respective port of the PPI has been initialized.
    // Bit 5:PORT_A, Bit 6:PORT_B, Bit 7:PORT_C.
    switch(pPPIInput->Port)
    {
        case PORT_A:
            deviceInfo->PPIPortValue[3] |= 0x20;
            break;
        case PORT_B:
            deviceInfo->PPIPortValue[3] |= 0x40;
            break;
        case PORT_C:
            deviceInfo->PPIPortValue[3] |= 0x80;
            break;
    };
    
    // Get the control port setting to send to the port.  The setting
    // is the value of control port entry in the PortValue array with
    // bits 5-7 stripped and then bit 7 forced to 1 to command a
    // mode set.
    portSetting = (deviceInfo->PPIPortValue[3] & 0x1f) | 0x80;

    // Send the value to the control port in I/O space.
    if (deviceInfo->PortMemoryType == 1)
        WRITE_PORT_UCHAR((UCHAR *)
            ((ULONG) deviceInfo->PPIBaseAddress + 3), portSetting);

    // Send the value to the control port in memory space.
    else
        WRITE_REGISTER_UCHAR((UCHAR *)
            ((ULONG) deviceInfo->PPIBaseAddress + 3), portSetting);

    // Successfully finished.
    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Write byte to the given port of the PPI.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOWrite(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                  IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Structure containing PPI data from application.
    PPI_IOCTL_INPUT *pPPIInput;

    // Bit to check for PPI port initialized.
    UCHAR initBit;

    // Bit(s) to check for PPI port current I/O mode.
    UCHAR modeBit;

    // **** Code Statements ****

    // No data will be returned.
    pIRP->IoStatus.Information = 0;

    // The input size the IRP is providing MUST be the size of the
    // PPI_IOCTL_INPUT structure.
    if (pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
        sizeof(PPI_IOCTL_INPUT))
        return STATUS_INVALID_PARAMETER;

    // Since there is no data to output, verify the output size the
    // IRP is expected to return is zero.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength != 0)
        return STATUS_INVALID_PARAMETER;

    // The device IOCTL is set as METHOD_BUFFERED.  NT copies the user's
    // input buffer to the IRP's SystemBuffer entry before entry and
    // copies the IRP's SystemBuffer entry to the user's output buffer
    // after return.  The output buffer is not used as there
    // is no output sent to the user.  Point to the SystemBuffer as
    // a PPI_IOCTL_INPUT structure.
    pPPIInput = (PPI_IOCTL_INPUT *) pIRP->AssociatedIrp.SystemBuffer;

    // Check to be sure the PPI port is valid (A, B or C).
    if (pPPIInput->Port > PORT_C)
        return STATUS_INVALID_PARAMETER;

    // Switch on the port to get the appropriate bits to test for
    // initialization and current I/O mode.
    switch(pPPIInput->Port)
    {
        case PORT_A:
            initBit = 0x20;
            modeBit = 0x10;
            break;
        case PORT_B:
            initBit = 0x40;
            modeBit = 0x02;
            break;
        case PORT_C:
            initBit = 0x80;
            modeBit = 0x09;
            break;
    }

    // Validate the PPI port has been initialized and the mode is
    // output.  If not, return error.
    if (((deviceInfo->PPIPortValue[3] & initBit) == 0) ||
        ((deviceInfo->PPIPortValue[3] & modeBit) != 0))
        return STATUS_ACCESS_VIOLATION;

    // Set the value in the output port entry in the PPIPortValue
    // array to the value to output to the port.
    deviceInfo->PPIPortValue[(ULONG) pPPIInput->Port] = pPPIInput->Byte;

    // Send the value to the output port in I/O space.
    if (deviceInfo->PortMemoryType == 1)
        WRITE_PORT_UCHAR((UCHAR *) ((ULONG) deviceInfo->PPIBaseAddress +
                                    (ULONG) pPPIInput->Port),
                         pPPIInput->Byte);

    // Send the value to the output port in memory space.
    else
        WRITE_REGISTER_UCHAR((UCHAR *) ((ULONG) deviceInfo->PPIBaseAddress +
                                        (ULONG) pPPIInput->Port),
                             pPPIInput->Byte);

    // Successfully finished.
    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Set specified bit(s) to the given port of the PPI.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOSet(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Structure containing PPI data from application.
    PPI_IOCTL_INPUT *pPPIInput;

    // Bit to check for PPI port initialized.
    UCHAR initBit;

    // Bit(s) to check for PPI port current I/O mode.
    UCHAR modeBit;

    // **** Code Statements ****

    // No data will be returned.
    pIRP->IoStatus.Information = 0;

    // The input size the IRP is providing MUST be the size of the
    // PPI_IOCTL_INPUT structure.
    if (pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
        sizeof(PPI_IOCTL_INPUT))
        return STATUS_INVALID_PARAMETER;

    // Since there is no data to output, verify the output size the
    // IRP is expected to return is zero.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength != 0)
        return STATUS_INVALID_PARAMETER;

    // The device IOCTL is set as METHOD_BUFFERED.  NT copies the user's
    // input buffer to the IRP's SystemBuffer entry before entry and
    // copies the IRP's SystemBuffer entry to the user's output buffer
    // after return.  The output buffer is not used as there
    // is no output sent to the user.  Point to the SystemBuffer as
    // a PPI_IOCTL_INPUT structure.
    pPPIInput = (PPI_IOCTL_INPUT *) pIRP->AssociatedIrp.SystemBuffer;

    // Check to be sure the PPI port is valid (A, B or C).
    if (pPPIInput->Port > PORT_C)
        return STATUS_INVALID_PARAMETER;

    // Switch on the port to get the appropriate bits to test for
    // initialization and current I/O mode.
    switch(pPPIInput->Port)
    {
        case PORT_A:
            initBit = 0x20;
            modeBit = 0x10;
            break;
        case PORT_B:
            initBit = 0x40;
            modeBit = 0x02;
            break;
        case PORT_C:
            initBit = 0x80;
            modeBit = 0x09;
            break;
    }

    // Validate the PPI port has been initialized and the mode is
    // output.  If not, return error.
    if (((deviceInfo->PPIPortValue[3] & initBit) == 0) ||
        ((deviceInfo->PPIPortValue[3] & modeBit) != 0))
        return STATUS_ACCESS_VIOLATION;

    // Set the value in the output port entry in the PPIPortValue
    // array to the value to output to the port.  This is the previously
    // output value or'ed with the given value (sets the given bits).
    deviceInfo->PPIPortValue[(ULONG) pPPIInput->Port] |= pPPIInput->Byte;

    // Send the value to the output port in I/O space.
    if (deviceInfo->PortMemoryType == 1)
        WRITE_PORT_UCHAR((UCHAR *) ((ULONG) deviceInfo->PPIBaseAddress +
                                    (ULONG) pPPIInput->Port),
                         deviceInfo->PPIPortValue[(ULONG) pPPIInput->Port]);

    // Send the value to the output port in memory space.
    else
        WRITE_REGISTER_UCHAR((UCHAR *) ((ULONG) deviceInfo->PPIBaseAddress +
                                        (ULONG) pPPIInput->Port),
                             deviceInfo->PPIPortValue[(ULONG) pPPIInput->Port]);

    // Successfully finished.
    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Reset specified bit(s) to the given port of the PPI.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOReset(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                  IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Structure containing PPI data from application.
    PPI_IOCTL_INPUT *pPPIInput;

    // Bit to check for PPI port initialized.
    UCHAR initBit;

    // Bit(s) to check for PPI port current I/O mode.
    UCHAR modeBit;

    // **** Code Statements ****

    // No data will be returned.
    pIRP->IoStatus.Information = 0;

    // The input size the IRP is providing MUST be the size of the
    // PPI_IOCTL_INPUT structure.
    if (pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
        sizeof(PPI_IOCTL_INPUT))
        return STATUS_INVALID_PARAMETER;

    // Since there is no data to output, verify the output size the
    // IRP is expected to return is zero.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength != 0)
        return STATUS_INVALID_PARAMETER;

    // The device IOCTL is set as METHOD_BUFFERED.  NT copies the user's
    // input buffer to the IRP's SystemBuffer entry before entry and
    // copies the IRP's SystemBuffer entry to the user's output buffer
    // after return.  The output buffer is not used as there
    // is no output sent to the user.  Point to the SystemBuffer as
    // a PPI_IOCTL_INPUT structure.
    pPPIInput = (PPI_IOCTL_INPUT *) pIRP->AssociatedIrp.SystemBuffer;

    // Check to be sure the PPI port is valid (A, B or C).
    if (pPPIInput->Port > PORT_C)
        return STATUS_INVALID_PARAMETER;

    // Switch on the port to get the appropriate bits to test for
    // initialization and current I/O mode.
    switch(pPPIInput->Port)
    {
        case PORT_A:
            initBit = 0x20;
            modeBit = 0x10;
            break;
        case PORT_B:
            initBit = 0x40;
            modeBit = 0x02;
            break;
        case PORT_C:
            initBit = 0x80;
            modeBit = 0x09;
            break;
    }

    // Validate the PPI port has been initialized and the mode is
    // output.  If not, return error.
    if (((deviceInfo->PPIPortValue[3] & initBit) == 0) ||
        ((deviceInfo->PPIPortValue[3] & modeBit) != 0))
        return STATUS_ACCESS_VIOLATION;

    // Set the value in the output port entry in the PPIPortValue
    // array to the value to output to the port.  This is the previously
    // output value and'ed with the inverse of the given value
    // (resets the given bits).
    deviceInfo->PPIPortValue[(ULONG) pPPIInput->Port] &= ~pPPIInput->Byte;

    // Send the value to the output port in I/O space.
    if (deviceInfo->PortMemoryType == 1)
        WRITE_PORT_UCHAR((UCHAR *) ((ULONG) deviceInfo->PPIBaseAddress +
                                    (ULONG) pPPIInput->Port),
                         deviceInfo->PPIPortValue[(ULONG) pPPIInput->Port]);

    // Send the value to the output port in memory space.
    else
        WRITE_REGISTER_UCHAR((UCHAR *) ((ULONG) deviceInfo->PPIBaseAddress +
                                        (ULONG) pPPIInput->Port),
                             deviceInfo->PPIPortValue[(ULONG) pPPIInput->Port]);

    // Successfully finished.
    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Read byte from the given port of the PPI.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIORead(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                 IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Structure containing PPI data from application.
    PPI_IOCTL_INPUT *pPPIInput;

    // Structure containing PPI data to application.
    PPI_IOCTL_OUTPUT *pPPIOutput;

    // Bit to check for PPI port initialized.
    UCHAR initBit;

    // **** Code Statements ****

    // No data yet to return.
    pIRP->IoStatus.Information = 0;

    // The input size the IRP is providing MUST be either the size of the
    // PPI_IOCTL_INPUT structure or the PPI_IOCTL_INPUT structure
    // minus the unused UCHAR parameter at the end of the structure.
    if ((pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
         sizeof(PPI_IOCTL_INPUT)) &&
        (pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
         (sizeof(PPI_IOCTL_INPUT) - sizeof(UCHAR))))
        return STATUS_INVALID_PARAMETER;

    // The data to return is a PPI_IOCTL_OUTPUT structure.  Verify the
    // output size the IRP is expected to return is the same size as the
    // structure.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength !=
        sizeof(PPI_IOCTL_OUTPUT))
        return STATUS_INVALID_PARAMETER;

    // The device IOCTL is set as METHOD_BUFFERED.  NT copies the user's
    // input buffer to the IRP's SystemBuffer entry before entry and
    // copies the IRP's SystemBuffer entry to the user's output buffer
    // after return.  Point to the SystemBuffer as both a
    // PPI_IOCTL_INPUT and a PPI_IOCTL_OUTPUT structure.
    pPPIInput = (PPI_IOCTL_INPUT *) pIRP->AssociatedIrp.SystemBuffer;
    pPPIOutput = (PPI_IOCTL_OUTPUT *) pIRP->AssociatedIrp.SystemBuffer;

    // Check to be sure the PPI port is valid (A, B or C).
    if (pPPIInput->Port > PORT_C)
        return STATUS_INVALID_PARAMETER;

    // Switch on the port to get the appropriate bits to test for
    // initialization and current I/O mode.
    switch(pPPIInput->Port)
    {
        case PORT_A:
            initBit = 0x20;
            break;
        case PORT_B:
            initBit = 0x40;
            break;
        case PORT_C:
            initBit = 0x80;
            break;
    }

    // Validate the PPI port has been initialized.  Note that the mode
    // can be input or output.  If not, return error.
    if ((deviceInfo->PPIPortValue[3] & initBit) == 0)
        return STATUS_ACCESS_VIOLATION;

    // Get the value from the input port in I/O space.
    if (deviceInfo->PortMemoryType == 1)
        pPPIOutput->Byte =
            READ_PORT_UCHAR((UCHAR *) ((ULONG) deviceInfo->PPIBaseAddress +
                                       (ULONG) pPPIInput->Port));

    // Get the value from the input port in memory space.
    else
        pPPIOutput->Byte =
            READ_REGISTER_UCHAR((UCHAR *) ((ULONG) deviceInfo->PPIBaseAddress +
                                           (ULONG) pPPIInput->Port));

    // Give the number of bytes returned.
    pIRP->IoStatus.Information = sizeof(PPI_IOCTL_OUTPUT);

    // Successfully finished.
    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Initialize the given timer counter.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOInitCounter(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                        IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Structure containing COUNTER data from application.
    COUNTER_IOCTL_INPUT *counterInput;

    // **** Code Statements ****

    // No data will be returned.
    pIRP->IoStatus.Information = 0;

    // If this device does not have a clock, return STATUS_NOT_IMPLEMENTED.
    if (!deviceInfo->HasClock)
        return STATUS_NOT_IMPLEMENTED;

    // The input size the IRP is providing MUST be the size of the
    // COUNTER_IOCTL_INPUT structure.
    if (pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
        sizeof(COUNTER_IOCTL_INPUT))
        return STATUS_INVALID_PARAMETER;

    // Since there is no data to output, verify the output size the
    // IRP is expected to return is zero.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength != 0)
        return STATUS_INVALID_PARAMETER;

    // The device IOCTL is set as METHOD_BUFFERED.  NT copies the user's
    // input buffer to the IRP's SystemBuffer entry before entry and
    // copies the IRP's SystemBuffer entry to the user's output buffer
    // after return.  The output buffer is not used as there
    // is no output sent to the user.  Point to the SystemBuffer as
    // a COUNTER_IOCTL_INPUT structure.
    counterInput = (COUNTER_IOCTL_INPUT *) pIRP->AssociatedIrp.SystemBuffer;

    // Check to be sure the given counter is valid (COUNTERA0 - COUNTERC2).
    if (counterInput->Counter > COUNTERC2)
        return STATUS_INVALID_PARAMETER;

    // Check to be sure the given mode is valid (0 - 5).
    if (counterInput->Mode > 5)
        return STATUS_INVALID_PARAMETER;

    // Call the DIOSetCounter function to set the counter's value and
    // initialize to the provided mode.
    DIOSetCounter(deviceInfo, counterInput, TRUE);

    // Successfully finished.
    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Set new values in given timer counter.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOWriteCounter(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                         IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Structure containing COUNTER data from application.
    COUNTER_IOCTL_INPUT *counterInput;

    // **** Code Statements ****

    // No data will be returned.
    pIRP->IoStatus.Information = 0;

    // If this device does not have a clock, return STATUS_NOT_IMPLEMENTED.
    if (!deviceInfo->HasClock)
        return STATUS_NOT_IMPLEMENTED;

    // The input size the IRP is providing MUST be either the size of the
    // COUNTER_IOCTL_INPUT structure or the COUNTER_IOCTL_INPUT structure
    // minus the unused UCHAR parameter at the end of the structure.
    if ((pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
         sizeof(COUNTER_IOCTL_INPUT)) &&
        (pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
         (sizeof(COUNTER_IOCTL_INPUT) - sizeof(UCHAR))))
        return STATUS_INVALID_PARAMETER;

    // Since there is no data to output, verify the output size the
    // IRP is expected to return is zero.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength != 0)
        return STATUS_INVALID_PARAMETER;

    // The device IOCTL is set as METHOD_BUFFERED.  NT copies the user's
    // input buffer to the IRP's SystemBuffer entry before entry and
    // copies the IRP's SystemBuffer entry to the user's output buffer
    // after return.  The output buffer is not used as there
    // is no output sent to the user.  Point to the SystemBuffer as
    // a COUNTER_IOCTL_INPUT structure.
    counterInput = (COUNTER_IOCTL_INPUT *) pIRP->AssociatedIrp.SystemBuffer;

    // Check to be sure the given counter is valid (COUNTERA0 - COUNTERC2).
    if (counterInput->Counter > COUNTERC2)
        return STATUS_INVALID_PARAMETER;

    // Call the DIOSetCounter function to set the counter's value without
    // changing the mode or stopping the counter.
    DIOSetCounter(deviceInfo, counterInput, FALSE);

    // Successfully finished.
    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Set the given timer counter value.
////////////////////////////////////////////////////////////////////////////////
void DIOSetCounter(DIO_DEVICE_EXTENSION *deviceInfo,
                   COUNTER_IOCTL_INPUT *counterInput,
                   BOOLEAN initialize)
{

    // **** Local Declarations ****

    // Timer control value.
    UCHAR timerControl;

    // Timer control register address for given counter.
    void *timerControlAddress;

    // Data address of given counter.
    void *counterDataAddress;

    // Union/structure to refer to the counter data as two bytes and as a
    // single short integer.
    union
    {
        USHORT CounterValue;
        struct
        {
            UCHAR LSB;
            UCHAR MSB;
        };
    } counterUnion;

    // Structure to hold context for synchronized call to set the
    // counter value.
    SYNC_COUNTER syncCounter;

    // **** Code Statements ****

    // Switch/case on the counter.  Create the proper timer control for
    // the given counter if the initialize flag is set.
    //
    //     Timer Control
    //        XX11_MMM0
    //        XX  _       Select Counter
    //          11_       Write LSB, MSB
    //            _MMM    Mode
    //            _   0   Binary Counter
    //
    switch (counterInput->Counter)
    {
        // Timer A Counters
        case COUNTERA0:
            timerControl = 0x30;
            timerControlAddress = deviceInfo->TimerControlA;
            counterDataAddress = (UCHAR *) deviceInfo->CounterA0;
            break;
        case COUNTERA1:
            timerControl = 0x70;
            timerControlAddress = deviceInfo->TimerControlA;
            counterDataAddress = (UCHAR *) deviceInfo->CounterA1;

            // Note:  When set in mode 2, this is the low order word
            // of the interrupt counter.  Save this value for the
            // interrupt service routine.
            if (initialize)
            {
                if (counterInput->Mode == 2)
                    deviceInfo->ClockPeriodicInterruptA1Set = TRUE;
                else
                    deviceInfo->ClockPeriodicInterruptA1Set = FALSE;
            }
            break;
        case COUNTERA2:
            timerControl = 0xb0;
            timerControlAddress = deviceInfo->TimerControlA;
            counterDataAddress = (UCHAR *) deviceInfo->CounterA2;

            // Note:  When set in mode 2, this is the high order word
            // of the interrupt counter.  Save this value for the
            // interrupt service routine.
            if (initialize)
            {
                if (counterInput->Mode == 2)
                    deviceInfo->ClockPeriodicInterruptA2Set = TRUE;
                else
                    deviceInfo->ClockPeriodicInterruptA2Set = FALSE;
            }
            break;

        // Timer B counters
        case COUNTERB0:
            timerControl = 0x30;
            timerControlAddress = deviceInfo->TimerControlB;
            counterDataAddress = (UCHAR *) deviceInfo->CounterB0;
            break;
        case COUNTERB1:
            timerControl = 0x70;
            timerControlAddress = deviceInfo->TimerControlB;
            counterDataAddress = (UCHAR *) deviceInfo->CounterB1;
            break;
        case COUNTERB2:
            timerControl = 0xb0;
            timerControlAddress = deviceInfo->TimerControlB;
            counterDataAddress = (UCHAR *) deviceInfo->CounterB2;
            break;

        // Timer C counters.
        case COUNTERC0:
            timerControl = 0x30;
            timerControlAddress = deviceInfo->TimerControlC;
            counterDataAddress = (UCHAR *) deviceInfo->CounterC0;
            break;
        case COUNTERC1:
            timerControl = 0x70;
            timerControlAddress = deviceInfo->TimerControlC;
            counterDataAddress = (UCHAR *) deviceInfo->CounterC1;
            break;
        case COUNTERC2:
            timerControl = 0xb0;
            timerControlAddress = deviceInfo->TimerControlC;
            counterDataAddress = (UCHAR *) deviceInfo->CounterC2;
            break;
    }

    // Set the mode bits and the initial counter value.
    if (initialize)
        timerControl |= (counterInput->Mode << 1);
    counterUnion.CounterValue = counterInput->InitValue;

    // Setup for synchronized call at device's IRQL to prevent the
    // registers from changing during the counter set.
    syncCounter.PortMemoryType = deviceInfo->PortMemoryType;
    if (initialize)
    {
        syncCounter.TimerControlAddress = timerControlAddress;
        syncCounter.TimerControl = timerControl;
    }
    else
    {
        syncCounter.TimerControlAddress = NULL;
        syncCounter.TimerControl = 0x00;
    }
    syncCounter.CounterDataAddress = counterDataAddress;
    syncCounter.CounterLSB = &counterUnion.LSB;
    syncCounter.CounterMSB = &counterUnion.MSB;

    // Perform synchronized call to set counter value.
    KeSynchronizeExecution(deviceInfo->Interrupt,
                           (PKSYNCHRONIZE_ROUTINE) DIOSyncCounterSet,
                           (PVOID) &syncCounter);
    return;
}

////////////////////////////////////////////////////////////////////////////////
// Called synchronously to set the counter value at high IRQL to prevent
// register changing during the set.
////////////////////////////////////////////////////////////////////////////////
VOID DIOSyncCounterSet(VOID *context)
{

    // **** Local Declarations ****

    // Structure to hold context for synchronized call to set the
    // counter value.
    SYNC_COUNTER *syncCounter = context;

    // **** Code Statements ****

    // If the time control port is provided output the timer control
    // to the control register port.  This will initialize the counter
    // timer to the provided value immediately and set the provided
    // mode.
    if (syncCounter->TimerControlAddress != NULL)
    {
 
        // Output the timer control to the timer control port in I/O space.
        if (syncCounter->PortMemoryType == 1)
        {
            WRITE_PORT_UCHAR((UCHAR *) syncCounter->TimerControlAddress,
                             syncCounter->TimerControl);
        }

        // Output the timer control to the timer control port in memory space.
        else
        {
            WRITE_REGISTER_UCHAR((UCHAR *) syncCounter->TimerControlAddress,
                                 syncCounter->TimerControl);
        }
    }

    // Output the counter low byte in I/O space.
    if (syncCounter->PortMemoryType == 1)
        WRITE_PORT_UCHAR(syncCounter->CounterDataAddress,
                         *syncCounter->CounterLSB);

    // Output the counter low byte in memory space.
    else
        WRITE_REGISTER_UCHAR(syncCounter->CounterDataAddress,
                             *syncCounter->CounterLSB);

    // Output the counter high byte in I/O space.
    if (syncCounter->PortMemoryType == 1)
        WRITE_PORT_UCHAR(syncCounter->CounterDataAddress,
                         *syncCounter->CounterMSB);

    // Output the counter high byte in memory space.
    else
        WRITE_REGISTER_UCHAR(syncCounter->CounterDataAddress,
                             *syncCounter->CounterMSB);
}

////////////////////////////////////////////////////////////////////////////////
// Read the given timer counter value.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOReadCounter(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                        IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Structure containing COUNTER data from application.
    COUNTER_IOCTL_INPUT *counterInput;

    // Structure containing COUNTER data to application.
    COUNTER_IOCTL_OUTPUT *counterOutput;

    // **** Code Statements ****

    // No data yet to return.
    pIRP->IoStatus.Information = 0;

    // If this device does not have a clock, return STATUS_NOT_IMPLEMENTED.
    if (!deviceInfo->HasClock)
        return STATUS_NOT_IMPLEMENTED;

    // The input size the IRP is providing MUST be the size of the
    // COUNTER_IOCTL_INPUT structure.
    if (pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
        sizeof(COUNTER_IOCTL_INPUT))
        return STATUS_INVALID_PARAMETER;

    // The data to return is a COUNTER_IOCTL_OUTPUT structure.  Verify
    // the output size the IRP is expected to return is the same size as
    // the structure.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength !=
        sizeof(COUNTER_IOCTL_OUTPUT))
        return STATUS_INVALID_PARAMETER;

    // The device IOCTL is set as METHOD_BUFFERED.  NT copies the user's
    // input buffer to the IRP's SystemBuffer entry before entry and
    // copies the IRP's SystemBuffer entry to the user's output buffer
    // after return.  Point to the SystemBuffer as both a
    // COUNTER_IOCTL_INPUT and a COUNTER_IOCTL_OUTPUT structure.
    counterInput =
        (COUNTER_IOCTL_INPUT *) pIRP->AssociatedIrp.SystemBuffer;
    counterOutput =
        (COUNTER_IOCTL_OUTPUT *) pIRP->AssociatedIrp.SystemBuffer;

    // Check to be sure the given counter is valid (COUNTERA0 - COUNTERC2).
    if (counterInput->Counter > COUNTERC2)
        return STATUS_INVALID_PARAMETER;

    // Call the DIOGetCounter function to get the counter's value.
    counterOutput->CounterValue = DIOGetCounter(deviceInfo,
                                                counterInput->Counter);

    // Give the number of bytes returned.
    pIRP->IoStatus.Information = sizeof(COUNTER_IOCTL_OUTPUT);

    // Successfully finished.
    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Get the current value of the given counter.
////////////////////////////////////////////////////////////////////////////////
USHORT DIOGetCounter(DIO_DEVICE_EXTENSION *deviceInfo, enum COUNTER counter)
{

    // **** Local Declarations ****

    // Timer control value.
    UCHAR timerControl;

    // Timer control register address for given counter.
    void *timerControlAddress;

    // Data address of given counter.
    void *counterDataAddress;

    // Union/structure to refer to the counter as two bytes and as a single
    // short integer.
    union
    {
        USHORT CounterValue;
        struct
        {
            UCHAR LSB;
            UCHAR MSB;
        };
    } counterUnion;

    // Structure to hold context for synchronized call to get the
    // counter value.
    SYNC_COUNTER syncCounter;

    // **** Code Statements ****

    // Switch/case on the counter.  Note that since this is only called
    // internally by the driver, we do not validate the counter (should
    // already be validated).  Create the proper timer control for the
    // given counter.
    //
    //     Timer Control
    //        XX00_MMM0
    //        XX  _       Select Counter
    //          00_       Latch Counter Command
    //            _MMM    Mode
    //            _   0   Binary Counter
    //
    switch (counter)
    {

        // Timer A counters
        case COUNTERA0:
            timerControl = 0x00;
            timerControlAddress = deviceInfo->TimerControlA;
            counterDataAddress = (UCHAR *) deviceInfo->CounterA0;
            break;
        case COUNTERA1:
            timerControl = 0x40;
            timerControlAddress = deviceInfo->TimerControlA;
            counterDataAddress = (UCHAR *) deviceInfo->CounterA1;
            break;
        case COUNTERA2:
            timerControl = 0x80;
            timerControlAddress = deviceInfo->TimerControlA;
            counterDataAddress = (UCHAR *) deviceInfo->CounterA2;
            break;

        // Timer B counters
        case COUNTERB0:
            timerControl = 0x00;
            timerControlAddress = deviceInfo->TimerControlB;
            counterDataAddress = (UCHAR *) deviceInfo->CounterB0;
            break;
        case COUNTERB1:
            timerControl = 0x40;
            timerControlAddress = deviceInfo->TimerControlB;
            counterDataAddress = (UCHAR *) deviceInfo->CounterB1;
            break;
        case COUNTERB2:
            timerControl = 0x80;
            timerControlAddress = deviceInfo->TimerControlB;
            counterDataAddress = (UCHAR *) deviceInfo->CounterB2;
            break;

        // Timer C counters
        case COUNTERC0:
            timerControl = 0x00;
            timerControlAddress = deviceInfo->TimerControlC;
            counterDataAddress = (UCHAR *) deviceInfo->CounterC0;
            break;
        case COUNTERC1:
            timerControl = 0x40;
            timerControlAddress = deviceInfo->TimerControlC;
            counterDataAddress = (UCHAR *) deviceInfo->CounterC1;
            break;
        case COUNTERC2:
            timerControl = 0x80;
            timerControlAddress = deviceInfo->TimerControlC;
            counterDataAddress = (UCHAR *) deviceInfo->CounterC2;
            break;
    }

    // Setup for synchronized call at device's IRQL to prevent an IRQ
    // generated ISR from getting in the middle of the counter read.
    syncCounter.PortMemoryType = deviceInfo->PortMemoryType;
    syncCounter.TimerControlAddress = timerControlAddress;
    syncCounter.TimerControl = timerControl;
    syncCounter.CounterDataAddress = counterDataAddress;
    syncCounter.CounterLSB = &counterUnion.LSB;
    syncCounter.CounterMSB = &counterUnion.MSB;

    // Perform synchronized call to get counter value.
    KeSynchronizeExecution(deviceInfo->Interrupt,
                           (PKSYNCHRONIZE_ROUTINE) DIOSyncCounterRead,
                           (PVOID) &syncCounter);

    // Return the counter value.
    return counterUnion.CounterValue;
}

////////////////////////////////////////////////////////////////////////////////
// Called synchronously to get the counter value at high IRQL to prevent
// register changing during the read.
////////////////////////////////////////////////////////////////////////////////
VOID DIOSyncCounterRead(VOID *context)
{

    // **** Local Declarations ****

    // Structure to hold context for synchronized call to get the
    // counter value.
    SYNC_COUNTER *syncCounter = context;

    // **** Code Statements ****

    // Output the timer control to the timer control port in I/O space.
    if (syncCounter->PortMemoryType == 1)
    {
        WRITE_PORT_UCHAR((UCHAR *) syncCounter->TimerControlAddress,
                         syncCounter->TimerControl);
    }

    // Output the timer control to the timer control port in memory space.
    else
    {
        WRITE_REGISTER_UCHAR((UCHAR *) syncCounter->TimerControlAddress,
                             syncCounter->TimerControl);
    }

    // Get the counter low byte in I/O space.
    if (syncCounter->PortMemoryType == 1)
        *syncCounter->CounterLSB =
            READ_PORT_UCHAR(syncCounter->CounterDataAddress);

    // Get the counter low byte in memory space.
    else
        *syncCounter->CounterLSB =
            READ_REGISTER_UCHAR(syncCounter->CounterDataAddress);

    // Get the counter high byte in I/O space.
    if (syncCounter->PortMemoryType == 1)
        *syncCounter->CounterMSB =
            READ_PORT_UCHAR(syncCounter->CounterDataAddress);

    // Get the counter high byte in memory space.
    else
        *syncCounter->CounterMSB =
            READ_REGISTER_UCHAR(syncCounter->CounterDataAddress);
}

////////////////////////////////////////////////////////////////////////////////
// Read all counters of the given timer latched together.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOReadTimer(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                      IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Structure containing TIMER data from application.
    TIMER_IOCTL_INPUT *timerInput;

    // Structure containing TIMER data to application.
    TIMER_IOCTL_OUTPUT *timerOutput;

    // **** Code Statements ****

    // No data yet to return.
    pIRP->IoStatus.Information = 0;

    // If this device does not have a clock, return STATUS_NOT_IMPLEMENTED.
    if (!deviceInfo->HasClock)
        return STATUS_NOT_IMPLEMENTED;

    // The input size the IRP is providing MUST be the size of the
    // TIMER_IOCTL_INPUT structure.
    if (pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
        sizeof(TIMER_IOCTL_INPUT))
        return STATUS_INVALID_PARAMETER;

    // The data to return is a TIMER_IOCTL_OUTPUT structure.  Verify
    // the output size the IRP is expected to return is the same size as
    // the structure.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength !=
        sizeof(TIMER_IOCTL_OUTPUT))
        return STATUS_INVALID_PARAMETER;

    // The device IOCTL is set as METHOD_BUFFERED.  NT copies the user's
    // input buffer to the IRP's SystemBuffer entry before entry and
    // copies the IRP's SystemBuffer entry to the user's output buffer
    // after return.  Point to the SystemBuffer as both a
    // TIMER_IOCTL_INPUT and a TIMER_IOCTL_OUTPUT structure.
    timerInput =
        (TIMER_IOCTL_INPUT *) pIRP->AssociatedIrp.SystemBuffer;
    timerOutput =
        (TIMER_IOCTL_OUTPUT *) pIRP->AssociatedIrp.SystemBuffer;

    // Check to be sure the given timer is valid (TIMERA - TIMERC).
    if (timerInput->Timer > TIMERC)
        return STATUS_INVALID_PARAMETER;

    // Call the DIOGetTimer function to get all counters' values.
    DIOGetTimer(deviceInfo, timerInput->Timer, timerOutput, TRUE);

    // Give the number of bytes returned.
    pIRP->IoStatus.Information = sizeof(TIMER_IOCTL_OUTPUT);

    // Successfully finished.
    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Get the current value of all counters latched together.  NOTE:  This
// is also called from the ISR and must not call KeSynchronizeExecution
// but may call the routine directly.
////////////////////////////////////////////////////////////////////////////////
void DIOGetTimer(DIO_DEVICE_EXTENSION *deviceInfo, enum TIMER timer,
                 TIMER_IOCTL_OUTPUT *counterValues, BOOLEAN syncISR)
{

    // **** Local Declarations ****

    // Timer control value.
    UCHAR timerControl;

    // Timer control register address for given counter.
    void *timerControlAddress;

    // Data address of all counters of the selected timer.
    void *counter0DataAddress;
    void *counter1DataAddress;
    void *counter2DataAddress;

    // Structure to hold context for synchronized call to get the
    // timer value.
    SYNC_TIMER syncTimer;

    // **** Code Statements ****

    // Switch/case on the timer.  Note that since this is only called
    // internally by the driver, we do not validate the timer (should
    // already be validated).  Create the proper timer control for the
    // given timer to latch all counters.
    //
    //     Timer Control
    //        1100_0000
    //        11  _       Read-Back Command
    //          0 _       Latch Value Of Selected Counters
    //           1_       Do Not Latch Status Of Selected Counters
    //            _111    Select All Counters (2, 1, and 0)
    //            _   0   Reserved - Must Be Zero
    //
    switch (timer)
    {

        case TIMERA:
            timerControlAddress = deviceInfo->TimerControlA;
            counter0DataAddress = (UCHAR *) deviceInfo->CounterA0;
            counter1DataAddress = (UCHAR *) deviceInfo->CounterA1;
            counter2DataAddress = (UCHAR *) deviceInfo->CounterA2;
            break;
        case TIMERB:
            timerControlAddress = deviceInfo->TimerControlB;
            counter0DataAddress = (UCHAR *) deviceInfo->CounterB0;
            counter1DataAddress = (UCHAR *) deviceInfo->CounterB1;
            counter2DataAddress = (UCHAR *) deviceInfo->CounterB2;
            break;
        case TIMERC:
            timerControlAddress = deviceInfo->TimerControlC;
            counter0DataAddress = (UCHAR *) deviceInfo->CounterC0;
            counter1DataAddress = (UCHAR *) deviceInfo->CounterC1;
            counter2DataAddress = (UCHAR *) deviceInfo->CounterC2;
            break;
    }

    // The timer control value will be the same since always getting
    // all three counter values with no status.
    timerControl = 0xDE;

    // Setup for synchronized call at device's IRQL to prevent an IRQ
    // generated ISR from getting in the middle of the timer read.
    syncTimer.PortMemoryType = deviceInfo->PortMemoryType;
    syncTimer.TimerControlAddress = timerControlAddress;
    syncTimer.TimerControl = timerControl;
    syncTimer.Counter0DataAddress = counter0DataAddress;
    syncTimer.Counter1DataAddress = counter1DataAddress;
    syncTimer.Counter2DataAddress = counter2DataAddress;
    syncTimer.Counter0Value = &(counterValues->Counter0Value);
    syncTimer.Counter1Value = &(counterValues->Counter1Value);
    syncTimer.Counter2Value = &(counterValues->Counter2Value);

    // If the syncISR flag is TRUE, perform synchronized call with the
    // ISR to get counter value (prevents the ISR from corrupting the
    // value during the read).  If the syncISR flag is FALSE, this
    // is being called from the ISR and cannot be synchronized but
    // may be called directly.
    if (syncISR)
        KeSynchronizeExecution(deviceInfo->Interrupt,
                               (PKSYNCHRONIZE_ROUTINE) DIOSyncTimerRead,
                               (PVOID) &syncTimer);
    else
        DIOSyncTimerRead((PVOID) &syncTimer);
}

////////////////////////////////////////////////////////////////////////////////
// Called synchronously to get the timer value at high IRQL to prevent
// register changing by the ISR during the read.  Called directly if
// from within the ISR.
////////////////////////////////////////////////////////////////////////////////
VOID DIOSyncTimerRead(VOID *context)
{

    // **** Local Declarations ****

    // Structure to hold context for synchronized call to get the
    // timer value.
    SYNC_TIMER *syncTimer = context;

    // **** Code Statements ****

    // Union/structure to refer to the counter as two bytes and as a single
    // short integer.
    union
    {
        USHORT CounterValue;
        struct
        {
            UCHAR LSB;
            UCHAR MSB;
        };
    } counterUnion;

    // Output the timer control to the timer control port in I/O space.
    if (syncTimer->PortMemoryType == 1)
    {
        WRITE_PORT_UCHAR((UCHAR *) syncTimer->TimerControlAddress,
                         syncTimer->TimerControl);
    }

    // Output the timer control to the timer control port in memory space.
    else
    {
        WRITE_REGISTER_UCHAR((UCHAR *) syncTimer->TimerControlAddress,
                             syncTimer->TimerControl);
    }

    //** Important note:  The counters are being read and, presumably,
    //** are being read together for the purpose of using a 32-bit or
    //** 48-bit counter that is, in some way, latched together and
    //** using MODE 2.  In this mode, the OUT line is toggled on the
    //** transition from 0x0002 to 0x0001 and not from the obvious
    //** transition from 0x0000 to the reset value of the counter.
    //** Due to this anomolie, to read the value properly, two must
    //** be subtracted from the value.  Otherwise the greater
    //** significant word will roll over of the lesser signficant
    //** word and produce incorrect values when the lesser
    //** significant word is 0x0001 or 0x0000.  When saving
    //** the value for return, two is subtracted.

    // Get the counter 0 low byte in I/O space.
    if (syncTimer->PortMemoryType == 1)
        counterUnion.LSB = READ_PORT_UCHAR(syncTimer->Counter0DataAddress);

    // Get the counter 0 low byte in memory space.
    else
        counterUnion.LSB = READ_REGISTER_UCHAR(syncTimer->Counter0DataAddress);

    // Get the counter 0 high byte in I/O space.
    if (syncTimer->PortMemoryType == 1)
        counterUnion.MSB = READ_PORT_UCHAR(syncTimer->Counter0DataAddress);

    // Get the counter 0 high byte in memory space.
    else
        counterUnion.MSB = READ_REGISTER_UCHAR(syncTimer->Counter0DataAddress);

    // Save the counter 0 value.
    *(syncTimer->Counter0Value) = counterUnion.CounterValue - 2;

    // Get the counter 1 low byte in I/O space.
    if (syncTimer->PortMemoryType == 1)
        counterUnion.LSB = READ_PORT_UCHAR(syncTimer->Counter1DataAddress);

    // Get the counter 1 low byte in memory space.
    else
        counterUnion.LSB = READ_REGISTER_UCHAR(syncTimer->Counter1DataAddress);

    // Get the counter 1 high byte in I/O space.
    if (syncTimer->PortMemoryType == 1)
        counterUnion.MSB = READ_PORT_UCHAR(syncTimer->Counter1DataAddress);

    // Get the counter 1 high byte in memory space.
    else
        counterUnion.MSB = READ_REGISTER_UCHAR(syncTimer->Counter1DataAddress);

    // Save the counter 1 value.
    *(syncTimer->Counter1Value) = counterUnion.CounterValue - 2;

    // Get the counter 2 low byte in I/O space.
    if (syncTimer->PortMemoryType == 1)
        counterUnion.LSB = READ_PORT_UCHAR(syncTimer->Counter2DataAddress);

    // Get the counter 2 low byte in memory space.
    else
        counterUnion.LSB = READ_REGISTER_UCHAR(syncTimer->Counter2DataAddress);

    // Get the counter 2 high byte in I/O space.
    if (syncTimer->PortMemoryType == 1)
        counterUnion.MSB = READ_PORT_UCHAR(syncTimer->Counter2DataAddress);

    // Get the counter 2 high byte in memory space.
    else
        counterUnion.MSB = READ_REGISTER_UCHAR(syncTimer->Counter2DataAddress);

    // Save the counter 2 value.
    *(syncTimer->Counter2Value) = counterUnion.CounterValue - 2;
}

////////////////////////////////////////////////////////////////////////////////
// Respond to a read or write request.  This driver does not support reads or
// writes.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOReadWrite(DEVICE_OBJECT *deviceObject, IRP *pIRP)
{

    // **** Local Declarations ****

    // Status to return from NT calls.
    NTSTATUS status;

    // Get the local data structure for the PCI-DIO24 device.
    DIO_DEVICE_EXTENSION *deviceInfo;

    // **** Code Statements ****

    // Get the device information and the driver object.
    deviceInfo = deviceObject->DeviceExtension;

    // If holding requests, queue this IRP.
    if (deviceInfo->HoldNewRequests)
    {
        status = DIOQueueRequest(deviceInfo, pIRP);
        if (NT_SUCCESS(status))
            status = STATUS_PENDING;
        return status;
    }

    // Simply return a not implemented error.
    pIRP->IoStatus.Information = 0;
    pIRP->IoStatus.Status = STATUS_NOT_IMPLEMENTED;
    IoCompleteRequest(pIRP, IO_NO_INCREMENT);
    return STATUS_NOT_IMPLEMENTED;
}

////////////////////////////////////////////////////////////////////////////////
// Send an error to the event logger.
////////////////////////////////////////////////////////////////////////////////
void LogDIOError(DRIVER_OBJECT *driverObject, DEVICE_OBJECT *deviceObject,
                 PHYSICAL_ADDRESS p1, PHYSICAL_ADDRESS p2,
                 ULONG sequenceNumber, UCHAR majorFunctionCode,
                 UCHAR retryCount, ULONG uniqueErrorValue,
                 NTSTATUS finalStatus, NTSTATUS specificIOStatus,
                 ULONG lengthOfInsert1, WCHAR *insert1,
                 ULONG lengthOfInsert2, WCHAR *insert2)
{

    // **** Arguments ****
    //
    // driverObject - A pointer to the driver object for the device.
    //
    // deviceObject - A pointer to the device object associated with the
    // device that had the error, early in initialization, one may not
    // yet exist.
    //
    // p1,p2 - If phyical addresses for the controller ports involved
    // with the error are available, put them through as dump data.
    //
    // sequenceNumber - A ulong value that is unique to an IRP over the
    // life of the irp in this driver - 0 generally means an error not
    // associated with an irp.
    //
    // majorFunctionCode - If there is an error associated with the irp,
    // this is the major function code of that irp.
    //
    // retryCount - The number of times a particular operation has been
    // retried.
    //
    // uniqueErrorValue - A unique long word that identifies the particular
    // call to this function.
    //
    // finalStatus - The final status given to the irp that was associated
    // with this error.  If this log entry is being made during one of
    // the retries this value will be STATUS_SUCCESS.
    //
    // specificIOStatus - The IO status for a particular error.
    //
    // lengthOfInsert1 - The length in bytes (including the terminating NULL)
    //                   of the first insertion string.
    //
    // insert1 - The first insertion string.
    //
    // lengthOfInsert2 - The length in bytes (including the terminating NULL)
    //                   of the second insertion string.  NOTE, there must
    //                   be a first insertion string for their to be
    //                   a second insertion string.
    //
    // insert2 - The second insertion string.

    // **** Local Declarations ****

    // Pointer to either driver or device object.
    void *objectToUse;

    // Number of bytes to allocate for the error log entry.
    SHORT dumpToAllocate = 0;

    // Error log entry pointer.
    IO_ERROR_LOG_PACKET *errorLogEntry;

    // Size of error log entry.
    SHORT errorLogSize = 0;

    // Pointers to string insertion points.
    UCHAR *ptrToFirstInsert;
    UCHAR *ptrToSecondInsert;

    // **** Code Statements ****

    // Verify the string length is zero if the string is NULL.
    if (insert1 == NULL)
        lengthOfInsert1 = 0;
    if (insert2 == NULL)
        lengthOfInsert2 = 0;

    // Use the device object if it is given, otherwise, use
    // the driver object.
    if (ARGUMENT_PRESENT(deviceObject))
        objectToUse = deviceObject;
    else
        objectToUse = driverObject;

    // Find how much dump space to allocate.
    if (p1.QuadPart != PhysicalZero.QuadPart)
        dumpToAllocate += (SHORT) sizeof(PHYSICAL_ADDRESS);
    if (p2.QuadPart != PhysicalZero.QuadPart)
        dumpToAllocate += (SHORT) sizeof(PHYSICAL_ADDRESS);

    // Find how many bytes required for error log entry.
    errorLogSize = (SHORT) (sizeof(IO_ERROR_LOG_PACKET) + dumpToAllocate +
                            lengthOfInsert1 + lengthOfInsert2);

    // Allocate an error log entry.
    errorLogEntry = IoAllocateErrorLogEntry(objectToUse, (UCHAR) errorLogSize);

    // Populate the error log entry if it was properly allocated.
    if (errorLogEntry != NULL)
    {
        errorLogEntry->ErrorCode = specificIOStatus;
        errorLogEntry->SequenceNumber = sequenceNumber;
        errorLogEntry->MajorFunctionCode = majorFunctionCode;
        errorLogEntry->RetryCount = retryCount;
        errorLogEntry->UniqueErrorValue = uniqueErrorValue;
        errorLogEntry->FinalStatus = finalStatus;
        errorLogEntry->DumpDataSize = dumpToAllocate;

        // Output p1 and/or p2 to dump and set the pointer to
        // the first and second string insertion points.
        ptrToFirstInsert = (UCHAR *) &errorLogEntry->DumpData[0];
        if (dumpToAllocate > 0)
        {
            ptrToFirstInsert += sizeof(PHYSICAL_ADDRESS);
            if (p1.QuadPart != PhysicalZero.QuadPart)
            {
                RtlCopyMemory(&errorLogEntry->DumpData[0], &p1,
                              sizeof(PHYSICAL_ADDRESS));
                if (p2.QuadPart != PhysicalZero.QuadPart)
                {
                    RtlCopyMemory(((UCHAR *) &errorLogEntry->DumpData[0])
                                  + sizeof(PHYSICAL_ADDRESS), &p2,
                                  sizeof(PHYSICAL_ADDRESS));
                    ptrToFirstInsert += sizeof(PHYSICAL_ADDRESS);
                }
            }
            else
                RtlCopyMemory(&errorLogEntry->DumpData[0], &p2,
                              sizeof(PHYSICAL_ADDRESS));
        }
        ptrToSecondInsert = ptrToFirstInsert + lengthOfInsert1;

        // Insert the strings.  If there is a second without a first,
        // use the second as the first.
        errorLogEntry->NumberOfStrings = 0;
        errorLogEntry->StringOffset = (USHORT) (ptrToFirstInsert -
                                      (UCHAR *)errorLogEntry);
        if (lengthOfInsert1)
        {
            errorLogEntry->NumberOfStrings++;
            RtlCopyMemory(ptrToFirstInsert, insert1, lengthOfInsert1);
        }
        if (lengthOfInsert2)
        {
            errorLogEntry->NumberOfStrings++;
            RtlCopyMemory(ptrToSecondInsert, insert2, lengthOfInsert2);
        }

        // Output the entry to the error log.
        IoWriteErrorLogEntry(errorLogEntry);
    }
}

////////////////////////////////////////////////////////////////////////////////
// Set the timer interrupt period in milliseconds.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOSetTimerInterruptPeriod(DIO_DEVICE_EXTENSION *deviceInfo,
                                    IRP *pIRP,
                                    IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Status to return from NT calls.
    NTSTATUS status;

    // Structure containing the timer interrupt period in milliseconds.
    PERIOD_IOCTL_INPUT *periodInput;

    // Structure containing the set counter values.
    TIMER_IOCTL_OUTPUT *setTimerOutput;

    // Structure for calling DIOSetCounter.
    COUNTER_IOCTL_INPUT counterInput;

    // Value of the period.
    ULONG period;

    // **** Code Statements ****

    // No data will be returned.
    pIRP->IoStatus.Information = 0;

    // If this device does not have a clock, return STATUS_NOT_IMPLEMENTED.
    if (!deviceInfo->HasClock)
        return STATUS_NOT_IMPLEMENTED;

    // The input size the IRP is providing MUST be the size of the
    // PERIOD_IOCTL_INPUT structure.
    if (pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
        sizeof(PERIOD_IOCTL_INPUT))
        return STATUS_INVALID_PARAMETER;

    // The data to return is a TIMER_IOCTL_OUTPUT structure.  Verify
    // the output size the IRP is expected to return is the same size as
    // the structure.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength !=
        sizeof(TIMER_IOCTL_OUTPUT))
        return STATUS_INVALID_PARAMETER;

    // The device IOCTL is set as METHOD_BUFFERED.  NT copies the user's
    // input buffer to the IRP's SystemBuffer entry before entry and
    // copies the IRP's SystemBuffer entry to the user's output buffer
    // after return.  The output buffer is not used as there
    // is no output sent to the user.  Point to the SystemBuffer as
    // a PERIOD_IOCTL_INPUT structure.
    periodInput = (PERIOD_IOCTL_INPUT *) pIRP->AssociatedIrp.SystemBuffer;
    setTimerOutput =
        (TIMER_IOCTL_OUTPUT *) pIRP->AssociatedIrp.SystemBuffer;

    //** The timer interrupt is controlled by COUNTERA1 and COUNTERA2.  The
    //** initial value of COUNTERA1 and COUNTERA2 makes up the periodic
    //** counter when the counters are set to mode 2.  The period  in
    //** microseconds is  given by multiplying the two counts.  As the
    //** period given in timerInput is in milliseconds, COUNTERA1 will be
    //** initialized to 500 (0x1f4) and COUNTERA2 will be set to the given
    //** period multiplied by 2 (the values 0 and 1 are invalid).  This
    //** gives a range of 1ms - 32767ms (.001s - 32.767s).

    // Validate the range of the given period.
    period = periodInput->Period;
    if ( period > 32767 )
        return STATUS_INVALID_PARAMETER;

    // If the period is specified as zero, initialize the counters to
    // non-periodic counters to stop timers.
    if (period == 0)
    {
        setTimerOutput->Counter0Value = 0x0000;
        counterInput.Counter = COUNTERA1;
        counterInput.InitValue = 0x0000;
        setTimerOutput->Counter1Value = 0x0000;
        counterInput.Mode = 0;
        DIOSetCounter(deviceInfo, &counterInput, TRUE);
        counterInput.Counter = COUNTERA2;
        counterInput.InitValue = 0x0000;
        setTimerOutput->Counter2Value = 0x0000;
        counterInput.Mode = 0;
        DIOSetCounter(deviceInfo, &counterInput, TRUE);
        deviceInfo->ClockInterruptPeriod = 0;
        pIRP->IoStatus.Information = sizeof(TIMER_IOCTL_OUTPUT);
        return STATUS_SUCCESS;
    }

    // Set the COUNTER_IOCTL_INPUT structure and initialize each counter.
    // Note that timer A is initialized as a periodic 32-bit timer (COUNTERA2
    // is MSW and COUNTERA1 is LSW and expiration of COUNTERA2 generates the
    // periodic interrupt.  Also, initialize timer B as a 32-bit microsecond
    // counter (COUNTERB2 is MSW and COUNTERB1 is LSW).
    setTimerOutput->Counter0Value = 0x0000;
    counterInput.Counter = COUNTERA1;
    counterInput.InitValue = 0x0002;
    setTimerOutput->Counter1Value = 0x0002;
    counterInput.Mode = 2;
    DIOSetCounter(deviceInfo, &counterInput, TRUE);
    counterInput.Counter = COUNTERA2;
    counterInput.InitValue = ((USHORT) ((period * 1000) >> 1));
    setTimerOutput->Counter2Value = counterInput.InitValue;
    counterInput.Mode = 2;
    DIOSetCounter(deviceInfo, &counterInput, TRUE);
    counterInput.Counter = COUNTERB1;
    counterInput.InitValue = 0x0000;
    counterInput.Mode = 2;
    DIOSetCounter(deviceInfo, &counterInput, TRUE);
    counterInput.Counter = COUNTERB2;
    counterInput.InitValue = 0x0000;
    counterInput.Mode = 2;
    DIOSetCounter(deviceInfo, &counterInput, TRUE);

    // Save the timer interrupt period (in MICROSECONDS) so the ISR
    // doesn't have to do any calculations.
    deviceInfo->ClockInterruptPeriod = period * 1000;

    // Give the number of bytes returned.
    pIRP->IoStatus.Information = sizeof(TIMER_IOCTL_OUTPUT);

    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Add event for timer/external interrupt handler to set.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOAddInterruptEvent(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                              IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Status to return from NT calls.
    NTSTATUS status;

    // Structure containing the number of ticks before setting an event and
    // a handle to the previously created user event to set.
    INTERRUPT_HANDLE_IOCTL_INPUT *interruptEventInput;

    // Pointer to event object retrieved from given handle.
    KEVENT *eventObject;

    // Pointer to newly created interrupt event information structure.
    INTERRUPT_EVENT_INFO *interruptEventInfo;

    // Pointers to event information structure while traversing the list.
    INTERRUPT_EVENT_INFO *nextInterruptEvent;
    INTERRUPT_EVENT_INFO *previousInterruptEvent;

    // Priority for next event in list.
    KPRIORITY priorityCount;

    // **** Code Statements ****

    // No data to be returned.
    pIRP->IoStatus.Information = 0;

    // Ensure that interrupts are not currently enabled.  Cannot add
    // interrupt events while interrupt processing is enabled.
    if (deviceInfo->InterruptHandled)
        return STATUS_DEVICE_BUSY;

    // The input size the IRP is providing MUST be the size of the
    // INTERRUPT_HANDLE_IOCTL_INPUT structure.
    if (pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
        sizeof(INTERRUPT_HANDLE_IOCTL_INPUT))
        return STATUS_INVALID_PARAMETER;

    // Since there is no data to output, verify the output size the
    // IRP is expected to return is zero.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength != 0)
        return STATUS_INVALID_PARAMETER;

    // The device IOCTL is set as METHOD_BUFFERED.  NT copies the user's
    // input buffer to the IRP's SystemBuffer entry before entry and
    // copies the IRP's SystemBuffer entry to the user's output buffer
    // after return.  The output buffer is not used as there
    // is no output sent to the user.  Point to the SystemBuffer as
    // a INTERRUPT_HANDLE_IOCTL_INPUT structure.
    interruptEventInput =
        (INTERRUPT_HANDLE_IOCTL_INPUT *) pIRP->AssociatedIrp.SystemBuffer;

    // A tick count of zero adds an event to be set for an external
    // interrupt.  A tick count greater than zero adds an event to be
    // set upon the given number of timer periods (i.e. if the
    // timer period is 0.01 seconds and the tick count is 10, the
    // event will be set every 0.10 seconds).  If the tick count
    // is greater than zero and this device does not have a clock,
    // return STATUS_NOT_IMPLEMENTED.
    if ((interruptEventInput->Ticks > 0) && (!deviceInfo->HasClock))
        return STATUS_NOT_IMPLEMENTED;

    // Verify there are not more than MAX_INTERRUPT_EVENTS already.
    if (deviceInfo->InterruptEventInfoCount >= MAX_INTERRUPT_EVENTS)
        return DIO_MAXIMUM_INTERRUPT_EVENTS;

    // Attempt to reference the event handle.
    status = ObReferenceObjectByHandle(interruptEventInput->EventHandle,
        0, NULL, UserMode, &eventObject, NULL);
    if (!NT_SUCCESS(status))
    {
        LogDIOError(deviceInfo->DriverObject, deviceInfo->DeviceObject,
                    PhysicalZero, PhysicalZero, 0,
                    0, 0, 0x51, status, DIO_CANNOT_REFERENCE_EVENT,
                    0, NULL, 0, NULL);
        return status;
    }

    // Ensure the event is initially reset (non signaled).
    KeClearEvent(eventObject);

    // Create interrupt event information structure.
    interruptEventInfo = (INTERRUPT_EVENT_INFO *) ExAllocatePool(NonPagedPool,
        sizeof(INTERRUPT_EVENT_INFO));
    if (interruptEventInfo == NULL)
    {
        LogDIOError(deviceInfo->DriverObject, deviceInfo->DeviceObject,
            PhysicalZero, PhysicalZero, 0, 0, 0, 0x52,
            STATUS_INSUFFICIENT_RESOURCES, DIO_INSUFFICIENT_RESOURCES,
            0, NULL, 0, NULL);
        ObDereferenceObject(eventObject);
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    // Clear structure.
    RtlZeroMemory(interruptEventInfo, sizeof(INTERRUPT_EVENT_INFO));

    // Initialize the members.
    interruptEventInfo->Ticks = interruptEventInput->Ticks;
    interruptEventInfo->InterruptEventHandle =
        interruptEventInput->EventHandle;
    interruptEventInfo->InterruptEventObject = eventObject;
    interruptEventInfo->ThreadHandle = interruptEventInput->ThreadHandle;
    interruptEventInfo->ThreadPriority = 0;

    // If the event list is empty, place the new one as the only one.
    if (deviceInfo->InterruptEventInfoList == NULL)
    {
        deviceInfo->InterruptEventInfoList = interruptEventInfo;
        deviceInfo->InterruptEventInfoCount = 1;
    }

    // If the event list is not empty, go through the link and place the
    // new one in the list (lowest tick first).  Note that a tick count
    // of zero is for an external interrupt and should be first.
    else
    {
        nextInterruptEvent = deviceInfo->InterruptEventInfoList;
        previousInterruptEvent = NULL;
        while (nextInterruptEvent != NULL)
        {

            // If the 'next' interrupt event has a greater tick count, place
            // in list BEFORE it.
            if (nextInterruptEvent->Ticks > interruptEventInfo->Ticks)
            {

                // If this is the first one, make the new one the first one.
                if (previousInterruptEvent == NULL)
                {
                    interruptEventInfo->NextInterruptEventInfo =
                        deviceInfo->InterruptEventInfoList;
                    deviceInfo->InterruptEventInfoList =
                        interruptEventInfo;
                    deviceInfo->InterruptEventInfoCount++;
                    break;
                }

                // If not the first one, make it between the previous and next.
                else
                {
                    interruptEventInfo->NextInterruptEventInfo =
                        nextInterruptEvent;
                    previousInterruptEvent->NextInterruptEventInfo =
                        interruptEventInfo;
                    deviceInfo->InterruptEventInfoCount++;
                    break;
                }
            }

            // If the 'next' interrupt event has the same tick count, it is
            // an error (only one of each tick count allowed).
            else if (nextInterruptEvent->Ticks == interruptEventInfo->Ticks)
            {
                ObDereferenceObject(eventObject);
                ExFreePool(interruptEventInfo);
                return DIO_DUPLICATE_EVENT;
            }

            // If the 'next' interrupt event has a lower tick count, goto the
            // 'next' interrupt event in the list.
            else
            {
                previousInterruptEvent = nextInterruptEvent;
                nextInterruptEvent =
                    nextInterruptEvent->NextInterruptEventInfo;

                // If past the end, place the new one at the end.
                if (nextInterruptEvent == NULL)
                {
                    previousInterruptEvent->NextInterruptEventInfo =
                        interruptEventInfo;
                    deviceInfo->InterruptEventInfoCount++;
                }
            }
        }
    }

    // Go through the entire list and make priorities in reverse order.
    // The highest priority (HIGH_PRIORITY) is for the first task event
    // in the list.  Each entry in list following the first is one
    // priority count lower.  Note that there CANNOT be more than 16
    // entries!  If the above errored due to an interrupt event in
    // the list already having the same tick count, this is not
    // reached...OK since nothing was added and priorities should be
    // correct.
    nextInterruptEvent = deviceInfo->InterruptEventInfoList;
    priorityCount = HIGH_PRIORITY;
    while (nextInterruptEvent != NULL)
    {

        // If the current priority is not the priority count, change
        // the priority.
        if (nextInterruptEvent->ThreadPriority != priorityCount)
        {
            nextInterruptEvent->ThreadPriority = priorityCount;
            status = ZwSetInformationThread(nextInterruptEvent->ThreadHandle,
                ThreadPriority, &nextInterruptEvent->ThreadPriority,
                sizeof(nextInterruptEvent->ThreadPriority));
            if (!NT_SUCCESS(status))
            {
                LogDIOError(deviceInfo->DriverObject,
                            deviceInfo->DeviceObject,
                            PhysicalZero, PhysicalZero, 0,
                            0, 0, 0x54, status,
                            DIO_CANNOT_SET_THREAD_PRIORITY,
                            0, NULL, 0, NULL);
            }
        }

        // Decrement the priorityCount for the next event in list and
        // goto the next event in list.
        priorityCount--;
        nextInterruptEvent = nextInterruptEvent->NextInterruptEventInfo;
    }

    // The new one has been place successfully in the list and the
    // priorities were reinitialized.
    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Remove event for timer/external interrupt handler.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIORemoveInterruptEvent(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                                 IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Status to return from NT calls.
    NTSTATUS status;

    // Structure containing the number of ticks before setting an event and
    // a handle to the previously created user event to set.
    EVENT_INFORMATION_IOCTL_INPUT *interruptEventInput;

    // Pointer to interrupt event information structure.
    INTERRUPT_EVENT_INFO *interruptEventInfo;

    // Pointers to event information structure while traversing the list.
    INTERRUPT_EVENT_INFO *nextInterruptEvent;
    INTERRUPT_EVENT_INFO *previousInterruptEvent;

    // Flag whether to start a new task.
    BOOLEAN startTask;

    // Priority for next event in list.
    KPRIORITY priorityCount;

    // **** Code Statements ****

    // No data to be returned.
    pIRP->IoStatus.Information = 0;

    // Ensure that interrupts are not currently enabled.  Cannot remove
    // interrupt events while interrupt processing is enabled.
    if (deviceInfo->InterruptHandled)
        return STATUS_DEVICE_BUSY;

    // The input size the IRP is providing MUST be the size of the
    // EVENT_INFORMATION_IOCTL_INPUT structure.
    if (pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
         sizeof(EVENT_INFORMATION_IOCTL_INPUT))
        return STATUS_INVALID_PARAMETER;

    // Since there is no data to output, verify the output size the
    // IRP is expected to return is zero.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength != 0)
        return STATUS_INVALID_PARAMETER;

    // The device IOCTL is set as METHOD_BUFFERED.  NT copies the user's
    // input buffer to the IRP's SystemBuffer entry before entry and
    // copies the IRP's SystemBuffer entry to the user's output buffer
    // after return.  The output buffer is not used as there
    // is no output sent to the user.  Point to the SystemBuffer as
    // a INTERRUPT_HANDLE_IOCTL_INPUT structure.
    interruptEventInput =
        (EVENT_INFORMATION_IOCTL_INPUT *) pIRP->AssociatedIrp.SystemBuffer;

    // Return immediately with a not found error if event not in list.
    if (deviceInfo->InterruptEventInfoList == NULL)
    {
        deviceInfo->InterruptEventInfoCount = 0;
        return DIO_EVENT_NOT_FOUND;
    }

    // Go through the list and find the given tick number to remove from
    // the list.
    nextInterruptEvent = deviceInfo->InterruptEventInfoList;
    previousInterruptEvent = NULL;
    while (nextInterruptEvent != NULL)
    {

        // If this is the tick to be removed, exit the loop.
        if (nextInterruptEvent->Ticks == interruptEventInput->Ticks)
            break;

        // If this is not the tick to be removed, goto the next in list.
        previousInterruptEvent = nextInterruptEvent;
        nextInterruptEvent = nextInterruptEvent->NextInterruptEventInfo;
    }

    // If here, event info not found.  Issue error!
    if (nextInterruptEvent == NULL)
        return DIO_EVENT_NOT_FOUND;

    // Look at the running flag.  If the thread is running, the next
    // event with the same thread context that is schedule must be
    // started.
    if (nextInterruptEvent->IsRunning)
        startTask = TRUE;
    else
        startTask = FALSE;

    // Remove this event from the interrupt event list.
    if (previousInterruptEvent == NULL)
    {
        interruptEventInfo = deviceInfo->InterruptEventInfoList;
        deviceInfo->InterruptEventInfoList =
            deviceInfo->InterruptEventInfoList->NextInterruptEventInfo;
        deviceInfo->InterruptEventInfoCount--;
    }
    else
    {
        interruptEventInfo = nextInterruptEvent;
        previousInterruptEvent->NextInterruptEventInfo =
            nextInterruptEvent->NextInterruptEventInfo;
        deviceInfo->InterruptEventInfoCount--;
    }

    // Go through the entire list and make priorities in reverse order.
    // The highest priority (HIGH_PRIORITY) is for the first task event
    // in the list.  Each entry in list following the first is one
    // priority count lower.  Note that there CANNOT be more than 16
    // entries!  If the above errored due to an interrupt event in
    // the list not being found, this is not reached...OK since
    // nothing was removed and priorities should be correct.
    nextInterruptEvent = deviceInfo->InterruptEventInfoList;
    priorityCount = HIGH_PRIORITY;
    while (nextInterruptEvent != NULL)
    {

        // If the current priority is not the priority count, change
        // the priority.
        if (nextInterruptEvent->ThreadPriority != priorityCount)
        {
            nextInterruptEvent->ThreadPriority = priorityCount;
            status = ZwSetInformationThread(nextInterruptEvent->ThreadHandle,
                ThreadPriority, &nextInterruptEvent->ThreadPriority,
                sizeof(nextInterruptEvent->ThreadPriority));
            if (!NT_SUCCESS(status))
            {
                LogDIOError(deviceInfo->DriverObject,
                            deviceInfo->DeviceObject,
                            PhysicalZero, PhysicalZero, 0,
                            0, 0, 0x62, status,
                            DIO_CANNOT_SET_THREAD_PRIORITY,
                            0, NULL, 0, NULL);
            }
        }

        // Decrement the priorityCount for the next event in list and
        // goto the next event in list.
        priorityCount--;
        nextInterruptEvent = nextInterruptEvent->NextInterruptEventInfo;
    }

    // If the flag to start task is set, go through the list again.
    if (startTask)
    {
        nextInterruptEvent = deviceInfo->InterruptEventInfoList;
        while (nextInterruptEvent != NULL)
        {

            // If startTask is still set and this event has the
            // same thread context and the scheduled flag is set,
            // start the task (set associated event) and break
            // out of loop to prevent another from being started.
            if (startTask &&
                (nextInterruptEvent->ThreadContext ==
                    interruptEventInfo->ThreadContext) &&
                nextInterruptEvent->IsScheduled)
            {
                nextInterruptEvent->IsScheduled = FALSE;
                nextInterruptEvent->IsRunning = TRUE;
                KeSetEvent(nextInterruptEvent->InterruptEventObject,
                    0, FALSE);
                break;
            }

            // Goto next in list and continue loop.
            nextInterruptEvent = nextInterruptEvent->NextInterruptEventInfo;
        }
    }
    
    // The event object of the removed event from the list must be
    // dereferenced and the event information freed.
    ObDereferenceObject(interruptEventInfo->InterruptEventObject);
    ExFreePool(interruptEventInfo);
    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Reset event tick counts for timer/external interrupt handler.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOResetInterruptEvents(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                                 IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Pointer to event information structure while traversing the list.
    INTERRUPT_EVENT_INFO *interruptEvent;

    // **** Code Statements ****

    // No data to be returned.
    pIRP->IoStatus.Information = 0;

    // No input so the input size the IRP is providing MUST be zero.
    if (pIRPStack->Parameters.DeviceIoControl.InputBufferLength != 0)
        return STATUS_INVALID_PARAMETER;

    // Since there is no data to output, verify the output size the
    // IRP is expected to return is zero.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength != 0)
        return STATUS_INVALID_PARAMETER;

    // Return immediately if there are no events in list.
    if (deviceInfo->InterruptEventInfoList == NULL)
    {
        deviceInfo->InterruptEventInfoCount = 0;
        return STATUS_SUCCESS;
    }

    // Go through the list and reset all tick counters.
    interruptEvent = deviceInfo->InterruptEventInfoList;
    while (interruptEvent != NULL)
    {

        // Reset the counter.  Note that reseting the counter
        // for tick counts of 0 (external interrupt) or 1 (run
        // upon each timer interrupt) does not have any
        // effect but does not result in any problems.
        interruptEvent->TickCounter = 0;

        // Goto the next in list.
        interruptEvent = interruptEvent->NextInterruptEventInfo;
    }
    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Enable interrupts (and handler).
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOEnableInterrupt(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                            IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Current state of PPI Port C.
    UCHAR portCState;

    // **** Code Statements ****

    // No data will be returned.
    pIRP->IoStatus.Information = 0;

    // No input so the input size the IRP is providing MUST be zero.
    if (pIRPStack->Parameters.DeviceIoControl.InputBufferLength != 0)
        return STATUS_INVALID_PARAMETER;

    // Since there is no data to output, verify the output size the
    // IRP is expected to return is zero.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength != 0)
        return STATUS_INVALID_PARAMETER;

    // Need to initialize the 'ExternalLineLow' flag if possible.  This
    // flag is used for detection of the rising of the external
    // interrupt (Port C, bit 3) line.  If the state is not set
    // properly, the first signal will be missed or the first timer
    // interrupt will be incorrectly identified as external.

    // Ensure PPI port C is set for input.
    if (((deviceInfo->PPIPortValue[3] & 0x80) != 0) &&
        ((deviceInfo->PPIPortValue[3] & 0x09) == 0x09))

    {
        if (deviceInfo->PortMemoryType == 1)
            portCState = READ_PORT_UCHAR(
                (UCHAR *) ((ULONG) deviceInfo->PPIBaseAddress +
                           (ULONG) PORT_C));
        else
            portCState = READ_REGISTER_UCHAR(
                (UCHAR *) ((ULONG) deviceInfo->PPIBaseAddress +
                           (ULONG) PORT_C));

        // If bit 3 of port C is low, set the flag showing the
        // line state as low.  Otherwise, reset the flag.
        if ((portCState & 0x08) == 0)
            deviceInfo->ExternalLineLow = TRUE;
        else
            deviceInfo->ExternalLineLow = FALSE;
    }

    // Set the interrupt handler flag.
    deviceInfo->InterruptHandled = TRUE;

    // Enable the interrupts.
    if (deviceInfo->PortMemoryType == 1)
        WRITE_PORT_UCHAR((UCHAR *) deviceInfo->EnableIRQAddress, 0x00);
    else
        WRITE_REGISTER_UCHAR((UCHAR *) deviceInfo->EnableIRQAddress, 0x00);

    // Successfully finished.
    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Disable interrupts (and handler).
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIODisableInterrupt(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                             IO_STACK_LOCATION *pIRPStack)
{

    // **** Code Statements ****

    // No data will be returned.
    pIRP->IoStatus.Information = 0;

    // No input so the input size the IRP is providing MUST be zero.
    if (pIRPStack->Parameters.DeviceIoControl.InputBufferLength != 0)
        return STATUS_INVALID_PARAMETER;

    // Since there is no data to output, verify the output size the
    // IRP is expected to return is zero.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength != 0)
        return STATUS_INVALID_PARAMETER;

    // Set the interrupt handler flag.
    deviceInfo->InterruptHandled = FALSE;

    // Disable the interrupts.
    if (deviceInfo->PortMemoryType == 1)
        WRITE_PORT_UCHAR((UCHAR *) deviceInfo->DisableIRQAddress, 0x00);
    else
        WRITE_REGISTER_UCHAR((UCHAR *) deviceInfo->DisableIRQAddress, 0x00);

    // Successfully finished.
    return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
// Get event information.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOGetEventInfo(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                         IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Structure containing the event information number of ticks (used in
    // setting a handle by DIOAddInterruptEvent.
    EVENT_INFORMATION_IOCTL_INPUT *eventInfoInput;

    // Structure containing the event information to be returned.
    EVENT_INFORMATION_IOCTL_OUTPUT *eventInfoOutput;

    // Pointer to interrupt event information structure.
    INTERRUPT_EVENT_INFO *interruptEventInfo;

    // **** Code Statements ****

    // No data yet to return.
    pIRP->IoStatus.Information = 0;

    // The input size the IRP is providing MUST be the size of the
    // EVENT_INFORMATION_IOCTL_INPUT structure.
    if (pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
         sizeof(EVENT_INFORMATION_IOCTL_INPUT))
        return STATUS_INVALID_PARAMETER;

    // The data to return is a EVENT_INFORMATION_IOCTL_OUTPUT structure.
    // Verify the output size the IRP is expected to return is the same
    // size as the structure.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength !=
        sizeof(EVENT_INFORMATION_IOCTL_OUTPUT))
        return STATUS_INVALID_PARAMETER;

    // The device IOCTL is set as METHOD_BUFFERED.  NT copies the user's
    // input buffer to the IRP's SystemBuffer entry before entry and
    // copies the IRP's SystemBuffer entry to the user's output buffer
    // after return.  Point to the SystemBuffer as both a
    // EVENT_INFORMATION_IOCTL_INPUT and a EVENT_INFORMATION_IOCTL_OUTPUT
    // structure.
    eventInfoInput =
        (EVENT_INFORMATION_IOCTL_INPUT *) pIRP->AssociatedIrp.SystemBuffer;
    eventInfoOutput =
        (EVENT_INFORMATION_IOCTL_OUTPUT *) pIRP->AssociatedIrp.SystemBuffer;

    // Go through the event information structure and find the associated
    // event with the given tick number.
    interruptEventInfo = deviceInfo->InterruptEventInfoList;
    while (interruptEventInfo != NULL)
    {
        if (interruptEventInfo->Ticks == eventInfoInput->Ticks)
        {

            // Set the appropriate values for return (note that the
            // eventInfoInput variable is clobbered and this loop
            // MUST be exited after completion of the settings!).
            eventInfoOutput->Overrun = interruptEventInfo->Overrun;
            eventInfoOutput->TimerValue = interruptEventInfo->TimerBValue;
            eventInfoOutput->Priority = interruptEventInfo->ThreadPriority;

            // Reset the overrun flag since it is being returned.
            interruptEventInfo->Overrun = FALSE;

            // Give the number of bytes returned.
            pIRP->IoStatus.Information =
                sizeof(EVENT_INFORMATION_IOCTL_OUTPUT);

            // Successfully finished.
            return STATUS_SUCCESS;
        }
        interruptEventInfo = interruptEventInfo->NextInterruptEventInfo;
    }

    // If here, event info not found.  Issue error!
    LogDIOError(deviceInfo->DriverObject, deviceInfo->DeviceObject,
        PhysicalZero, PhysicalZero, 0, 0, 0, 0x70,
        DIO_EVENT_NOT_FOUND, DIO_EVENT_NOT_FOUND,
        0, NULL, 0, NULL);
    return DIO_EVENT_NOT_FOUND;
}

////////////////////////////////////////////////////////////////////////////////
// Reset event after completion.
////////////////////////////////////////////////////////////////////////////////
NTSTATUS DIOResetEvent(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                       IO_STACK_LOCATION *pIRPStack)
{

    // **** Local Declarations ****

    // Status to return from NT calls.
    NTSTATUS status;

    // Structure containing the event information number of ticks (used in
    // setting a handle by DIOAddInterruptEvent.
    EVENT_INFORMATION_IOCTL_INPUT *eventInfoInput;

    // Pointers to interrupt event information structures.
    INTERRUPT_EVENT_INFO *interruptEventInfo;
    INTERRUPT_EVENT_INFO *nextInterruptEventInfo;

    // **** Code Statements ****

    // No data will be returned.
    pIRP->IoStatus.Information = 0;

    // The input size the IRP is providing MUST be the size of the
    // EVENT_INFORMATION_IOCTL_INPUT structure.
    if (pIRPStack->Parameters.DeviceIoControl.InputBufferLength !=
         sizeof(EVENT_INFORMATION_IOCTL_INPUT))
        return STATUS_INVALID_PARAMETER;

    // Since there is no data to output, verify the output size the
    // IRP is expected to return is zero.
    if (pIRPStack->Parameters.DeviceIoControl.OutputBufferLength != 0)
        return STATUS_INVALID_PARAMETER;

    // The device IOCTL is set as METHOD_BUFFERED.  NT copies the user's
    // input buffer to the IRP's SystemBuffer entry before entry and
    // copies the IRP's SystemBuffer entry to the user's output buffer
    // after return.  The output buffer is not used as there
    // is no output sent to the user.  Point to the SystemBuffer as
    // a EVENT_INFORMATION_IOCTL_INPUT structure.
    eventInfoInput =
        (EVENT_INFORMATION_IOCTL_INPUT *) pIRP->AssociatedIrp.SystemBuffer;

    // Go through the event information structure and find the associated
    // event with the given tick number.
    interruptEventInfo = deviceInfo->InterruptEventInfoList;
    while (interruptEventInfo != NULL)
    {
        if (interruptEventInfo->Ticks == eventInfoInput->Ticks)
        {

            // The event is complete.  Mark it as complete and reset
            // this event.
            interruptEventInfo->IsRunning = FALSE;
            interruptEventInfo->IsScheduled = FALSE;
            KeClearEvent(interruptEventInfo->InterruptEventObject);

            // Starting from this point in the interrupt event info
            // list, continue to the end and find the next event that
            // is scheduled with the same thread context.
            nextInterruptEventInfo =
                interruptEventInfo->NextInterruptEventInfo;
            while (nextInterruptEventInfo != NULL)
            {
                if ((nextInterruptEventInfo->IsScheduled) &&
                    (nextInterruptEventInfo->ThreadContext ==
                     interruptEventInfo->ThreadContext))
                {

                    // This is the next event to start.  Before the
                    // scheduled flag is EVER set, the event has been
                    // tested to be sure it is not already scheduled,
                    // running and the event is reset.  No need to
                    // check again.
                    nextInterruptEventInfo->IsScheduled = FALSE;
                    nextInterruptEventInfo->IsRunning = TRUE;
                    KeSetEvent(nextInterruptEventInfo->InterruptEventObject,
                        0, FALSE);

                    // Found and started the first one following the one
                    // that is completing that has the same thread context.
                    // Do not start any more so return successful.
                    return STATUS_SUCCESS;
                }
                nextInterruptEventInfo =
                    nextInterruptEventInfo->NextInterruptEventInfo;
            }

            // At this point the event that is signaling completion and
            // reseting the event has been found and there was no event
            // with the same thread context scheduled to start.  All
            // events for that thread context have been completed.
            // Return successful.
            return STATUS_SUCCESS;
        }

        // Goto the next in the list and continue looking.
        interruptEventInfo = interruptEventInfo->NextInterruptEventInfo;
    }

    // If here, event info not found.  Issue error!
    LogDIOError(deviceInfo->DriverObject, deviceInfo->DeviceObject,
        PhysicalZero, PhysicalZero, 0, 0, 0, 0x80,
        DIO_EVENT_NOT_FOUND, DIO_EVENT_NOT_FOUND,
        0, NULL, 0, NULL);
    return DIO_EVENT_NOT_FOUND;
}

////////////////////////////////////////////////////////////////////////////////
// Called synchronously by the ISR DPC to perform a check on the DPCInterlock
// variable (the ISR may use this variable directly).  This is to ensure
// all queued interrupts are processed and the DPC is not re-entered.
////////////////////////////////////////////////////////////////////////////////
VOID DPCInterlockOperation(VOID *context)
{

    // **** Local Declarations ****

    // Structure to hold value to send to the thread and multiprocessor safe
    // interlock operation.
    DPC_INTERLOCK_CHECK *interlockCheck = context;

    // **** Code Statements ****

    // If the DPC is being entered, check the interlock value.  If the
    // interlock value is FALSE, the DPC should process data as normal
    // and the interlock and process data flags are set. If the interlock
    // value is TRUE, there is already a DPC running.  Therefore, this
    // DPC (the one calling this at this time) should not process
    // data.
    if (interlockCheck->Entry)
    {
        if (*(interlockCheck->Interlock) == FALSE)
        {
            *(interlockCheck->Interlock) = TRUE;
            *(interlockCheck->ProcessData) = TRUE;
        }
        else
        {
            *(interlockCheck->ProcessData) = FALSE;
        }
    }

    // If the DPC is being exited, check the number of interrupts
    // queued.  If there are interrupts still queued, set the
    // flag so the DPC will continue processing.  If there are no
    // interrupts queued, reset the interlock and process data
    // flags so the DPC will finish and exit.
    else
    {
        if (*(interlockCheck->InterruptsQueued) > 0)
        {
            *(interlockCheck->ProcessData) = TRUE;
        }
        else
        {
            *(interlockCheck->Interlock) = FALSE;
            *(interlockCheck->ProcessData) = FALSE;
        }
    }
}

////////////////////////////////////////////////////////////////////////////////
// Called synchronously by the ISR DPC to extract the next interrupt from
// the queue.  The ISR may perform operations directly on the queue, but
// the DPC must call this at the same IRQL as the ISR to access the queue
// to provide thread and multiprocessor safe operations on the queue.
////////////////////////////////////////////////////////////////////////////////
VOID DPCGetQueuedInterrupt(VOID *context)
{

    // **** Local Declarations ****

    // Structure to hold pointer to the queued interrupt to send to the
    // thread and multiprocessor safe queue extraction.
    DPC_GET_QUEUED_INTERRUPT *getQueuedInterrupt = context;

    // **** Code Statements ****

    // Ensure the queue is not empty.  If empty set the interrupt
    // valid flag to false.
    if (getQueuedInterrupt->DeviceInfo->InterruptsQueued == 0)
        getQueuedInterrupt->InterruptValid = FALSE;

    // Get the interrupt type and saved timer values from the first
    // interrupt in the queue.  Then move the pointer to the first
    // queued interrupt to the next position and if past the queue
    // reset the position back to the start of the queue buffer
    // (circular queue).  Then subtract one from the number of
    // interrupts queued.
    else
    {
        getQueuedInterrupt->QueuedInterrupt->InterruptType =
            getQueuedInterrupt->DeviceInfo->InterruptQueueFirst->
                InterruptType;
        getQueuedInterrupt->QueuedInterrupt->InterruptingTimerBValue =
            getQueuedInterrupt->DeviceInfo->InterruptQueueFirst->
                InterruptingTimerBValue;
        getQueuedInterrupt->InterruptValid = TRUE;
        getQueuedInterrupt->DeviceInfo->InterruptQueueFirst++;
        if (getQueuedInterrupt->DeviceInfo->InterruptQueueFirst >=
            getQueuedInterrupt->DeviceInfo->InterruptQueueEnd)
            getQueuedInterrupt->DeviceInfo->InterruptQueueFirst =
                getQueuedInterrupt->DeviceInfo->InterruptQueue;
        getQueuedInterrupt->DeviceInfo->InterruptsQueued--;
    }
}

////////////////////////////////////////////////////////////////////////////////
// The Interrupt Service Routine.
////////////////////////////////////////////////////////////////////////////////
BOOLEAN DIO_ISR(PKINTERRUPT interrupt, void *context)
{

    // **** Local Declarations ****

    // The device object associated with the context.
    DEVICE_OBJECT *deviceObject;

    // Get the local data structure for the PCI-DIO24 device.
    DIO_DEVICE_EXTENSION *deviceInfo;

    // Counter values for timer B.
    TIMER_IOCTL_OUTPUT timerBValues;

    // Combined timer B value.
    ULONG timerB;

    // Current state of PPI Port C.
    UCHAR portCState;

    // Delta timer B value from last timer interrupt.
    ULONG deltaTimerB;

    // Flag to which kind of interrupt occured.
    INTERRUPT_TYPE interruptType;

    // **** Code Statements ****

    // Get the device object.
    deviceObject = context;

    // Get the device information.
    deviceInfo = deviceObject->DeviceExtension;

    // Clear the interrupt status.
    if (deviceInfo->PortMemoryType == 1)
        WRITE_PORT_UCHAR((UCHAR *) deviceInfo->ClearIRQAddress, 0x00);
    else
        WRITE_REGISTER_UCHAR((UCHAR *) deviceInfo->ClearIRQAddress, 0x00);

    // See if processing the interrupts have been enabled.
    if (!deviceInfo->InterruptHandled)
        return FALSE;

    // Initially set the interrupt type to INVALID.
    interruptType = INVALID;

    // If there is no clock, the interrupt could not be from
    // the timer and had to be from external interrupt line.
    if (!deviceInfo->HasClock)
        interruptType = EXTERNAL;

    // If there is a clock, immediately get the timer B time as a 32-bit
    // time (COUNTERB2 is MSW and COUNTERB1 is LSW).
    else
    {
        DIOGetTimer(deviceInfo, TIMERB, &timerBValues, FALSE);
        timerB = (((ULONG) timerBValues.Counter2Value) << 16) |
                  ((ULONG) timerBValues.Counter1Value);

        // If there is NO clock interrupt period (no timer interrupt
        // has been set), the interrupt MUST be external (C3).
        if (deviceInfo->ClockInterruptPeriod == 0)
            interruptType = EXTERNAL;

        else
        {

            //** It has been determined there is a clock, a currently
            //** set timer interrupt period, and the current 1MHz
            //** timer value has been read.  The cause of the interrupt
            //** must be determined (if it didn't have a clock or timer
            //** interrupt period, it had to be external).  For this to
            //** work correctly, a timer must be set to below the duty
            //** cycle of the external interrupt signal (C3).

            //** For example, if the external interrupt signal has an 80%
            //** duty cycle at 1Hz (80% high, 20% low), a timer interrupt
            //** must be set to trigger at least once during the 20% the
            //** signal is low and therefore must be GREATER than 5Hz.  This
            //** will detect when the signal has changed from low to high
            //** and, therefore, properly detect the external interrupt.

            // If PPI port C is not set for input, the interrupt MUST
            // be a timer interrupt.
//            if (((deviceInfo->PPIPortValue[3] & 0x80) == 0) ||
//                ((deviceInfo->PPIPortValue[3] & 0x09) != 0x09))
//                interruptType = TIMER;

            // If PPI port C is set for input, read the current state.
//            else
//            {
                if (deviceInfo->PortMemoryType == 1)
                    portCState = READ_PORT_UCHAR(
                        (UCHAR *) ((ULONG) deviceInfo->PPIBaseAddress +
                                   (ULONG) PORT_C));
                else
                    portCState = READ_REGISTER_UCHAR(
                        (UCHAR *) ((ULONG) deviceInfo->PPIBaseAddress +
                                   (ULONG) PORT_C));

                // If bit 3 of port C is low, set the flag showing the
                // line state as low.  Keep the interrupt type as INVALID!
                if ((portCState & 0x08) == 0)
                    deviceInfo->ExternalLineLow = TRUE;

                // If bit 3 of port C is NOT low and the flag shows it
                // was low, then there has been a transition.  This is
                // therefore an external interrupt.  Otherwise,
                // keep the interrupt type as INVALID!
                else if (deviceInfo->ExternalLineLow)
                {
                    deviceInfo->ExternalLineLow = FALSE;
                    interruptType = EXTERNAL;
                }

                //** Detection of EXTERNAL interrupts from bit 3 of Port C is
                //** complete and it has been determined that this interrupt
                //** could be from the timer.  Now determine if it is time for
                //** the timer interrupt.  Since it is possible the timer and
                //** external interrupt could occur too close together to
                //** generate separate interrupts, the interrupt can be from
                //** both.  Detect whether it is time for a timer interrupt
                //** and 'or' the TIMER bit in the interrupt type.  At this
                //** point, the interrupt type could be BOTH.  However, if
                //** it is NOT time for a timer interrupt, do not set
                //** the interrupt type.  It could still be INVALID if the
                //** previous interrupt was detected as an external interrupt
                //** and it was close enough to the expected time of the
                //** timer interrupt it was handled as both and yet this is
                //** the true timer interrupt a fraction of time later.  In
                //** this event, disregard this interrupt.

                // Get the delta time between the last processed timer
                // interrupt and the current interrupt.
                if (timerB > deviceInfo->ClockInterruptSavedTimerBValue)
                    deltaTimerB =
                        timerB -
                        deviceInfo->ClockInterruptSavedTimerBValue;
                else
                    deltaTimerB =
                        deviceInfo->ClockInterruptSavedTimerBValue -
                        timerB;

                // If it is NOT an external interrupt and the previous
                // interrupt was greater than TIMER_INTERRUPT_THRESHOLD away,
                // it MUST be a timer interrupt.
                if (interruptType != EXTERNAL)
                {
                    if (deltaTimerB > TIMER_INTERRUPT_THRESHOLD)
                        interruptType = TIMER;
                }

                // If it is an external interrupt, check to see if the clock
                // is within TIMER_INTERRUPT_THRESHOLD of the next expected
                // timer interrupt.
                else
                {

                    // If the delta is greater than or equal to the clock
                    // interrupt period, it is a timer interrupt.
                    if (deltaTimerB >= deviceInfo->ClockInterruptPeriod)
                        interruptType = BOTH;

                    // Fudge factorized.  If the delta is less than the clock
                    // interrupt period but within TIMER_INTERRUPT_THRESHOLD,
                    // assume it is a timer interrupt.
                    else if ((deviceInfo->ClockInterruptPeriod - deltaTimerB) <
                        TIMER_INTERRUPT_THRESHOLD)
                        interruptType = BOTH;
                }
//            }
        }
    }

    // The interrupt type should be set.  If INVALID, ignore this interrupt.
    // As previously indicated, this should only occur when 1) There is
    // a clock, 2) there is a timer set, 3) there is an external interrupt
    // source, 4) the previous interrupt was from the external source but
    // was so close to the expected timer interrupt it was handled as both,
    // and 5) this is the 'true' timer interrupt which was handled in
    // the previous interrupt.
    if (interruptType == INVALID)
        return FALSE;

    // Place the interrupt on the queue.  Note that it is acceptable to
    // access the queue directly in the ISR as all interrupts for this
    // are masked.  Note that if the interrupt can't be queued, drop
    // the interrupt and issue a system log message.
    if (deviceInfo->InterruptsQueued >= INTERRUPT_QUEUE_LENGTH)
    {
        LogDIOError(deviceInfo->DriverObject, deviceInfo->DeviceObject,
            PhysicalZero, PhysicalZero, 0, 0, 0, 0x90,
            DIO_INTERRUPT_QUEUE_FULL, DIO_INTERRUPT_QUEUE_FULL,
            0, NULL, 0, NULL);
        return TRUE;
    }

    deviceInfo->InterruptQueueNext->InterruptType = interruptType;
    deviceInfo->InterruptQueueNext->InterruptingTimerBValue = timerB;
    deviceInfo->InterruptQueueNext++;
    if (deviceInfo->InterruptQueueNext >= deviceInfo->InterruptQueueEnd)
        deviceInfo->InterruptQueueNext = deviceInfo->InterruptQueue;
    deviceInfo->InterruptsQueued++;

    // If there is not a DPC currently executing, queue the DPC to process
    // the queued interrupt.  If there is a DPC currently executing, the
    // queued IRQ will be processed.  At the current IRQL it is safe to
    // access the DPCInterlock variable directly.
    if (!deviceInfo->DPCInterlock)
        IoRequestDpc(deviceInfo->DeviceObject, NULL, NULL);

    return TRUE;
}

////////////////////////////////////////////////////////////////////////////////
// The Deferred Procedure Call for the ISR to complete at DISPATCH IRQL.
////////////////////////////////////////////////////////////////////////////////
VOID DIO_ISRDPC(KDPC *pDPC, DEVICE_OBJECT *deviceObject, IRP *pIRP,
                void *context)
{

    // **** Local Declarations ****

    // Get the local data structure for the PCI-DIO24 device.
    DIO_DEVICE_EXTENSION *deviceInfo;

    // Structure to hold value to send to the thread and multiprocessor safe
    // interlock operation.
    DPC_INTERLOCK_CHECK interlockCheck;

    // Set to allow the DPC to loop.  This allows the DPC to identify any
    // IRQ's that come in during processing of the current IRQ.
    BOOLEAN moreProcessing;

    // Structure to hold pointer to the queued interrupt to send to the
    // thread and multiprocessor safe queue extraction.
    DPC_GET_QUEUED_INTERRUPT getQueuedInterrupt;

    // Value for the queued interrupt.
    INTERRUPT_QUEUE queuedInterrupt;

    // Flag whether a scheduled event was started.
    BOOLEAN eventStarted;

    // Pointer to interrupt event information structure.
    INTERRUPT_EVENT_INFO *interruptEventInfo;

    // Never referenced parameters.
    UNREFERENCED_PARAMETER(pDPC);
    UNREFERENCED_PARAMETER(pIRP);
    UNREFERENCED_PARAMETER(context);

    // **** Code Statements ****

    // Get the device information.
    deviceInfo = deviceObject->DeviceExtension;

    // The DPCInterlock value in the device information is used by the ISR
    // and DPC to prevent re-entrency of the DPC.  If the moreProcessing
    // value is TRUE, then the DPC should attempt to process events in the
    // queue.  If the value of moreProcessing is FALSE, the DPC should
    // exit without processing any data.  The synchronized call to
    // DPCInterlockOperation provides a thread and multiprocessor safe
    // method to access the value of the DPCInterlock value.  The
    // interlock value is used by the ISR to schedule a DPC if it is not
    // set.  This call with entry set to TRUE will set the interlock
    // if it is not already set.
    interlockCheck.Interlock = &deviceInfo->DPCInterlock;
    interlockCheck.Entry = TRUE;
    interlockCheck.ProcessData = &moreProcessing;
    interlockCheck.InterruptsQueued = &deviceInfo->InterruptsQueued;
    KeSynchronizeExecution(deviceInfo->Interrupt,
                           (PKSYNCHRONIZE_ROUTINE) DPCInterlockOperation,
                           (PVOID) &interlockCheck);

    // Processing loop.  The DPC will loop until all queued interrupts
    // have been processed.  If an IRQ is generated during DPC processing,
    // the number of interrupts queued will be incremented again and
    // result in more than one iteration within the loop.
    while (moreProcessing)
    {

        // Get the first interrupt to be processed from the queue.  This
        // must be done thread/multi-processor safe with a synchronized
        // execution (at the ISR IRQL).  Prepare for the call with the
        // pointer to a INTERRUPT_QUEUE pointer returned as the
        // next interrupt in the queue to process.
        getQueuedInterrupt.DeviceInfo = deviceInfo;
        getQueuedInterrupt.QueuedInterrupt = &queuedInterrupt;
        getQueuedInterrupt.InterruptValid = FALSE;
        KeSynchronizeExecution(deviceInfo->Interrupt,
                               (PKSYNCHRONIZE_ROUTINE) DPCGetQueuedInterrupt,
                               (PVOID) &getQueuedInterrupt);

        // Has a valid queued interrupt, process the interrupt.
        if (getQueuedInterrupt.InterruptValid)
        {

            // Process until the queue is empty.
            while (getQueuedInterrupt.InterruptValid)
            {


                // Process timer interrupts.
                if ((queuedInterrupt.InterruptType == TIMER) ||
                    (queuedInterrupt.InterruptType == BOTH))
                {

                    // Save the timer interrupt delta time.
                    deviceInfo->ClockInterruptSavedTimerBValue =
                        queuedInterrupt.InterruptingTimerBValue;

                    // Get a new thread context for this pass (for bookkeeping
                    // only and is OK to wrap to 0).
                    deviceInfo->ThreadContext++;

                    // Traverse the interrupt event list and find ALL items with
                    // a non-zero tick count to schedule and start the first and
                    // only the first one scheduled.
                    eventStarted = FALSE;
                    interruptEventInfo = deviceInfo->InterruptEventInfoList;
                    while (interruptEventInfo != NULL)
                    {
                        if (interruptEventInfo->Ticks > 0)
                        {

                            // Increment the tick counter and check to see if it
                            // is greater than or equal to the desired ticks.
                            interruptEventInfo->TickCounter++;
                            if (interruptEventInfo->TickCounter >=
                                interruptEventInfo->Ticks)
                            {

                                // Have reached time to schedule, reset the
                                // counter.
                                interruptEventInfo->TickCounter = 0;

                                // Time to schedule.  Ensure it is not already
                                // scheduled or running and that the event has
                                // been reset.  If it is scheduled or running or
                                // the event is still set, it is an overrun so
                                // break out of loop and don't schedule anything
                                // else!
                                if (interruptEventInfo->IsScheduled ||
                                    interruptEventInfo->IsRunning)
                                    interruptEventInfo->Overrun = TRUE;
    
                                // Start the first and only the first task for
                                // this timer interrupt.  Check to be sure an
                                // event has not already been started in this
                                // loop.
                                else if (!eventStarted)
                                {

                                    // This is the first event that should be
                                    // scheduled due to the timer interrupt.
                                    // This event should be started immediately,
                                    // set the running flag, set the time, set
                                    // the thread context and set the event to
                                    // start the task.
                                    interruptEventInfo->IsRunning = TRUE;
                                    interruptEventInfo->TimerBValue =
                                        queuedInterrupt.InterruptingTimerBValue;
                                    interruptEventInfo->ThreadContext =
                                        deviceInfo->ThreadContext;
                                    KeSetEvent(
                                        interruptEventInfo->InterruptEventObject,
                                         0, FALSE);

                                    // Set the flag that an event has already been
                                    // started so another won't be started
                                    // immediately.
                                    eventStarted = TRUE;
                                }

                                // If an event has already been started.  Simply
                                // mark the event to start by setting the
                                // scheduled flag, the time and the thread
                                // context.
                                else
                                {
                                    interruptEventInfo->IsScheduled = TRUE;
                                    interruptEventInfo->TimerBValue =
                                        queuedInterrupt.InterruptingTimerBValue;
                                    interruptEventInfo->ThreadContext =
                                        deviceInfo->ThreadContext;
                                }
                            }
                        }

                        // Goto next interrupt event in list.
                        interruptEventInfo =
                            interruptEventInfo->NextInterruptEventInfo;
                    }
                }

                // Process external interrupts.
                if ((queuedInterrupt.InterruptType == EXTERNAL) ||
                    (queuedInterrupt.InterruptType == BOTH))
                {

                    // Check the interrupt event list.
                    if (deviceInfo->InterruptEventInfoList != NULL)
                    {
                        interruptEventInfo = deviceInfo->InterruptEventInfoList;

                        // Check if first in list has a tick count of 0 (the
                        // external interrupt event info).  Note that if it is
                        // present it is the first in the list.
                        if (interruptEventInfo->Ticks == 0)
                        {

                            // Ensure the event is NOT currently running (can't
                            // be scheduled since there can be no other events
                            // scheduled for external interrupts).  If running,
                            // then set the overrun flag.  Also check to be sure
                            // the event is reset.  If set, then it is also an
                            // overrun.
                            if (interruptEventInfo->IsRunning)
                                interruptEventInfo->Overrun = TRUE;

                            // The event was not scheduled or running.  Set the
                            // scheduled and running flag, set the time, set the
                            // thread context and set the event to start the
                            // task.  Note that the thread context is only for
                            // internal book-keeping and is not for the external
                            // interrupt event which can only have one event so it
                            // is set explicitly to zero.  Also note that the
                            // external event will always be immediately run
                            // so the IsScheduled flag has no meaning and is
                            // left to be FALSE.
                            else
                            {
                                interruptEventInfo->IsRunning = TRUE;
                                interruptEventInfo->TimerBValue =
                                    queuedInterrupt.InterruptingTimerBValue;
                                interruptEventInfo->ThreadContext = 0;
                                KeSetEvent(
                                    interruptEventInfo->InterruptEventObject, 0,
                                    FALSE);
                            }
                        }
                    }
                }

                // Get the next interrupt to be processed from the queue.  This
                // must be done thread/multi-processor safe with a synchronized
                // execution (at the ISR IRQL).  Prepare for the call with the
                // pointer to a INTERRUPT_QUEUE pointer returned as the
                // next interrupt in the queue to process.  Note that this will
                // attempt to get subsequent interrupts from the queue and it
                // is normal if it returns an InterruptValid of FALSE in which
                // case the InterruptValue loop will exit.
                getQueuedInterrupt.DeviceInfo = deviceInfo;
                getQueuedInterrupt.QueuedInterrupt = &queuedInterrupt;
                getQueuedInterrupt.InterruptValid = FALSE;
                KeSynchronizeExecution(deviceInfo->Interrupt,
                                       (PKSYNCHRONIZE_ROUTINE) DPCGetQueuedInterrupt,
                                       (PVOID) &getQueuedInterrupt);
            }
        }

        // Attempt to exit the DPC by resetting the DPCInterlock.  If the
        // moreProcessing flag is set on return, the interrupt queue was not
        // empty and the DPC should begin to process queued interrupts again.
        // If the queue is empty, the interlock will be reset and the
        // moreProcessing flag will be reset and the DPC will exit the
        // moreProcessing loop and terminate.
        interlockCheck.Interlock = &deviceInfo->DPCInterlock;
        interlockCheck.Entry = FALSE;
        interlockCheck.ProcessData = &moreProcessing;
        interlockCheck.InterruptsQueued = &deviceInfo->InterruptsQueued;
        KeSynchronizeExecution(deviceInfo->Interrupt,
                               (PKSYNCHRONIZE_ROUTINE) DPCInterlockOperation,
                               (PVOID) &interlockCheck);
    }
}
