////////////////////////////////////////////////////////////////////////////////
//
//      Title:          PCIDIO24.h
//
//      Date Created:   7 August 1999
//
//      Created By:     Ben-z Lawrence
//      Organization:   Control Systems Research
//
//      Target OS:      Microsoft Windows 2000
//      Compiler:       Microsoft Visual C++ 6.0
//
//      Description:    Include file for PCI-DIO24 Port I/O Driver.  Defines
//                      driver function prototypes and structures.
//
//      Modifications:
//
//      Version 1.1:    Ben-z Lawrence; August 15, 2000
//
//                      Error discovered in methodology of ISR.  Added DPC
//                      to complete ISR processing (as required by NT/2000)
//                      [function DIO_ISRDPC] and moved actual processing
//                      of thread control to the DPC.  Placed
//                      multiprocessor-safe code in DPC to prevent re-entrancy
//                      and a function [DPCInterlockOperation] to prevent
//                      concurrent access to the variable used to prevent
//                      re-entrancy.  Added queuing of the interrupts received
//                      for processing of the interrupts by the DPC and a
//                      function [DPCGetQueuedInterrupt] to retrieve queued
//                      interrupts in a thread/multiprocessor safe method.
//
//                      Added multiprocessor-safe code and functions to
//                      counter/timer set/read functions [DIOSyncCounterSet,
//                      DIOSyncCounterRead, and DIOSyncTimerRead].
//
//                      Added check to ensure interrupt events are not added
//                      or removed while interrupts are enabled.
//
//      Version 1.2:    Ben-z Lawrence; August 14, 2001
//
//                      Required capability of synchronizing the timer
//                      interrupt to the external interrupt and changing the
//                      timer counter reload values at termination of current
//                      count with new values without stopping the counters.
//
//                      The DIOWriteCounter was added.
//
//
//      Version 1.3:    Ben-z Lawrence; December 10, 2001
//
//                      Required capability of synchronizing all the interrupt
//                      events timing with the external interrupt.  The
//                      internal tick counters can be reset (will generally
//                      be called by the user's external interrupt event
//                      handler).
//
//                      The DIOResetInterruptEvents was added.
//
////////////////////////////////////////////////////////////////////////////////

// Global constant defining a zero physical address.
static const PHYSICAL_ADDRESS PhysicalZero = {0};

// Registry key name for flagging whether the card has a clock or not.
#define HAS_CLOCK_KEY_NAME L"DIOHasClock"

// First part of interrupt event name and maximum length of name.
#define INTERRUPT_EVENT_NAME L"EVENT_TICK_"
#define INTERRUPT_EVENT_NAME_LENGTH 64

// Define an invalid handle value (don't include WINBASE.H!).
#define INVALID_HANDLE_VALUE ((HANDLE)-1)

// Define the number of concurrent interrupts the driver can queue.
#define INTERRUPT_QUEUE_LENGTH 10

// Set the maximum number of interrupt events.  Note that due to
// priorities, this maximum MUST be less than 16.
#define MAX_INTERRUPT_EVENTS 10

// Timer interrupt threshold in microseconds (to determine interrupt cause).
#define TIMER_INTERRUPT_THRESHOLD 20

// Number of ports without and with clock.
#define NUM_PORTS_NO_CLOCK 0x10
#define NUM_PORTS_WITH_CLOCK 0x20

// Offsets to addresses.
#define PPI_OFFSET  0x00
#define BUFFER_ENABLE_OFFSET 0x0a
#define IRQ_DISABLE_OFFSET 0x0d
#define IRQ_ENABLE_OFFSET 0x0e
#define IRQ_CLEAR_OFFSET 0x0f
#define TIMER_A_OFFSET 0x10
#define TIMER_B_OFFSET 0x14
#define TIMER_C_OFFSET 0x18

// Structure and external to hold global values during driver installation.
typedef struct _DIO_GLOBALS
{
    void *PAGEDIO_Handle;
    BOOLEAN IsLocked;
    UNICODE_STRING RegistryPath;
} DIO_GLOBALS;
extern DIO_GLOBALS DIOGlobals;

// Possible device states.
typedef enum _DEVICE_PNP_STATE
{
    NotStarted = 0,         // Not started yet
    Started,                // Device has received the START_DEVICE IRP
    StopPending,            // Device has received the QUERY_STOP IRP
    Stopped,                // Device has received the STOP_DEVICE IRP
    RemovePending,          // Device has received the QUERY_REMOVE IRP
    SurpriseRemovePending,  // Device has received the SURPRISE_REMOVE IRP
    Deleted                 // Device has received the REMOVE_DEVICE IRP
} DEVICE_PNP_STATE;

// Interrupt event information.
typedef struct _INTERRUPT_EVENT_INFO
{
    ULONG Ticks;
    ULONG TickCounter;
    HANDLE InterruptEventHandle;
    KEVENT *InterruptEventObject;
    ULONG ThreadContext;
    HANDLE ThreadHandle;
    KPRIORITY ThreadPriority;
    BOOLEAN IsScheduled;
    BOOLEAN IsRunning;
    BOOLEAN Overrun;
    ULONG TimerBValue;
    struct _INTERRUPT_EVENT_INFO *NextInterruptEventInfo;
} INTERRUPT_EVENT_INFO;

// Type of interrupt for ISR to give to DPC.
typedef enum _INTERRUPT_TYPE
{
    INVALID = 0,
    TIMER = 1,
    EXTERNAL = 2,
    BOTH = 3
} INTERRUPT_TYPE;

// Structure for synchronized read of counter value.
typedef struct _SYNC_COUNTER
{
    ULONG PortMemoryType;
    VOID *TimerControlAddress;
    UCHAR TimerControl;
    VOID *CounterDataAddress;
    UCHAR *CounterLSB;
    UCHAR *CounterMSB;
} SYNC_COUNTER;

// Structure for synchronized read of timer value.
typedef struct _SYNC_TIMER
{
    ULONG PortMemoryType;
    VOID *TimerControlAddress;
    UCHAR TimerControl;
    VOID *Counter0DataAddress;
    VOID *Counter1DataAddress;
    VOID *Counter2DataAddress;
    USHORT *Counter0Value;
    USHORT *Counter1Value;
    USHORT *Counter2Value;
} SYNC_TIMER;

// Structure for interrupt queue.
typedef struct _INTERRUPT_QUEUE
{
    ULONG InterruptingTimerBValue;
    INTERRUPT_TYPE InterruptType;
} INTERRUPT_QUEUE;

// Driver data structure for device object.
typedef struct _DIO_DEVICE_EXTENSION
{
    DRIVER_OBJECT *DriverObject;
    DEVICE_OBJECT *PhysicalDeviceObject;
    DEVICE_OBJECT *DeviceObject;
    DEVICE_PNP_STATE DevicePnPState;
    ULONG CurrentDeviceState;
    BOOLEAN HoldNewRequests;
    LIST_ENTRY NewRequestsQueue;
    KSPIN_LOCK QueueLock;
    KEVENT RemoveEvent;
    KEVENT StopEvent;
    ULONG OutstandingIO;
    ULONG IsOpened;
    SYSTEM_POWER_STATE SystemPowerState;
    DEVICE_POWER_STATE DevicePowerState;
    UCHAR  *PPIPortValue;
    DEVICE_OBJECT *LowerDeviceObject;
    UNICODE_STRING DeviceClassSymbolicName;
    BOOLEAN HasClock;
    INTERFACE_TYPE InterfaceType;
    ULONG BusNumber;
    PHYSICAL_ADDRESS DeviceBusAddress;
    ULONG DeviceBusAddressLength;
    ULONG PortMemoryType;
    KIRQL DeviceIRQL;
    ULONG DeviceIRQV;
    KAFFINITY DeviceIRQAffinity;
    KINTERRUPT_MODE DeviceInterruptMode;
    PHYSICAL_ADDRESS DeviceTrBusAddress;
    ULONG DeviceTrBusAddressLength;
    KIRQL DeviceTrIRQL;
    ULONG DeviceTrIRQV;
    VOID *VirtualAddress;
    BOOLEAN Mapped;
    VOID *PPIBaseAddress;
    VOID *BufferEnableAddress;
    VOID *DisableIRQAddress;
    VOID *EnableIRQAddress;
    VOID *ClearIRQAddress;
    VOID *CounterA0;
    VOID *CounterA1;
    VOID *CounterA2;
    VOID *TimerControlA;
    VOID *CounterB0;
    VOID *CounterB1;
    VOID *CounterB2;
    VOID *TimerControlB;
    VOID *CounterC0;
    VOID *CounterC1;
    VOID *CounterC2;
    VOID *TimerControlC;
    PKSERVICE_ROUTINE DIOIsr;
    VOID *DIOIsrContext;
    PKINTERRUPT Interrupt;
    BOOLEAN InterruptHandled;
    BOOLEAN ClockPeriodicInterruptA1Set;
    BOOLEAN ClockPeriodicInterruptA2Set;
    ULONG ClockInterruptPeriod;
    ULONG ClockInterruptSavedTimerBValue;
    BOOLEAN ExternalLineLow;
    ULONG InterruptEventInfoCount;
    INTERRUPT_EVENT_INFO *InterruptEventInfoList;
    ULONG ThreadContext;
    KSPIN_LOCK InterruptSpinLock;
    BOOLEAN DPCInterlock;
    INTERRUPT_QUEUE *InterruptQueue;
    INTERRUPT_QUEUE *InterruptQueueFirst;
    INTERRUPT_QUEUE *InterruptQueueNext;
    INTERRUPT_QUEUE *InterruptQueueEnd;
    ULONG InterruptsQueued;
} DIO_DEVICE_EXTENSION;

// Operation context for DPCInterlockOperation.
typedef struct _DPC_INTERLOCK_CHECK
{
    BOOLEAN *Interlock;
    BOOLEAN Entry;
    BOOLEAN *ProcessData;
    ULONG *InterruptsQueued;
} DPC_INTERLOCK_CHECK;

// Operation context for DPCGetQueuedInterrupt
typedef struct _DPC_GET_QUEUED_INTERRUPT
{
    DIO_DEVICE_EXTENSION *DeviceInfo;
    INTERRUPT_QUEUE *QueuedInterrupt;
    BOOLEAN InterruptValid;
} DPC_GET_QUEUED_INTERRUPT;

// Driver entry.
NTSTATUS DriverEntry(DRIVER_OBJECT *driverObject,
                     UNICODE_STRING *registryPath);

// Increments the number of requests the device receives.
LONG DIOIoIncrement(DIO_DEVICE_EXTENSION *deviceInfo);

// Decrements the number of requests as the request completes.
LONG DIOIoDecrement(DIO_DEVICE_EXTENSION *deviceInfo);

// Synchronization completion call-back - set event flag when complete.
NTSTATUS DIOSyncCompletion(DEVICE_OBJECT *deviceObject, IRP *pIRP,
                           KEVENT *syncEvent);

// Power Synchronization completion call-back from PoRequestPowerIrp.
void DIOPoRequestComplete(DEVICE_OBJECT *deviceObject, UCHAR minorFunction,
                          POWER_STATE powerState, IRP *pIRP,
                          IO_STATUS_BLOCK *ioStatus);

// Power completion routine call-back from PoCallDriver.
NTSTATUS DIOPowerCompletionRoutine(DEVICE_OBJECT *deviceObject, IRP *pIRP,
                                   void *context);

// Queue IRP.
NTSTATUS DIOQueueRequest(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP);

// Cancel queued IRP call-back.
void DIOCancelQueued(DEVICE_OBJECT *deviceObject, IRP *pIRP);

// Process queued IRPs.
VOID DIOProcessQueuedRequests(DIO_DEVICE_EXTENSION *deviceInfo);

// Create the device object.
NTSTATUS DIOAddDevice(DRIVER_OBJECT *driverObject,
                      DEVICE_OBJECT *lowerDeviceObject);

// Dispatch plug and play requests.
NTSTATUS DIOPnpDispatch(DEVICE_OBJECT *deviceObject, IRP *pIRP);

// Dispatch power requests.
NTSTATUS DIOPowerDispatch(DEVICE_OBJECT *deviceObject, IRP *pIRP);

// Create or close a device handle (when the instance is created or
// closed by a user).
NTSTATUS DIOCreateClose(DEVICE_OBJECT *deviceObject, IRP *pIRP);

// Clean-up before closing the file handle (called before IRP_MJ_CLOSE).
NTSTATUS DIOCleanup(DEVICE_OBJECT *deviceObject, IRP *pIRP);

// Unload the device driver from the system.  Free all resources allocated or
// registed and symbolic links created by the driver.
void DIOUnload(DRIVER_OBJECT *driverObject);

// Dispatch handler for the PCI-DIO24 driver.  Handles I/O Request Packets
// (IRP) from the I/O Manager.  Specifically an application call to
// DeviceIoControl is routed to the dispatch handler which performs the
// appropriate action as given in the IO Control Code.
NTSTATUS DIODispatch(DEVICE_OBJECT *deviceObject, IRP *pIRP);

// Process the IOCTL_DIO_SETMODE to set the I/O mode of the given
// port of the PPI.
NTSTATUS DIOSetMode(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                    IO_STACK_LOCATION *pIRPStack);

// Process the IOCTL_DIO_WRITE to write byte to the given port of the PPI.
NTSTATUS DIOWrite(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                  IO_STACK_LOCATION *pIRPStack);

// Process the IOCTL_DIO_SET to set specified bit(s) to the given port
// of the PPI.
NTSTATUS DIOSet(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                IO_STACK_LOCATION *pIRPStack);

// Process the IOCTL_DIO_RESET to reset specified bit(s) to the given port
// of the PPI.
NTSTATUS DIOReset(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                  IO_STACK_LOCATION *pIRPStack);

// Process the IOCTL_CIO_READ to read byte from the given port of the PPI.
NTSTATUS DIORead(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                 IO_STACK_LOCATION *pIRPStack);

// Process the IOCTL_DIO_INIT_COUNTER to initialize the given timer counter.
NTSTATUS DIOInitCounter(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                        IO_STACK_LOCATION *pIRPStack);

// Process the IOCTL_DIO_WRITE_COUNTER to provide a new count value for
// the given timer counter.
NTSTATUS DIOWriteCounter(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                         IO_STACK_LOCATION *pIRPStack);

// Set the given timer counter value.
void DIOSetCounter(DIO_DEVICE_EXTENSION *deviceInfo,
                   COUNTER_IOCTL_INPUT *counterInput,
                   BOOLEAN initialize);

// Thread/multiprocessor safe call to set counters.
VOID DIOSyncCounterSet(VOID *context);

// Process the IOCTL_DIO_READ_COUNTER to read the given timer counter value.
NTSTATUS DIOReadCounter(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                        IO_STACK_LOCATION *pIRPStack);

// Get the current value of the given counter.
USHORT DIOGetCounter(DIO_DEVICE_EXTENSION *deviceInfo, enum COUNTER counter);

// Thread/multiprocessor safe call to read counters.
VOID DIOSyncCounterRead(VOID *context);

// Process the IOCTL_DIO_READ_TIMER to read all counters of the given
// timer latched together.
NTSTATUS DIOReadTimer(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                      IO_STACK_LOCATION *pIRPStack);

// Get the current value of all counters latched together.
void DIOGetTimer(DIO_DEVICE_EXTENSION *deviceInfo, enum TIMER timer,
                 TIMER_IOCTL_OUTPUT *counterValues, BOOLEAN syncISR);

// Thread/multiprocessor safe call to read timers.
VOID DIOSyncTimerRead(VOID *context);

// Respond to a read or write request.  This driver does not support reads or
// writes.
NTSTATUS DIOReadWrite(DEVICE_OBJECT *deviceObject, IRP *pIRP);

// Send an error to the event logger.
void LogDIOError(DRIVER_OBJECT *driverObject, DEVICE_OBJECT *deviceObject,
                 PHYSICAL_ADDRESS p1, PHYSICAL_ADDRESS p2,
                 ULONG sequenceNumber, UCHAR majorFunctionCode,
                 UCHAR retryCount, ULONG uniqueErrorValue,
                 NTSTATUS finalStatus, NTSTATUS specificIOStatus,
                 ULONG lengthOfInsert1, WCHAR *insert1,
                 ULONG lengthOfInsert2, WCHAR *insert2);

// Set the timer interrupt period.
NTSTATUS DIOSetTimerInterruptPeriod(DIO_DEVICE_EXTENSION *deviceInfo,
                                    IRP *pIRP,
                                    IO_STACK_LOCATION *pIRPStack);

// Add event for timer/external interrupt handler to set.
NTSTATUS DIOAddInterruptEvent(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                              IO_STACK_LOCATION *pIRPStack);

// Remove event for timer/external interrupt handler.
NTSTATUS DIORemoveInterruptEvent(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                                 IO_STACK_LOCATION *pIRPStack);

// Reset event tick counts for timer/external interrupt handler.
NTSTATUS DIOResetInterruptEvents(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                                 IO_STACK_LOCATION *pIRPStack);

// Enable interrupts (and handler).
NTSTATUS DIOEnableInterrupt(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                            IO_STACK_LOCATION *pIRPStack);

// Disable interrupts (and handler).
NTSTATUS DIODisableInterrupt(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                             IO_STACK_LOCATION *pIRPStack);

// Get event information.
NTSTATUS DIOGetEventInfo(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                         IO_STACK_LOCATION *pIRPStack);

// Reset event after completion.
NTSTATUS DIOResetEvent(DIO_DEVICE_EXTENSION *deviceInfo, IRP *pIRP,
                       IO_STACK_LOCATION *pIRPStack);

// Thread and multiprocessor safe variable increment and decrement.
VOID DPCInterlockOperation(VOID *context);

// Thread and multiprocessor safe interrupt queue extraction.
VOID DPCGetQueuedInterrupt(VOID *context);

// The Interrupt Service Routine (ISR).
BOOLEAN DIO_ISR(PKINTERRUPT interrupt, void *context);

// The Deferred Procedure Call (DPC) for the ISR (DpcForIsr) routine.
VOID DIO_ISRDPC(KDPC *pDPC, DEVICE_OBJECT *deviceObject, IRP *pIRP,
                void *context);
